Package: igvShiny
Title: igvShiny: a wrapper of Integrative Genomics Viewer (IGV - an
        interactive tool for visualization and exploration integrated
        genomic data)
Version: 1.9.43
Date: 2026-07-29
Authors@R: c(
    person("Paul","Shannon", role = c("aut"), 
    email = "paul.thurmond.shannon@gmail.com"),
    person("Arkadiusz", "Gladki", role=c("aut", "cre"), 
    email="gladki.arkadiusz@gmail.com", comment = c(ORCID = "0000-0002-7059-6378")),
    person("Karolina","Scigocka", role = c("aut"),
    email = "scigocka.karolina@gmail.com"),
    person("Carolina", "Heimann", role = c("ctb"),
    email = "carolina.heimann@gmail.com"),
    person("Steffen", "Klasberg", role = c("ctb"),
    email = "klasberg@dkms-lab.de"),
    person("Vincent", "Carey", role = c("ctb"),
    email = "stvjc@channing.harvard.edu"),
    person("Parv", "Sachdeva", role = c("ctb"),
    email = "parv.sachdeva@sonomabio.com"),
    person("Mateusz", "Gladki", role = c("ctb"),
    email = "gladkiimateusz@gmail.com"),
    person("R Consortium", role = "fnd")
    )
Description: This package is a wrapper of Integrative Genomics Viewer (IGV).
    It comprises an htmlwidget version of IGV. It can be used as a module 
    in Shiny apps. 
URL: https://github.com/gladkia/igvShiny,
        https://gladkia.github.io/igvShiny/
BugReports: https://github.com/gladkia/igvShiny/issues
Depends: R (>= 3.5.0), GenomicRanges, methods, shiny
Imports: BiocGenerics, checkmate, futile.logger, GenomeInfoDbData,
        htmlwidgets, httr, jsonlite, randomcoloR, utils
Suggests: BiocStyle, bslib, covr, GenomicAlignments, knitr, Rsamtools,
        rtracklayer, testthat, shinytest2, VariantAnnotation
biocViews: Software, ShinyApps, Sequencing, Coverage
License: MIT + file LICENSE
Encoding: UTF-8
Collate: 'constants.R' 'genomeSpec.R' 'GWASTrack.R' 'igvShiny.R'
        'packages.R' 'utils.R' 'zzz.R'
RoxygenNote: 7.3.3
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/igvShiny
git_branch: devel
git_last_commit: d5bca1f
git_last_commit_date: 2026-08-22
Repository: Bioconductor 3.24
Date/Publication: 2026-08-23
NeedsCompilation: no
Packaged: 2026-08-23 22:36:47 UTC; biocbuild
Author: Paul Shannon [aut],
  Arkadiusz Gladki [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7059-6378>),
  Karolina Scigocka [aut],
  Carolina Heimann [ctb],
  Steffen Klasberg [ctb],
  Vincent Carey [ctb],
  Parv Sachdeva [ctb],
  Mateusz Gladki [ctb],
  R Consortium [fnd]
Maintainer: Arkadiusz Gladki <gladki.arkadiusz@gmail.com>

# Generated by roxygen2: do not edit by hand

export(GWASTrack)
export(getGenomicRegion)
export(get_basic_genomes)
export(get_cas_genomes)
export(get_css_genomes)
export(get_tracks_dir)
export(igvShiny)
export(igvShinyOutput)
export(loadBamTrackFromLocalData)
export(loadBamTrackFromURL)
export(loadBedGraphTrack)
export(loadBedGraphTrackFromURL)
export(loadBedTrack)
export(loadCramTrackFromURL)
export(loadGFF3TrackFromLocalData)
export(loadGFF3TrackFromURL)
export(loadGwasTrack)
export(loadSegTrack)
export(loadVcfTrack)
export(parseAndValidateGenomeSpec)
export(removeTracksByName)
export(removeUserAddedTracks)
export(renderIgvShiny)
export(showGenomicRegion)
exportClasses(GWASTrack)
exportMethods(display)
exportMethods(getUrl)
import(BiocGenerics)
import(GenomeInfoDbData)
import(GenomicRanges)
import(httr)
import(shiny)
importFrom(futile.logger,flog.debug)
importFrom(htmlwidgets,createWidget)
importFrom(htmlwidgets,shinyRenderWidget)
importFrom(htmlwidgets,shinyWidgetOutput)
importFrom(methods,is)
importFrom(methods,new)
importFrom(randomcoloR,distinctColorPalette)
importFrom(utils,write.table)

# igvShiny

<!-- badges: start -->
[![Bioconductor build](https://bioconductor.org/shields/build/release/bioc/igvShiny.svg)](https://bioconductor.org/packages/release/bioc/html/igvShiny.html)
[![R-CMD-check (BiocCheck)](https://github.com/gladkia/igvShiny/actions/workflows/check-bioc.yml/badge.svg)](https://github.com/gladkia/igvShiny/actions/workflows/check-bioc.yml)
[![Bioconductor downloads](https://bioconductor.org/shields/downloads/release/igvShiny.svg)](https://bioconductor.org/packages/stats/bioc/igvShiny/)
[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](LICENSE.md)
<!-- badges: end -->

An [htmlwidget](https://www.htmlwidgets.org/) wrapper of the
[Integrative Genomics Viewer (IGV)](https://igv.org/) — embed an interactive
genome browser in your [Shiny](https://shiny.posit.co/) apps, and drive it from
R. One of only two Bioconductor packages bridging IGV to R.

## 🔬 Live demo

**[gladkia-igvshiny-demo.share.connect.posit.cloud](https://gladkia-igvshiny-demo.share.connect.posit.cloud)**

Click through BED / BedGraph / GWAS / BAM / CRAM tracks in a running app — no
install required. (Hosted on [Posit Connect Cloud](https://connect.posit.cloud/);
source in [`demo/posit-connect/`](demo/posit-connect/).)

![igvShiny demo — hg38 BAM and CRAM alignment tracks over the APP region on chr21 shown as stacked panels in the embedded IGV browser, alongside the bslib control sidebar](man/figures/demo.jpg)

## Installation

Release version from [Bioconductor](https://bioconductor.org/packages/igvShiny):

```r
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("igvShiny")
```

Development version from GitHub:

```r
remotes::install_github("gladkia/igvShiny")
```

## Quick start

```r
library(shiny)
library(igvShiny)

options <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "NDUFS2")

ui <- fluidPage(
  igvShinyOutput("igv")
)

server <- function(input, output, session) {
  output$igv <- renderIgvShiny({
    igvShiny(options)
  })
}

shinyApp(ui, server)
```

From there, load tracks reactively with the `load*Track*` functions
(`loadBedTrack`, `loadBedGraphTrack`, `loadGwasTrack`, `loadBamTrackFromURL`,
`loadCramTrackFromURL`, …) and move the view with `showGenomicRegion()`.

## Features

- Interactive IGV genome browser as a Shiny `htmlwidget`, usable as a Shiny module.
- Stock genomes (hg38, hg19, mm10, tair10, …) and custom genomes from local or remote FASTA.
- Track loaders for BED, BedGraph, bed9, GWAS, SEG, VCF, BAM (URL / local), and CRAM (URL).
- Navigate and query the current view from R (`showGenomicRegion()`, `getGenomicRegion()`).
- Track-click events surfaced back to the Shiny server.

Runnable apps live in [`inst/demos/`](inst/demos/), one per topic:
`igvShinyDemo.R` (most of the API — this is what the
[live demo](https://gladkia-igvshiny-demo.share.connect.posit.cloud) serves),
`tiny.R`, `genomes.R`, `gwas.R`, `local-data.R`, `modules.R`,
`two-instances.R`. Run one with
`shiny::runApp(system.file("demos", "tiny.R", package = "igvShiny"))`.

## Documentation

- 📦 Reference & articles: <https://gladkia.github.io/igvShiny/>
- 📖 Vignette: `vignette("igvShiny")`
- 🐛 Issues / feature requests: <https://github.com/gladkia/igvShiny/issues>

## Contributing

Contributions are welcome — please open an issue or pull request. The package
follows Bioconductor coding and review standards (see [`AGENTS.md`](AGENTS.md)).

## License

MIT © the igvShiny authors (see [`LICENSE.md`](LICENSE.md) / [`DESCRIPTION`](DESCRIPTION)).
Originally created by Paul Shannon; lead developer and maintainer: Arkadiusz Gladki.

library(shiny)
library(igvShiny)
library(GenomicAlignments)

#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
print(dim(tbl.gwas))
printf <- function(...) print(noquote(sprintf(...)))
#----------------------------------------------------------------------------------------------------
tbl.bed <- data.frame(chr=c("1","1", "1"),
                      start=c(7432951, 7437000, 7438000),
                      end=  c(7436000, 7437500, 7440000),
                      value=c(-2.239, 3.0, 0.5),
                      sampleID=c("sample1", "sample2", "sample3"),
                      stringsAsFactors=FALSE)
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(

  sidebarLayout(
     sidebarPanel(
        actionButton("searchButton", "Search"),
        textInput("roi", label=""),
        h5("One simple data.frame, three igv formats:"),
        actionButton("addBedTrackButton", "Add as Bed"),
        actionButton("addBedGraphTrackButton", "Add as BedGraph"),
        actionButton("addSegTrackButton", "Add as SEG"),
        br(),
        actionButton("addGwasTrackButton", "Add GWAS Track"),
        actionButton("addBamViaHttpButton", "BAM from URL"),
        actionButton("addBamLocalFileButton", "BAM local data"),
        actionButton("addCramViaHttpButton", "CRAM from URL"),
        actionButton("removeUserTracksButton", "Remove User Tracks"),
        actionButton("getChromLoc", "Get Region"),
        htmlOutput("chromLocDisplay"),
        hr(),
        width=2
        ),
     mainPanel(
        igvShinyOutput('igvShiny_0'),
        igvShinyOutput('igvShiny_1'),
        width=10
        )
     ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {

   observeEvent(input$searchButton, {
      printf("--- search")
      searchString = isolate(input$roi)
      if(nchar(searchString) > 0)
        showGenomicRegion(session, id="igvShiny_0", searchString)
      })

   observeEvent(input$addBedTrackButton, {
      showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
      loadBedTrack(session, id="igvShiny_0", trackName="bed", tbl=tbl.bed, color="green");
      })

   observeEvent(input$addBedGraphTrackButton, {
      showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
      loadBedGraphTrack(session, id="igvShiny_0", trackName="wig", tbl=tbl.bed, color="blue", autoscale=TRUE)
      })

   observeEvent(input$addSegTrackButton, {
      showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
      loadSegTrack(session, id="igvShiny_0", trackName="seg", tbl=tbl.bed)
      })

   observeEvent(input$addGwasTrackButton, {
      printf("---- addGWASTrack")
      printf("current working directory: %s", getwd())
      showGenomicRegion(session, id="igvShiny_0", "chr19:45,248,108-45,564,645")
      loadGwasTrack(session, id="igvShiny_0", trackName="gwas", tbl=tbl.gwas, deleteTracksOfSameName=FALSE)
      })

   observeEvent(input$addBamViaHttpButton, {
      printf("---- addBamViaHttpTrack")
      showGenomicRegion(session, id="igvShiny_0", "chr5:88,733,959-88,761,606")
      base.url <- "https://1000genomes.s3.amazonaws.com/phase3/data/HG02450/alignment"
      url <- sprintf("%s/%s", base.url, "HG02450.mapped.ILLUMINA.bwa.ACB.low_coverage.20120522.bam")
      indexURL <- sprintf("%s/%s", base.url, "HG02450.mapped.ILLUMINA.bwa.ACB.low_coverage.20120522.bam.bai")
      loadBamTrackFromURL(session, id="igvShiny_0",trackName="1kg.bam", bamURL=url, indexURL=indexURL)
      })

   observeEvent(input$addBamLocalFileButton, {
      printf("---- addBamLocalFileButton")
      showGenomicRegion(session, id="igvShiny_0", "chr21:10,397,614-10,423,341")
      bamFile <- system.file(package="igvShiny", "extdata", "tumor.bam")
      x <- readGAlignments(bamFile)
      loadBamTrackFromLocalData(session, id="igvShiny_0", trackName="tumor.bam", data=x)
      })

   observeEvent(input$addCramViaHttpButton, {
      printf("---- addCramViaHttpTrack")
      showGenomicRegion(session, id="igvShiny_0", "chr5:88,733,959-88,761,606")
      base.url <- "https://s3.amazonaws.com/1000genomes/phase3/data/HG00096/exome_alignment"
      url <- sprintf("%s/%s", base.url, "HG00096.mapped.ILLUMINA.bwa.GBR.exome.20120522.bam.cram")
      indexURL <- sprintf("%s/%s", base.url, "HG00096.mapped.ILLUMINA.bwa.GBR.exome.20120522.bam.cram.crai")
      loadCramTrackFromURL(session, id="igvShiny_0",trackName="CRAM", cramURL=url, indexURL=indexURL)
      })

   observeEvent(input$removeUserTracksButton, {
      printf("---- removeUserTracks")
      removeUserAddedTracks(session, id="igvShiny_0")
      })


   observeEvent(input$trackClick, {
       printf("--- trackclick event")
       x <- input$trackClick
       print(x)
       })

   observeEvent(input[["igv-trackClick"]], {
       printf("--- igv-trackClick event")
       x <- input[["igv-trackClick"]]
       print(x)
       })

   observeEvent(input$getChromLoc, {
      printf("--- getChromLoc event")
      output$chromLocDisplay <- renderText({" "})
      getGenomicRegion(session, id="igvShiny_0")
      })

   observeEvent(input$currentGenomicRegion, {
      printf("--- currentGenomicRegion event")
      chromLocRegion <- input$currentGenomicRegion
      output$chromLocDisplay <- renderText({
         chromLocRegion
         })
       })

   genomes <- c("hg38", "hg19", "mm10", "tair10", "rhos")
   loci <- c("chr5:88,466,402-89,135,305", "MEF2C", "Mef2c", "1:7,432,931-7,440,395", "NC_007494.2:370,757-378,078")
   i <- 2

   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <- parseAndValidateGenomeSpec(genomeName=genomes[i],  initialLocus=loci[i])
     igvShiny(genomeOptions)
     })

   output$igvShiny_1 <- renderIgvShiny({
     genomeOptions <- parseAndValidateGenomeSpec(genomeName="hg38",
                                                 initialLocus="chr2:232,983,999-233,283,872")
     igvShiny(genomeOptions)
     })

} # server
#----------------------------------------------------------------------------------------------------
print(sessionInfo())
runApp(shinyApp(ui = ui, server = server), port=9832)
#shinyApp(ui = ui, server = server)

library(shiny)
library(igvShiny)
library(htmlwidgets)
# library(biomaRt)

#----------------------------------------------------------------------------------------------------

  base.url <- "https://gladki.pl/igvr/testFiles"
  title <- "ribo remote"
  fasta.file <- sprintf("%s/%s", base.url, "ribosomal-RNA-gene.fasta")
  fastaIndex.file <- sprintf("%s/%s", base.url, "ribosomal-RNA-gene.fasta.fai")
  annotation.file <- sprintf("%s/%s", base.url, "ribosomal-RNA-gene.gff3")
  locus <- "U13369.1:7,276-8,225"

  #title <- "ensembl gz"
  #fasta.file <- "http://ftp.ensembl.org/pub/release-106/fasta/homo_sapiens/dna_index/Homo_sapiens.GRCh38.dna.toplevel.fa.gz"
  #fastaIndex.file <- "http://ftp.ensembl.org/pub/release-106/fasta/homo_sapiens/dna_index/Homo_sapiens.GRCh38.dna.toplevel.fa.gz.fai"
  #locus <- "all"

genomeOptions <- parseAndValidateGenomeSpec(genomeName=title,
                                            initialLocus=locus,
                                            stockGenome=FALSE,
                                            dataMode="http",
                                            fasta=fasta.file,
                                            fastaIndex=fastaIndex.file,
                                            genomeAnnotation=annotation.file)
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(
    igvShinyOutput('igvShiny_0'),
    ))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {


   output$igvShiny_0 <- renderIgvShiny({
     cat("--- starting renderIgvShiny\n");
     x <- igvShiny(genomeOptions,
                   displayMode="SQUISHED",
                   )
     cat("--- ending renderIgvShiny\n");
     return(x)
     })

} # server
#----------------------------------------------------------------------------------------------------
if(grepl("hagfish", Sys.info()[["nodename"]]) & !interactive()){
   runApp(shinyApp(ui, server), port=6867)
   } else {
   shinyApp(ui, server)
   }

library(shiny)
library(igvShiny)
library(dplyr)
library(GenomicAlignments)

#----------------------------------------------------------------------------------------------------
# a largish gwas file hosted remotely.  this appears to have some
# zero pvalues, which when transformed by -log10, become
# infinite, not conducive to autoscaled display.
#----------------------------------------------------------------
url <- "https://s3.amazonaws.com/igv.org.demo/gwas_sample.tsv.gz"
url.gwasTrack <- GWASTrack("remote url gwas",
                           url,
                           chrom.col=3,
                           pos.col=4,
                           pval.col=10,
                           trackHeight=100,
                           autoscale=FALSE,
                           minY=0,
                           maxY=300)

#----------------------------------------------------------------------------------------------------
# a gwas file hosted locally
#----------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
tbl.gwas <- get(load(f))
tbl.gwasTrack <- GWASTrack("data.frame gwas",
                           tbl.gwas,
                           chrom.col=3,
                           pos.col=4,
                           pval.col=10,
                           trackHeight=100)
#----------------------------------------------------------------------------------------------------
printf <- function(...) print(noquote(sprintf(...)))
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(

  sidebarLayout(
     sidebarPanel(
        actionButton("loadRemoteUrlGwasTrackButton", "Load remote url GWAS Track"),
        actionButton("loadLocalTableGwasTrackButton", "Load local table GWAS Track"),
        actionButton("showApoeButton", "APOE region"),
        actionButton("showAllButton",  "All Regions"),
        width=2
        ),
     mainPanel(
        igvShinyOutput('igvShiny_0'),
        width=10
        )
     ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {


   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <- parseAndValidateGenomeSpec(genomeName="hg19",  initialLocus="all")
     igvShiny(genomeOptions)
     })

   observeEvent(input$loadRemoteUrlGwasTrackButton, {
      display(url.gwasTrack, session, id="igvShiny_0")
      })

   observeEvent(input$loadLocalTableGwasTrackButton, {
      display(tbl.gwasTrack, session, id="igvShiny_0")
      })

   observeEvent(input$showApoeButton, {
       showGenomicRegion(session, "igvShiny_0", "chr19:45,303,720-45,463,047")
       })

   observeEvent(input$showAllButton, {
       showGenomicRegion(session, "igvShiny_0", "all")
       })

   #shiny::observeEvent(input$igvReady, ignoreInit=TRUE, {
   #   tbl.gwas <- get(load("../extdata/gwas.RData"))
   #   containerID <- input$igvReady
   #   loadGwasTrack(session, id="igvShiny_0", trackName="demo", tbl=tbl.gwas, deleteTracksOfSameName=TRUE)
   #   })

} # server

#------------------------------------------------------------------------------------------------------------------------
runApp(shinyApp(ui, server), port=6869)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
    )

## ----eval=FALSE---------------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("igvShiny")

## ----include=TRUE,results="hide",message=FALSE,warning=FALSE------------------
library(igvShiny)

## ----eval=FALSE---------------------------------------------------------------
# library(igvShiny)
# options <- parseAndValidateGenomeSpec(genomeName="hg38",  initialLocus="NDUFS2")
# if (interactive()) {
#     ui <- shinyUI(fluidPage(igvShinyOutput('igvShiny'), width = 10))
#     server <-
#         function(input, output, session) {
#             output$igvShiny <- renderIgvShiny({
#                 igvShiny(options)
#             })
#             }
#     runApp(shinyApp(ui = ui, server = server))
#     }

## ----echo=TRUE----------------------------------------------------------------
igvShiny::parseAndValidateGenomeSpec("mm10")

## ----echo=TRUE----------------------------------------------------------------
igvShiny::get_css_genomes()

## ----echo=TRUE----------------------------------------------------------------
library(igvShiny)
demo_app_file <- system.file(package= "igvShiny", "demos", "stockGenomesDemo.R")
    if (interactive()) {
        shiny::runApp(demo_app_file)
        }

## ----echo=TRUE----------------------------------------------------------------
library(igvShiny)

# defining custom genome with data provided via URLs
base_url <- "https://gladki.pl/igvr/testFiles"
title <- "ribo remote"
fasta_file <- sprintf("%s/%s", base_url, "ribosomal-RNA-gene.fasta")
fastaIndex_file <- sprintf("%s/%s", base_url, "ribosomal-RNA-gene.fasta.fai")
annotation_file <- sprintf("%s/%s", base_url, "ribosomal-RNA-gene.gff3")
locus <- "U13369.1:7,276-8,225"
genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = title,
    initialLocus = locus,
    stockGenome = FALSE,
    dataMode = "http",
    fasta = fasta_file,
    fastaIndex = fastaIndex_file,
    genomeAnnotation = annotation_file
    )
genomeOptions

# defining custom genome with data provided with local files
data_directory <- system.file(package = "igvShiny", "extdata")
fasta_file      <- file.path(data_directory, "ribosomal-RNA-gene.fasta")
fastaIndex_file <- file.path(data_directory, "ribosomal-RNA-gene.fasta.fai")
annotation_file <- file.path(data_directory, "ribosomal-RNA-gene.gff3")
genomeOptions2 <- parseAndValidateGenomeSpec(
    genomeName = "ribo local",
    initialLocus = "U13369.1:7,276-8,225",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = fasta_file,
    fastaIndex = fastaIndex_file,
    genomeAnnotation = annotation_file
    )
genomeOptions2

## ----echo=TRUE----------------------------------------------------------------
library(igvShiny)
demo_app_file <- system.file(
    package = "igvShiny", 
    "demos", 
    "igvShinyDemo-customGenome-http.R"
    )
if (interactive()) {
    shiny::runApp(demo_app_file)
    }

## ----echo=TRUE----------------------------------------------------------------
library(igvShiny)
demo_app_file <-
    system.file(
        package = "igvShiny", 
        "demos", 
        "igvShinyDemo-customGenome-localFiles.R")
if (interactive()) {
    shiny::runApp(demo_app_file)
    }

## ----echo=TRUE----------------------------------------------------------------
library(igvShiny)
demo_app_file <- system.file(package= "igvShiny", "demos", "igvShinyDemo.R")
if (interactive()) {
    shiny::runApp(demo_app_file)
    }

## -----------------------------------------------------------------------------
sessionInfo()

library(shiny)
library(igvShiny)
library(htmlwidgets)

ui <-
  fluidPage(
    sidebarLayout(
      sidebarPanel(
        #actionButton("addAutoscaledGroupBedGraphTrackButton", "Update/Add bigwig"),
        actionButton("searchButton", "Search"),
        textInput("roi", label=""),
        hr(),
        div(style="background-color: white; width: 200px; height:30px; padding-left: 5px;
                   margin-top: 10px; border: 1px solid blue;",
            htmlOutput("chromLocDisplay"))
      ),
      mainPanel(
        igvShinyOutput('igvShiny_0')
      )
    ) # sidebarLayout
  )

tbl.wig <- data.frame(
  chrom=c("8","8", "8","8","20","20","20","20"),
  start=c(127735432, 127736432, 127734632,127735632,32358060,32359060,32369060,32399960),
  end=  c(127735462, 127736632, 127734632,127735932,32358260,32359160,32369260,32401060),
  score=c(-2.239, 3.0, 0.5, 5, 1,2.5, 0.2, 2),
  stringsAsFactors=FALSE
)

server <- function(input, output, session) {
  output$igvShiny_0 <- igvShiny::renderIgvShiny(
    igvShiny::igvShiny(
      list(
        genomeName="hg38",
        initialLocus="MYC",
        displayMode="SQUISHED"
        )
      )
    ) # renderIgvShiny


  ## observeEvent(input$igvReady, {
  ## observeEvent(input$addAutoscaledGroupBedGraphTrackButton, {
  newGenomicRegion <- reactive({
     printf("--- new GenomicREgion arrived")
     input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
     })

  #observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
  # newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
  #observeEvent(newGenomicRegion(), ignoreInit=TRUE, {
  observeEvent(newGenomicRegion(), ignoreInit=TRUE, {
    newLoc <- newGenomicRegion()
    #output$chromLocDisplay <- renderText({newLoc})
    printf("--- observing roi #1: %s", newLoc)
    tbl.wig1b <- tbl.wig
    tbl.wig1b$score <-tbl.wig1b$score * 10
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig1a", tbl=tbl.wig, color="blue",
                      autoscale=TRUE, autoscaleGroup=1, quiet=FALSE,
                      deleteTracksOfSameName=TRUE)
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig1b", tbl=tbl.wig1b, color="brown",
                      autoscale=TRUE, autoscaleGroup=1, quiet=FALSE,
                      deleteTracksOfSameName=TRUE)
    })

  #observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
  #  printf("--- observing roi #2")
  #  newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
  #})

  observeEvent(input$removeUserTracksButton, {
    print("---- removeUserTracks")
    removeUserAddedTracks(session, id="igvShiny_0")
  })
  observeEvent(input$searchButton, {
    print("--- search")
    searchString = isolate(input$roi)
    if(base::nchar(searchString) > 0)
      showGenomicRegion(session, id="igvShiny_0", searchString)
  })

} # server


app <- shinyApp(ui=ui, server=server)

library(shiny)
library(igvShiny)
library(cyjShiny)
library(GenomicAlignments)

#cyjshiny object

tbl.nodes <- data.frame(id=c("A", "B", "C"),
                        type=c("kinase", "TF", "glycoprotein"),
                        lfc=c(-3, 1, 1),
                        count=c(0, 0, 0),
                        stringsAsFactors=FALSE)

tbl.edges <- data.frame(source=c("A", "B", "C"),
                        target=c("B", "C", "A"),
                        interaction=c("phosphorylates", "synthetic lethal", "unknown"),
                        stringsAsFactors=FALSE)

graph.json <- toJSON(dataFramesToJSON(tbl.edges, tbl.nodes), auto_unbox=TRUE)
#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
print(dim(tbl.gwas))
printf <- function(...) print(noquote(sprintf(...)))
ns.sep <- "."
#----------------------------------------------------------------------------------------------------
igv_ui = function(id){

  ns <- NS(id)

  shinyUI(fluidPage(
      igvShinyOutput(ns('igvShiny_0')),
      cyjShiny::cyjShinyOutput(ns('cyjShiny')),
      width=10
  )
  )
}
#----------------------------------------------------------------------------------------------------
igv_server <-  function(input, output, session) {

   ns <- session$ns

   genomes <- c("hg38", "hg19", "mm10", "tair10", "rhos")
   loci <- c("chr5:88,466,402-89,135,305", "MEF2C", "Mef2c", "1:7,432,931-7,440,395", "NC_007494.2:370,757-378,078")
   i <- 2

   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <-
       parseAndValidateGenomeSpec(genomeName = genomes[i],  initialLocus = loci[i])
     igvShiny(genomeOptions, displayMode = "SQUISHED")
   })
   
   output$cyjShiny <- cyjShiny::renderCyjShiny({
     cyjShiny::cyjShiny(graph=graph.json, layoutName="cola")
   })

} # server
#----------------------------------------------------------------------------------------------------

server <- function(input, output, session){
  callModule(igv_server, "igv")
}

ui <- fluidPage(
  igv_ui(id="igv")
)

runApp(shinyApp(ui = ui, server = server), port=9834)

library(shiny)
library(igvShiny)
library(GenomicAlignments)

#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
print(dim(tbl.gwas))
printf <- function(...) print(noquote(sprintf(...)))
ns.sep <- "."
#----------------------------------------------------------------------------------------------------
tbl.bed <- data.frame(chr=c("1","1", "1"),
                      start=c(7432951, 7437000, 7438000),
                      end=  c(7436000, 7437500, 7440000),
                      value=c(-2.239, 3.0, 0.5),
                      sampleID=c("sample1", "sample2", "sample3"),
                      stringsAsFactors=FALSE)
#----------------------------------------------------------------------------------------------------
igv_ui = function(id){

  ns <- NS(id)
  printf("namespace: '%s'", ns("foo"))

  shinyUI(fluidPage(

    sidebarLayout(
      sidebarPanel(
        actionButton(ns("searchButton"), "Search"),
        textInput(ns("roi"), label=""),
        h5("One simple data.frame, three igv formats:"),
        actionButton(ns("addBedTrackButton"), "Add as Bed"),
        actionButton(ns("addBedGraphTrackButton"), "Add as BedGraph"),
        actionButton(ns("addSegTrackButton"), "Add as SEG"),
        br(),
        actionButton(ns("addGwasTrackButton"), "Add GWAS Track"),
        actionButton(ns("addBamViaHttpButton"), "BAM from URL"),
        actionButton(ns("addBamLocalFileButton"), "BAM local data"),
        actionButton(ns("addCramViaHttpButton"), "CRAM from URL"),
        actionButton(ns("removeUserTracksButton"), "Remove User Tracks"),
        actionButton(ns("getChromLocButton"), "Get Region"),
        actionButton(ns("clearChromLocButton"), "Clear Region"),
        div(style="background-color: white; width: 200px; height:30px; padding-left: 5px;
                   margin-top: 10px; border: 1px solid blue;",
            htmlOutput(ns("chromLocDisplay"))),
        hr(),
        width=2
      ),
      mainPanel(
        igvShinyOutput(ns('igvShiny_0')),
        # igvShinyOutput('igvShiny_1'),
        width=10
      )
    ) # sidebarLayout
  ))
}
#----------------------------------------------------------------------------------------------------
igv_server <-  function(input, output, session) {

   ns <- session$ns

   observeEvent(input$searchButton, {
      printf("--- search")
      searchString = isolate(input$roi)
      if(nchar(searchString) > 0)
        showGenomicRegion(session, id="igvShiny_0", searchString)
      })

   observeEvent(input$addBedTrackButton, {
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr1:7,426,231-7,453,241")
      loadBedTrack(session, id=ns("igvShiny_0"), trackName="bed", tbl=tbl.bed, color="green");
      })

   observeEvent(input$addBedGraphTrackButton, {
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr1:7,426,231-7,453,241")
      loadBedGraphTrack(session, id=ns("igvShiny_0"), trackName="wig", tbl=tbl.bed, color="blue", autoscale=TRUE)
      })

   observeEvent(input$addSegTrackButton, {
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr1:7,426,231-7,453,241")
      loadSegTrack(session, id=ns("igvShiny_0"), trackName="seg", tbl=tbl.bed)
      })

   observeEvent(input$addGwasTrackButton, {
      printf("---- addGWASTrack")
      printf("current working directory: %s", getwd())
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr19:45,248,108-45,564,645")
      loadGwasTrack(session, id=ns("igvShiny_0"), trackName="gwas", tbl=tbl.gwas, deleteTracksOfSameName=FALSE)
      })

   observeEvent(input$addBamViaHttpButton, {
      printf("---- addBamViaHttpTrack")
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr5:88,733,959-88,761,606")
      base.url <- "https://1000genomes.s3.amazonaws.com/phase3/data/HG02450/alignment"
      url <- sprintf("%s/%s", base.url, "HG02450.mapped.ILLUMINA.bwa.ACB.low_coverage.20120522.bam")
      indexURL <- sprintf("%s/%s", base.url, "HG02450.mapped.ILLUMINA.bwa.ACB.low_coverage.20120522.bam.bai")
      loadBamTrackFromURL(session, id=ns("igvShiny_0"),trackName="1kg.bam", bamURL=url, indexURL=indexURL)
      })

   observeEvent(input$addBamLocalFileButton, {
      printf("---- addBamLocalFileButton")
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr21:10,397,614-10,423,341")
      bamFile <- system.file(package="igvShiny", "extdata", "tumor.bam")
      x <- readGAlignments(bamFile)
      loadBamTrackFromLocalData(session, id=ns("igvShiny_0"), trackName="tumor.bam", data=x)
      })

   observeEvent(input$addCramViaHttpButton, {
      printf("---- addCramViaHttpTrack")
      showGenomicRegion(session, id=ns("igvShiny_0"), "chr5:88,733,959-88,761,606")
      base.url <- "https://s3.amazonaws.com/1000genomes"
      url <- sprintf("%s/%s", base.url, "1000G_2504_high_coverage/additional_698_related/data/ERR3989250/HG04160.final.cram")
      indexURL <- sprintf("%s.%s", url, "crai")
      loadCramTrackFromURL(session, id=ns("igvShiny_0"),trackName="CRAM", cramURL=url, indexURL=indexURL)
      
      })

   observeEvent(input$removeUserTracksButton, {
      printf("---- removeUserTracks")
      removeUserAddedTracks(session, id=ns("igvShiny_0"))
      })


   observeEvent(input$trackClick, {
       printf("--- trackclick event")
       x <- input$trackClick
       print(x)
   })

   observeEvent(input$getChromLocButton, {
      # printf("--- getChromLoc event")
      # sends message to igv.js in browser; currentGenomicRegion.<id> event sent back
      # see below for how that can be captured and displayed
      getGenomicRegion(session, id=ns("igvShiny_0"))
      print(sprintf("getChromLocButton, currentGenomicRegion.%s", ns("igvShiny_0")))
      })

   observeEvent(input$clearChromLocButton, {
       printf("clearing chromLocDisplay after clearChromLocButton click")
       printf("---- names(input)")
       print(names(input))
      output$chromLocDisplay <- renderText({" "})
      })

   observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
      newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
      #observeEvent(input$genomicRegionChanged, {
      #newLoc <- input$genomicRegionChanged
      printf("new chromLocString: %s", newLoc)
      output$chromLocDisplay <- renderText({newLoc})
      })

   genomes <- c("hg38", "hg19", "mm10", "tair10", "rhos")
   loci <- c("chr5:88,466,402-89,135,305", "MEF2C", "Mef2c", "1:7,432,931-7,440,395", "NC_007494.2:370,757-378,078")
   i <- 2

   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <- parseAndValidateGenomeSpec(genomeName=genomes[i], initialLocus=loci[i])
     igvShiny(genomeOptions)
     })

} # server
#----------------------------------------------------------------------------------------------------
print(sessionInfo())

server <- function(input, output, session){
  callModule(igv_server, "igv")
}

ui <- fluidPage(
  igv_ui(id="igv")
)

runApp(shinyApp(ui = ui, server = server), port=9834)

library(shiny)
library(igvShiny)
library(htmlwidgets)
# library(biomaRt)

#----------------------------------------------------------------------------------------------------
data.directory <- system.file(package="igvShiny", "extdata", "sarsGenome")
fasta.file <- file.path(data.directory, "Sars_cov_2.ASM985889v3.dna.toplevel.fa")
fastaIndex.file <- file.path(data.directory, "Sars_cov_2.ASM985889v3.dna.toplevel.fa.fai")
annotation.file <- file.path(data.directory, "Sars_cov_2.ASM985889v3.101.gff3")

title <- "SARS-CoV-2"
genomeOptions <- parseAndValidateGenomeSpec(genomeName=title,
                                            initialLocus= NA,
                                            stockGenome=FALSE,
                                            dataMode="localFiles",
                                            fasta=fasta.file,
                                            fastaIndex=fastaIndex.file,
                                            genomeAnnotation=annotation.file
                                            )
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(
    igvShinyOutput('igvShiny_0'),
    ))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {

   output$igvShiny_0 <- renderIgvShiny({
     cat("--- starting renderIgvShiny\n");
     x <- igvShiny(genomeOptions, displayMode="SQUISHED")
     return(x)
     })

} # server
#----------------------------------------------------------------------------------------------------
if(grepl("hagfish", Sys.info()[["nodename"]]) & !interactive()){
   runApp(shinyApp(ui, server), port=6867)
   } else {
   shinyApp(ui, server)
   }

# igvShiny — alignments and variants that are already in R: the *FromLocalData
# loaders write a file into the tracks directory and have shiny serve it, so
# unlike the URL loaders they need the parsing package on the server side.
# Loaded on click, not at startup: a missing suggested package should grey out
# one button rather than the whole app. Junctions are in junctions.R, GFF3 in
# gff3.R.

library(shiny)
library(bslib)
library(igvShiny)

needs <- function(pkg) {
  ok <- requireNamespace(pkg, quietly = TRUE)
  if (!ok) showNotification(sprintf("install %s to use this button", pkg),
                            type = "error")
  ok
}

ui <- page_sidebar(
  title = "igvShiny — tracks from local data",
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  fillable = TRUE,
  sidebar = sidebar(
    width = 280,
    actionButton("addLocalBamButton", "BAM (readGAlignments)", class = "w-100 mb-2"),
    actionButton("addLocalVcfButton", "VCF (readVcf)", class = "w-100 mb-2"),
    actionButton("removeUserTracksButton", "Remove user tracks",
                 class = "btn-outline-danger w-100")
  ),
  card(full_screen = TRUE,
       card_body(class = "p-0", igvShinyOutput("igvShiny_0", height = "100%")))
)

server <- function(input, output, session) {
  observeEvent(input$addLocalBamButton, {
    if (!needs("GenomicAlignments")) return()
    showGenomicRegion(session, id = "igvShiny_0", "chr21:10,397,614-10,423,341")
    bam <- system.file(package = "igvShiny", "extdata", "tumor.bam")
    loadBamTrackFromLocalData(session, id = "igvShiny_0", trackName = "tumor.bam",
                              data = GenomicAlignments::readGAlignments(bam))
  })

  observeEvent(input$addLocalVcfButton, {
    if (!needs("VariantAnnotation")) return()
    showGenomicRegion(session, id = "igvShiny_0", "chr19:33,298,144-33,301,107")
    f <- system.file(package = "igvShiny", "extdata", "chr19-cebpaRegion.vcf.gz")
    loadVcfTrack(session, id = "igvShiny_0", trackName = "vcf",
                 vcfData = VariantAnnotation::readVcf(f, "hg38"))
  })

  observeEvent(input$removeUserTracksButton,
               removeUserAddedTracks(session, id = "igvShiny_0"))

  output$igvShiny_0 <- renderIgvShiny({
    igvShiny(parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "NDUFS2"))
  })
}

shinyApp(ui = ui, server = server)

# igvShiny — the three ways to specify a genome: a stock name igv.js streams
# itself, your own fasta + index + annotation from disk, or the same three
# files fetched from a URL. Only the first needs the igv.js registry, so an
# outage there leaves the stock genomes blank and the custom ones working.

library(shiny)
library(bslib)
library(igvShiny)

data.dir <- system.file(package = "igvShiny", "extdata")
base.url <- "https://gladki.pl/igvr/testFiles"
genomeSpec <- function(mode, stockName = "danRer11") {
  switch(
    mode,
    stock = parseAndValidateGenomeSpec(genomeName = stockName, initialLocus = "all"),
    localFiles = parseAndValidateGenomeSpec(
      genomeName = "ribosomal RNA gene (local)",
      initialLocus = "U13369.1:7,276-8,225",
      stockGenome = FALSE, dataMode = "localFiles",
      fasta = file.path(data.dir, "ribosomal-RNA-gene.fasta"),
      fastaIndex = file.path(data.dir, "ribosomal-RNA-gene.fasta.fai"),
      genomeAnnotation = file.path(data.dir, "ribosomal-RNA-gene.gff3")),
    http = parseAndValidateGenomeSpec(
      genomeName = "ribosomal RNA gene (url)",
      initialLocus = "U13369.1:7,276-8,225",
      stockGenome = FALSE, dataMode = "http",
      fasta = sprintf("%s/ribosomal-RNA-gene.fasta", base.url),
      fastaIndex = sprintf("%s/ribosomal-RNA-gene.fasta.fai", base.url),
      genomeAnnotation = sprintf("%s/ribosomal-RNA-gene.gff3", base.url))
  )
}

ui <- page_sidebar(
  title = "igvShiny — stock and custom genomes",
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  fillable = TRUE,
  sidebar = sidebar(
    width = 300,
    radioButtons("mode", "Genome source",
                 c("stock (igv.js registry)" = "stock",
                   "custom, local files" = "localFiles",
                   "custom, files by URL" = "http")),
    conditionalPanel("input.mode == 'stock'",
                     selectInput("genomeChooser", "Stock genome",
                                 sort(get_css_genomes()), selected = "danRer11"))
  ),
  card(full_screen = TRUE,
       card_body(class = "p-0", igvShinyOutput("igvShiny_0", height = "100%")))
)

server <- function(input, output, session) {
  # re-rendering rebuilds the browser: a live igv.js instance cannot swap genome
  output$igvShiny_0 <- renderIgvShiny({
    req(input$mode, input$genomeChooser)
    igvShiny(genomeSpec(input$mode, input$genomeChooser))
  })
}

shinyApp(ui = ui, server = server)

# igvShiny — flagship demo: most of the API in one app
#
# This is also what the public Posit Connect deploy serves
# (demo/posit-connect/app.R runs this very file) and what
# tests/testthat/test-shinyApp.R drives, so the input ids are part of the
# contract: renaming one breaks a test and the live demo at the same time.
#
# Deliberately no *FromLocalData loader here: those pull Rsamtools /
# GenomicAlignments / VariantAnnotation, heavy C-compiled Bioconductor deps
# that would have to be installed on the cloud host just to serve a demo
# button. They live in local-data.R instead. Everything below is either built
# from a data.frame or fetched client-side by igv.js from a URL.

library(shiny)
library(bslib)
library(igvShiny)
library(htmlwidgets)

#----------------------------------------------------------------------------
# sample data
#----------------------------------------------------------------------------
f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))

tbl.bed5 <- data.frame(chr = c("1", "1", "1"),
                       start = c(7432951, 7437000, 7438000),
                       end   = c(7436000, 7437500, 7440000),
                       value = c(-2.239, 3.0, 0.5),
                       sampleID = c("sample1", "sample2", "sample3"),
                       stringsAsFactors = FALSE)

base.loc <- 161200000
tbl.bed9 <- data.frame(chrom = rep("chr1", 3),
                       start = c(base.loc, base.loc + 100, base.loc + 250),
                       end = c(base.loc + 50, base.loc + 120, base.loc + 290),
                       name = c("red", "green", "blue"),
                       score = c(0.35, 0.62, 0.88),
                       strand = c("+", "-", "+"),
                       thickStart = c(base.loc + 10, base.loc + 110, base.loc + 260),
                       thickEnd = c(base.loc + 20, base.loc + 130, base.loc + 280),
                       itemRgb = c("255,0,128", "0,255,0", "0,0,255"),
                       stringsAsFactors = FALSE)

#----------------------------------------------------------------------------
# ui
#----------------------------------------------------------------------------
# a full-width action button with a leading icon, styled consistently
demoButton <- function(id, label, icon_name, class = "btn-outline-primary") {
  actionButton(id, label, icon = icon(icon_name),
               class = paste("w-100 mb-2 text-start", class))
}

theme <- bs_theme(
  version = 5,
  primary = "#2c6faa",
  base_font = font_google("Inter", local = FALSE),
  heading_font = font_google("Inter", local = FALSE)
)

ui <- page_sidebar(
  title = "igvShiny — interactive genome browser",
  theme = theme,
  fillable = TRUE,

  sidebar = sidebar(
    width = 300,
    title = "Controls",

    # keep the search box always visible above the accordion
    div(
      class = "mb-3",
      textInput("roi", label = "Search locus / gene",
                placeholder = "e.g. MEF2C or chr1:7,426,231-7,453,241"),
      actionButton("searchButton", "Search", icon = icon("magnifying-glass"),
                   class = "btn-primary w-100")
    ),

    accordion(
      open = c("Sample-data tracks", "Tracks from URL"),

      accordion_panel(
        "Sample-data tracks", icon = icon("table"),
        demoButton("addBedTrackButton", "BED", "align-left"),
        demoButton("addBedGraphTrackButton", "BedGraph", "chart-area"),
        demoButton("addBedGraphWithAltColorTrackButton", "BedGraph (AltColor)", "palette"),
        demoButton("addBed9TrackButton", "bed9", "grip-lines"),
        demoButton("addGwasTrackButton", "GWAS", "chart-column"),
        demoButton("addGwasCustomTrackButton", "GWAS (columns + colors)", "palette")
      ),

      accordion_panel(
        "Tracks from URL", icon = icon("cloud-arrow-down"),
        demoButton("addBedGraphTrackFromURLButton", "BedGraph (URL)", "chart-area", "btn-outline-secondary"),
        demoButton("addBamViaHttpButton", "BAM (URL)", "dna", "btn-outline-secondary"),
        demoButton("addCramViaHttpButton", "CRAM (URL)", "dna", "btn-outline-secondary"),
        demoButton("addSpliceJunctionTrackButton", "Splice junctions (URL)", "bezier-curve", "btn-outline-secondary")
      ),

      accordion_panel(
        "Region tools", icon = icon("location-crosshairs"),
        demoButton("getChromLocButton", "Get region", "crosshairs", "btn-outline-dark"),
        demoButton("clearChromLocButton", "Clear region", "eraser", "btn-outline-dark"),
        div(class = "small text-muted mt-1 mb-1", "Current region:"),
        div(class = "border rounded p-2 small bg-body-tertiary font-monospace",
            textOutput("chromLocDisplay"))
      ),

      accordion_panel(
        "Remove tracks", icon = icon("trash-can"),
        textInput("trackNameToRemove", label = NULL, value = "bed5",
                  placeholder = "track name"),
        demoButton("removeTrackByNameButton", "Remove by name", "eraser", "btn-outline-danger"),
        demoButton("removeUserTracksButton", "Remove all user tracks", "trash-can", "btn-outline-danger")
      )
    ),

    # footer: quick links back to the project
    tags$div(
      class = "mt-auto pt-2 small",
      tags$a(href = "https://github.com/gladkia/igvShiny", target = "_blank",
             class = "link-secondary text-decoration-none me-3",
             icon("github"), " GitHub"),
      tags$a(href = "https://gladkia.github.io/igvShiny/", target = "_blank",
             class = "link-secondary text-decoration-none",
             icon("book"), " Docs"),
      # the Connect deploy pins igvShiny to one commit, so this app can run
      # against an older package than the source tree - print which one
      tags$div(class = "text-muted mt-1 font-monospace",
               sprintf("igvShiny %s", utils::packageVersion("igvShiny")))
    )
  ),

  card(
    full_screen = TRUE,
    card_header(
      class = "d-flex align-items-center gap-2",
      icon("dna"), "Genome viewer",
      tags$span(class = "badge text-bg-light ms-auto", "hg38")
    ),
    card_body(
      class = "p-0",
      igvShinyOutput("igvShiny_0", height = "100%")
    )
  )
)

#----------------------------------------------------------------------------
# server
#----------------------------------------------------------------------------
server <- function(input, output, session) {

  observeEvent(input$searchButton, {
    searchString <- isolate(input$roi)
    if (nchar(searchString) > 0)
      showGenomicRegion(session, id = "igvShiny_0", searchString)
  })

  observeEvent(input$addBedTrackButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:7,426,231-7,453,241")
    loadBedTrack(session, id = "igvShiny_0", trackName = "bed5", tbl = tbl.bed5)
  })

  observeEvent(input$addBed9TrackButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:161,199,757-161,201,277")
    loadBedTrack(session, id = "igvShiny_0", trackName = "bed9", tbl = tbl.bed9)
  })

  observeEvent(input$addBedGraphTrackButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:7,426,231-7,453,241")
    loadBedGraphTrack(session, id = "igvShiny_0", trackName = "wig/bedGraph/local",
                      tbl = tbl.bed5, color = "blue", autoscale = TRUE)
  })

  observeEvent(input$addBedGraphWithAltColorTrackButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:7,426,231-7,453,241")
    loadBedGraphTrack(session, id = "igvShiny_0", trackName = "wig/bedGraph/local/alt",
                      tbl = tbl.bed5, color = "blue",
                      trackConfig = list(altColor = "red"), autoscale = TRUE)
  })

  observeEvent(input$addBedGraphTrackFromURLButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:154,946,914-155,080,475")
    url <- "https://www.encodeproject.org/files/ENCFF356YES/@@download/ENCFF356YES.bigWig"
    loadBedGraphTrackFromURL(session, id = "igvShiny_0", trackName = "bedGraph/remote",
                             url = url, color = "brown",
                             trackHeight = 50, autoscale = TRUE)
  })

  observeEvent(input$addGwasTrackButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr19:45,248,108-45,564,645")
    loadGwasTrack(session, id = "igvShiny_0", trackName = "gwas", tbl.gwas = tbl.gwas,
                  deleteTracksOfSameName = FALSE)
  })

  observeEvent(input$addGwasCustomTrackButton, {
    # headers igv.js has no chance of guessing: the same table draws an empty
    # track unless the column numbers are spelled out
    tbl <- tbl.gwas
    names(tbl)[c(3, 4, 10)] <- c("chromosome_label", "bp_location", "significance")
    showGenomicRegion(session, id = "igvShiny_0", "chr19:45,248,108-45,564,645")
    track <- GWASTrack("gwas, custom columns", tbl,
                       chrom.col = 3, pos.col = 4, pval.col = 10,
                       trackHeight = 200,
                       chromosomeColorMap = list("19" = "purple", "*" = "gray"))
    display(track, session, id = "igvShiny_0", deleteTracksOfSameName = FALSE)
  })

  observeEvent(input$addBamViaHttpButton, {
    # the browser runs on hg38, so the reads have to be aligned to hg38 too.
    # This used to point at a 1000 Genomes phase 3 bam, which is GRCh37: the
    # track loaded, and every read came up a wall of mismatches, because it was
    # being drawn against the wrong reference (spotted on the 2026-08 LinkedIn
    # post). NA19240 below is the igv.js structural-variant tutorial sample,
    # aligned to hg38
    showGenomicRegion(session, id = "igvShiny_0", "chr21:26,000,000-26,004,000")
    url <- paste0("https://raw.githubusercontent.com/igvteam/igv-data/",
                  "main/data/tutorials/svs/NA19240.bam")
    loadBamTrackFromURL(session, id = "igvShiny_0", trackName = "NA19240.bam",
                        bamURL = url, indexURL = sprintf("%s.bai", url))
  })

  observeEvent(input$addCramViaHttpButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr5:88,733,959-88,761,606")
    url <- paste0("https://s3.amazonaws.com/1000genomes/1000G_2504_high_coverage/",
                  "additional_698_related/data/ERR3989250/HG04160.final.cram")
    loadCramTrackFromURL(session, id = "igvShiny_0", trackName = "CRAM",
                         cramURL = url, indexURL = sprintf("%s.crai", url))
  })

  observeEvent(input$addSpliceJunctionTrackButton, {
    # junctions read as arcs only against the coverage they were spliced out
    # of, so load the sample's bigWig underneath: that pairing is what a
    # sashimi plot shows, and igv.js has no sashimi of its own
    showGenomicRegion(session, id = "igvShiny_0", "chr15:92,835,700-93,031,800")
    base.url <- paste0("https://raw.githubusercontent.com/igvteam/igv-data/",
                       "main/data/test/splice_junctions")
    stem <- "splice_junction_track_test_cases_sampleA.chr15-92835700-93031800"
    loadBedGraphTrackFromURL(session, id = "igvShiny_0", trackName = "sampleA coverage",
                             url = sprintf("%s/%s.bigWig", base.url, stem),
                             color = "gray", trackHeight = 60, autoscale = TRUE)
    loadSpliceJunctionTrackFromURL(
      session, id = "igvShiny_0", trackName = "sampleA junctions",
      url = sprintf("%s/%s.SJ.out.bed.gz", base.url, stem),
      indexURL = sprintf("%s/%s.SJ.out.bed.gz.tbi", base.url, stem),
      trackConfig = list(colorBy = "motif", labelWith = "uniquelyMapped"))
  })

  observeEvent(input$removeTrackByNameButton, {
    name <- isolate(input$trackNameToRemove)
    if (nchar(name) > 0)
      removeTracksByName(session, id = "igvShiny_0", trackNames = name)
  })

  observeEvent(input$removeUserTracksButton, {
    removeUserAddedTracks(session, id = "igvShiny_0")
  })

  observeEvent(input[["igv-trackClick"]], {
    x <- input[["igv-trackClick"]]
    tbl <- data.frame(name = as.character(x)[grep("name", names(x))],
                      value = as.character(x)[grep("value", names(x))],
                      stringsAsFactors = FALSE)
    showModal(modalDialog(HTML(renderTable(tbl)()), easyClose = TRUE))
  })

  observeEvent(input$getChromLocButton, {
    getGenomicRegion(session, id = "igvShiny_0")
  })

  # one output, fed by a reactiveVal: assigning output$… inside an observer
  # re-registers the output on every event
  chromLoc <- reactiveVal(" ")
  output$chromLocDisplay <- renderText(chromLoc())

  observeEvent(input$clearChromLocButton, chromLoc(" "))

  observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
    chromLoc(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]])
  })

  output$igvShiny_0 <- renderIgvShiny({
    genomeOptions <- parseAndValidateGenomeSpec(
      genomeName = "hg38", initialLocus = "chr1:6,575,383-8,304,088")
    igvShiny(genomeOptions, displayMode = "SQUISHED", tracks = list())
  })

} # server

shinyApp(ui = ui, server = server)

.onLoad <- function(libname, pkgname) {
  # assure proper value for "tracks" resource path; .tracksDir() re-points it
  # whenever the tracks directory moves, so every write site goes through it too
  .tracksDir()
}

library(shiny)
library(igvShiny)
library(htmlwidgets)

shinyUI(
  fluidPage(
    sidebarLayout(
      sidebarPanel(
        #actionButton("addAutoscaledGroupBedGraphTrackButton", "Update/Add bigwig"),
        actionButton("searchButton", "Search"),
        textInput("roi", label=""),
        hr(),
        div(style="background-color: white; width: 200px; height:30px; padding-left: 5px;
                   margin-top: 10px; border: 1px solid blue;",
            htmlOutput("chromLocDisplay"))
      ),
      mainPanel(
        igvShinyOutput('igvShiny_0')
      )
    ) # sidebarLayout
  )
)

#' @name GWASTrack-class
#' @rdname GWASTrack-class
#' @importFrom methods is new
#' @importFrom utils write.table
#' @keywords GWASTrack_class

#' @exportClass GWASTrack

.GWASTrack <-
  setClass(
    "GWASTrack",
    slots = representation(
      trackName = "character",
      data.mode = "character",
      url = "character",
      chrom.col = "numeric",
      pos.col = "numeric",
      pval.col = "numeric",
      chromosomeColorMap = "list",
      trackHeight = "numeric",
      autoscale = "logical",
      minY = "numeric",
      maxY = "numeric"
    )
  )

setGeneric("display",
           signature = "obj",
           function(obj,
                    session,
                    id,
                    deleteTracksOfSameName = TRUE) {
             standardGeneric("display")
           })

setGeneric("getUrl",
           signature = "obj",
           function(obj) {
             standardGeneric("getUrl")
           })

#-------------------------------------------------------------------------------
#' Check a chromosome color map before it reaches igv.js
#'
#' @param map a named list or character vector of colors, possibly empty
#' @return the map as a named list, empty if nothing was supplied
#' @keywords GWASTrack_class
.sanitizeChromosomeColorMap <- function(map) {
  if (is.null(map) || length(map) == 0) {
    return(list())
  }

  if (!is.list(map) && !is.character(map)) {
    stop("error: chromosomeColorMap must be a named list or character vector")
  }

  map <- as.list(map)
  # anyNA first: with an NA name, any(names == "") is NA and `if` errors
  if (is.null(names(map)) || anyNA(names(map)) || any(names(map) == "")) {
    stop("error: every chromosomeColorMap color needs a chromosome name")
  }

  # a colorTable is a JSON object, so a repeated chromosome keeps whichever
  # color happens to be written last - better to say so than to pick one
  if (anyDuplicated(names(map))) {
    stop(sprintf("error: chromosomeColorMap names one chromosome twice: %s",
                 toString(unique(names(map)[duplicated(names(map))]))))
  }

  valid <- vapply(map, function(color) {
    is.character(color) && length(color) == 1L && !is.na(color)
  }, logical(1))
  if (!all(valid)) {
    stop(sprintf("error: chromosomeColorMap colors must be single strings: %s",
                 toString(names(map)[!valid])))
  }

  map
}
#-------------------------------------------------------------------------------
#' Constructor for GWASTrack
#'
#' \code{GWASTrack} creates an \code{IGV} manhattan track from GWAS data
#'
#' @name GWASTrack
#' @rdname GWASTrack-class
#'
#' @param trackName A character string, used as track label by igv,
#' we recommend unique names per track.
#' @param data a data.frame or a url pointing to one,
#' whose structure is described by chrom.col, pos.col, pval.col
#' @param chrom.col numeric, the column number of the chromosome column
#' @param pos.col numeric, the column number of the position column
#' @param pval.col numeric, the column number of the GWAS pvalue column
#' @param chromosomeColorMap a named list or character vector mapping
#' chromosome names, as they are spelled in the data, to colors,
#' e.g. \code{list("1" = "red", "2" = "blue")}. A \code{"*"} entry colors every
#' chromosome not named explicitly. Empty by default, which leaves the igv.js
#' chromosome palette in place
#' @param trackHeight numeric in pixels
#' @param autoscale logical
#' @param minY numeric for explicit (non-auto) scaling
#' @param maxY numeric for explicit (non-auto) scaling
#'
#' @return A GWASTrack object
#'
#' @examples
#' file <-
#'   # a local gwas file
#'   system.file(package = "igvShiny", "extdata", "gwas-5k.tsv.gz")
#' tbl.gwas <- read.table(file,
#'                        sep = "\t",
#'                        header = TRUE,
#'                        quote = "")
#' dim(tbl.gwas)
#' track <-
#'   GWASTrack(
#'     "gwas 5k",
#'     tbl.gwas,
#'     chrom.col = 12,
#'     pos.col = 13,
#'     pval.col = 28
#'   )
#' getUrl(track)
#'
#' # a remote gwas file: the constructor checks that the url resolves, so this
#' # block reaches the network and stays out of R CMD check
#' \donttest{
#' url <- "https://gladki.pl/igvShiny/gwas_sample.tsv.gz"
#' track <- GWASTrack(
#'   "remote url gwas",
#'   url,
#'   chrom.col = 3,
#'   pos.col = 4,
#'   pval.col = 10,
#'   autoscale = FALSE,
#'   minY = 0,
#'   maxY = 300,
#'   trackHeight = 100
#' )
#' getUrl(track)
#' }
#'
#' # colors picked per chromosome, with "*" covering the rest
#' track <- GWASTrack(
#'   "gwas 5k, custom colors",
#'   tbl.gwas,
#'   chrom.col = 12,
#'   pos.col = 13,
#'   pval.col = 28,
#'   chromosomeColorMap = list("1" = "red", "2" = "blue", "*" = "gray")
#' )
#'
#'
#' @keywords GWASTrack_class
#' @export
#'


GWASTrack <- function(trackName,
                      data,
                      chrom.col,
                      pos.col,
                      pval.col,
                      chromosomeColorMap = list(),
                      trackHeight = 50,
                      autoscale = TRUE,
                      minY = 0,
                      maxY = 30) {
  data.class <- class(data)
  stopifnot(data.class %in% c("data.frame", "character"))

  chromosomeColorMap <- .sanitizeChromosomeColorMap(chromosomeColorMap)
  columns <- list(chrom.col = chrom.col, pos.col = pos.col,
                  pval.col = pval.col)
  for (name in names(columns)) {
    value <- columns[[name]]
    if (!is.numeric(value) || length(value) != 1L || is.na(value) ||
        value != round(value) || value < 1) {
      stop(sprintf("error: %s must be a single column number, at least 1",
                   name))
    }
  }

  if (data.class == "data.frame") {
    # a column number past the end of the table silently yields an empty track,
    # so it is worth catching here, where the table is still in hand
    too.big <- vapply(columns, function(value) value > NCOL(data), logical(1))
    if (any(too.big)) {
      stop(sprintf(
        "error: %s beyond the %d columns of the gwas table",
        toString(names(columns)[too.big]), NCOL(data)
      ))
    }

    mode <- "local.url"
    tdir <- .tracksDir()
    x <- NULL
    url <-
      tempfile(tmpdir = tdir, fileext = ".gwas") # expanded in javascript
    write.table(
      data,
      sep = "\t",
      row.names = FALSE,
      quote = FALSE,
      file = url
    )
  }

  if (data.class == "character") {
    if (httr::http_error(data)) {
      # was a legitimate url provided?
      error.message <-
        sprintf("error: putative gwas file url unreachable: '%s'",
                data)
      stop(error.message)
    }
    mode <- "remote.url"
    url <- data
  }

  obj <- .GWASTrack(
    trackName = trackName,
    data.mode = mode,
    url = url,
    chrom.col = chrom.col,
    pos.col = pos.col,
    pval.col = pval.col,
    chromosomeColorMap = chromosomeColorMap,
    trackHeight = trackHeight,
    autoscale = autoscale,
    minY = minY,
    maxY = maxY
  )
  obj

} # GWASTrack
#-------------------------------------------------------------------------------
#' display the already constructed and configured track
#'
#' @rdname display
#' @aliases display
#'
#' @param obj An object of class GWASTrack
#' @param session a Shiny session object
#' @param id character the identifier of the target igv object in the browser
#' @param deleteTracksOfSameName logical to avoid duplications in track names
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return nothing
#'
#' @keywords GWASTrack_class
#' @export
#'
setMethod("display",
          "GWASTrack",

          function(obj,
                   session,
                   id,
                   deleteTracksOfSameName = TRUE) {
            if (deleteTracksOfSameName) {
              removeTracksByName(session, id, obj@trackName)
            }

            state[["userAddedTracks"]] <-
              unique(c(state[["userAddedTracks"]], obj@trackName))

            # javascript function consults dataMode,
            # modifies dataUrl if local.url,
            # prepending the http host of the modest RStudio/Shiny webserver
            # make sure the embedded shiny webserver can find it by:
            #   - adding a resource path with a convenient shorthand name,
            #   pointing to the typically long and cryptic actual local host
            #   temporary directory
            #   - adjusting message$dataUrl to use that shorthand directory name

            message <- list(
              elementID = id,
              trackName = obj@trackName,
              dataMode = obj@data.mode,
              dataUrl = obj@url,
              # without an explicit mapping igv.js guesses the columns from the
              # header names, which only works for tables spelled its way
              columns = list(
                chromosome = obj@chrom.col,
                position = obj@pos.col,
                value = obj@pval.col
              ),
              trackHeight = obj@trackHeight,
              autoscale = obj@autoscale,
              min = obj@minY,
              max = obj@maxY
            )

            if (length(obj@chromosomeColorMap) > 0) {
              message$colorTable <- obj@chromosomeColorMap
            }

            if (obj@data.mode == "local.url") {
              directory.name <- dirname(obj@url)
              file.name      <-  basename(obj@url)
              message$dataUrl <- file.path("tracks", file.name)
            }

            session$sendCustomMessage("loadGwasTrackFlexibleSource", message)

          }) # display
#-------------------------------------------------------------------------------
#' the url of the gwas table
#'
#' @rdname getUrl
#' @aliases getUrl
#'
#' @param obj An object of class GWASTrack
#'
#' @examples
#' file <-
#'   # a local gwas file
#'   system.file(package = "igvShiny", "extdata", "gwas-5k.tsv.gz")
#' tbl.gwas <- read.table(file,
#'                        sep = "\t",
#'                        header = TRUE,
#'                        quote = "")
#' track <-
#'   GWASTrack(
#'     "gwas 5k",
#'     tbl.gwas,
#'     chrom.col = 12,
#'     pos.col = 13,
#'     pval.col = 28
#'   )
#' getUrl(track)
#'
#' @return character
#'
#' @keywords GWASTrack_class
#' @export
#'
setMethod("getUrl",
          "GWASTrack",
          function(obj) {
            obj@url
          })
#-------------------------------------------------------------------------------
#' Display a short summary of a GWASTrack object
#'
#' @rdname show
#' @aliases show,GWASTrack-method
#'
#' @importFrom methods show
#'
#' @param object An object of class GWASTrack
#'
#' @examples
#' file <-
#'   system.file(package = "igvShiny", "extdata", "gwas-5k.tsv.gz")
#' tbl.gwas <- read.table(file, sep = "\t", header = TRUE, quote = "")
#' track <-
#'   GWASTrack("gwas 5k", tbl.gwas, chrom.col = 12, pos.col = 13, pval.col = 28)
#' track
#'
#' @return nothing; called for the side effect of printing a summary
#'
#' @keywords GWASTrack_class
#' @export
#'
setMethod("show",
          "GWASTrack",
          function(object) {
            cat("GWASTrack object\n")
            cat("  trackName:  ", object@trackName, "\n", sep = "")
            cat("  dataMode:   ", object@data.mode, "\n", sep = "")
            cat("  url:        ", object@url, "\n", sep = "")
            cat("  columns:    chrom=", object@chrom.col,
                " pos=", object@pos.col,
                " pval=", object@pval.col, "\n", sep = "")
            if (length(object@chromosomeColorMap) > 0) {
              cat("  colors:     ", length(object@chromosomeColorMap),
                  " chromosomes\n", sep = "")
            }
            cat("  trackHeight:", object@trackHeight, "\n")
            cat("  autoscale:  ", object@autoscale, "\n", sep = "")
            if (isFALSE(object@autoscale)) {
              cat("  yRange:     [", object@minY, ", ", object@maxY, "]\n",
                  sep = "")
            }
          })

library(shiny)
library(igvShiny)
library(htmlwidgets)

##  Install the latest version of this package by entering the following in R: 
# install.packages("remotes")
# remotes::install_github("paul-shannon/igvShiny")

tbl.wig <- data.frame(
  chrom=c("8","8", "8","8","20","20","20","20"),
  start=c(127735432, 127736432, 127734632,127735632,32358060,32359060,32369060,32399960),
  end=  c(127735462, 127736632, 127734632,127735932,32358260,32359160,32369260,32401060),
  score=c(-2.239, 3.0, 0.5, 5, 1,2.5, 0.2, 2),
  stringsAsFactors=FALSE
)

shinyServer(function(input, output, session) {
  output$igvShiny_0 <- igvShiny::renderIgvShiny(
    igvShiny::igvShiny(
      list(
        genomeName="hg38",
        initialLocus="MYC",
        displayMode="SQUISHED"
      )
    )
  )
  
  ## observeEvent(input$igvReady, {
  ## observeEvent(input$addAutoscaledGroupBedGraphTrackButton, {
  observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig1a", tbl=tbl.wig, color="blue",
                      autoscale=TRUE, autoscaleGroup=1)
    tbl.wig1b <- tbl.wig
    tbl.wig1b$score <-tbl.wig1b$score * 10
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig1b", tbl=tbl.wig1b, color="brown",
                      autoscale=TRUE, autoscaleGroup=1)
  })
  
  observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
    newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
    output$chromLocDisplay <- renderText({newLoc})
  })
  
  observeEvent(input$removeUserTracksButton, {
    print("---- removeUserTracks")
    removeUserAddedTracks(session, id="igvShiny_0")
  })
  observeEvent(input$searchButton, {
    print("--- search")
    searchString = isolate(input$roi)
    if(base::nchar(searchString) > 0)
      showGenomicRegion(session, id="igvShiny_0", searchString)
  }) 
  
})

library(shiny)
library(igvShiny)
library(dplyr)
library(GenomicAlignments)

#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))

printf <- function(...) print(noquote(sprintf(...)))

#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(

  sidebarLayout(
     sidebarPanel(
        actionButton("loadGwasTrackButton", "Load GWAS Track"),
        actionButton("showApoeButton", "APOE region"),
        actionButton("showAllButton",  "All Regions"),
        width=2
        ),
     mainPanel(
        igvShinyOutput('igvShiny_0'),
        width=10
        )
     ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {


   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <- parseAndValidateGenomeSpec(genomeName="hg19",  initialLocus="all")
     igvShiny(genomeOptions)
     })

   observeEvent(input$loadGwasTrackButton, {
      loadGwasTrack(session, id="igvShiny_0", trackName="demo", tbl=tbl.gwas, deleteTracksOfSameName=TRUE)
      })

   observeEvent(input$showApoeButton, {
       showGenomicRegion(session, "igvShiny_0", "chr19:45,303,720-45,463,047")
       })

   observeEvent(input$showAllButton, {
       showGenomicRegion(session, "igvShiny_0", "all")
       })

   #shiny::observeEvent(input$igvReady, ignoreInit=TRUE, {
   #   tbl.gwas <- get(load("../extdata/gwas.RData"))
   #   containerID <- input$igvReady
   #   loadGwasTrack(session, id="igvShiny_0", trackName="demo", tbl=tbl.gwas, deleteTracksOfSameName=TRUE)
   #   })

} # server

#------------------------------------------------------------------------------------------------------------------------
runApp(shinyApp(ui, server), port=6867)

library(shiny)
library(igvShiny)
library(htmlwidgets)
# library(biomaRt)

#----------------------------------------------------------------------------------------------------
data.directory <- system.file(package="igvShiny", "extdata")
fasta.file      <- file.path(data.directory, "ribosomal-RNA-gene.fasta")
fastaIndex.file <- file.path(data.directory, "ribosomal-RNA-gene.fasta.fai")
annotation.file <- file.path(data.directory, "ribosomal-RNA-gene.gff3")

genomeOptions <- parseAndValidateGenomeSpec(genomeName="ribo local",
                                            initialLocus="U13369.1:7,276-8,225",
                                            stockGenome=FALSE,
                                            dataMode="localFiles",
                                            fasta=fasta.file,
                                            fastaIndex=fastaIndex.file,
                                            genomeAnnotation=annotation.file)

#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(
    igvShinyOutput('igvShiny_0'),
    ))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {


   output$igvShiny_0 <- renderIgvShiny({
     cat("--- starting renderIgvShiny\n");
     x <- igvShiny(genomeOptions,
                   displayMode="SQUISHED",
                   )
     cat("--- ending renderIgvShiny\n");
     return(x)
     })

} # server
#----------------------------------------------------------------------------------------------------
if(grepl("hagfish", Sys.info()[["nodename"]]) & !interactive()){
   runApp(shinyApp(ui, server), port=6867)
   } else {
   shinyApp(ui, server)
   }

# THE FOLLOWING WAS MOVED OUT OF doc section for igvShiny
# param options a list, with required elements "genomeName" and "initialLocus".
#   Local or remote custom genomes can be used by setting "genomeName" to
#   "local" or "remote". The necessary fasta and index files are provided via
#   "fasta" and "index" arguments, either as path on disk or as URL.

#-------------------------------------------------------------------------------
# A list of safe, known igv.js track configuration parameters.
# This allowlist prevents arbitrary code injection and invalid options.
#
# Every name here has to be one igv.js actually reads. A name it does not know
# is worse than a rejected one: it passes validation, reaches the browser, and
# is ignored there, so the user gets no warning from either side and a track
# drawn with defaults. test-track-option-allowlist.R holds the list to that,
# checking each name against the bundled igv.js build.
.validIgvTrackOptions <- c(
  "name", "type", "format", "url", "indexURL", "indexed", "order",
  "displayMode", "color", "altColor", "borderColor",
  "trait", "height", "autoHeight", "minHeight", "maxHeight", "removable",
  ## seg tracks read their two gradients from these; the shorter negColor and
  ## posColor, which look like the obvious spelling, are read by nothing
  "negColorScale", "posColorScale",
  ## only the lower-case s spelling is a track option. autoScaleGroup used to
  ## sit here next to it, which meant a capitalised typo passed validation and
  ## produced an ungrouped track with nothing said about it
  "autoscale", "autoscaleGroup",
  "visibilityWindow", "searchable", "autoScale",
  "min", "max", "logScale", "graphType",
  "flipAxis", "stroke", "fill", "featureHeight",
  "font", "colorTable",
  ## the gwas parser reads these 1-based; without them it guesses the layout
  ## from the header names it knows, and any other spelling draws nothing
  "columns",
  "showAllBases", "samplingWindowSize", "samplingDepth", "maxRows",
  "oauthToken", "headers", "viewAsPairs", "pairsSupported",
  "wholeGenomeView", "roi", "queryable",
  ## alignment tracks sort their reads at load time from this object, the same
  ## one the igv.js right-click menu builds; "TAG" plus a tag covers #104
  "sort",
  ## splice junction filters, thickness and label rules, read by
  ## SpliceJunctionTrack straight off the track config. Names come from the
  ## igv.js 3.8.4 source, not from its spliceJunctionTrack.html example: that
  ## example sets labelUniqueReadCount and four siblings, and no such option
  ## exists in the library any more
  "minUniquelyMappedReads", "minTotalReads", "maxFractionMultiMappedReads",
  "minSplicedAlignmentOverhang", "minJunctionEndsVisible",
  "minSamplesWithThisJunction", "maxSamplesWithThisJunction",
  "minPercentSamplesWithThisJunction", "maxPercentSamplesWithThisJunction",
  "thicknessBasedOn", "bounceHeightBasedOn", "colorBy",
  "colorByNumReadsThreshold", "labelWith", "labelWithInParen",
  "hideAnnotatedJunctions", "hideUnannotatedJunctions", "hideMotifs",
  "hideStrand",
  ## MergedTrack passes this one down to every member track it draws, which is
  ## what makes overlaid coverage readable under the junction arcs
  "alpha"
  # Add other valid igv.js track options here as needed in the future
  # "tracks" is deliberately absent: it is meaningful on a merged track only,
  # and .sanitizeTrack handles it there.
)

#-------------------------------------------------------------------------------
# A merged track may itself hold merged tracks - igv.js builds its members
# through the same factory as top-level tracks, and nothing there stops the
# recursion. One level is all that has a use (coverage under junction arcs);
# the cap only bounds a pathological config, it states no policy.
.maxTrackNestingDepth <- 3L

#-------------------------------------------------------------------------------
#' Sanitize and merge track configuration options
#' @param baseOptions A list of default options set by the R function.
#' @param userOptions A list of options provided by the user via trackConfig.
#' @return A merged and sanitized list of options ready to be sent to
#' JavaScript.
#' @keywords igvShiny
.sanitizeAndMergeOptions <- function(baseOptions, userOptions) {
  if (is.null(userOptions) || length(userOptions) == 0) {
    return(baseOptions)
  }

  # anyNA guards the NA-name case: without it, `any(names == "")` evaluates to
  # NA when a name is NA and the `if` errors instead of warning and ignoring.
  if (!is.list(userOptions) || is.null(names(userOptions)) ||
      anyNA(names(userOptions)) || any(names(userOptions) == "")) {
    warning("trackConfig must be a named list. Ignoring.")
    return(baseOptions)
  }

  # Identify and warn about conflicting keys that would override explicit
  # function arguments
  conflictingKeys <- intersect(names(baseOptions), names(userOptions))
  if (length(conflictingKeys) > 0) {
    fmt <- paste("User-provided trackConfig options conflict with function",
                 "arguments and will be ignored: %s")
    warning(sprintf(fmt, toString(conflictingKeys)))
    # Prioritize base options for security and clarity
    userOptions[conflictingKeys] <- NULL
  }

  # Filter user options against the allowlist of valid igv.js parameters
  invalidKeys <- setdiff(names(userOptions), .validIgvTrackOptions)
  if (length(invalidKeys) > 0) {
    fmt <- paste("Ignoring invalid or unsupported track options in",
                 "trackConfig: %s")
    warning(sprintf(fmt, toString(invalidKeys)))
    userOptions[invalidKeys] <- NULL
  }

  # Merge the sanitized user options with the base options
  return(c(baseOptions, userOptions))
}
#-------------------------------------------------------------------------------
#' Sanitize a list of startup track specifications
#' @param tracks A list of named lists, each an igv.js track configuration.
#' @param depth An integer, how deep this list sits in a merged track; 1 for
#' the startup list itself.
#' @return A sanitized list of track configurations; invalid entries or keys
#' are dropped with a warning.
#' @keywords igvShiny
.sanitizeTracks <- function(tracks, depth = 1L) {
  if (is.null(tracks) || length(tracks) == 0) {
    return(list())
  }

  sanitized <- Filter(Negate(is.null), lapply(tracks, .sanitizeTrack,
                                              depth = depth))
  # Names have to go: igv.js walks a merged track with `for (let tconf of
  # this.config.tracks)` and calls .every() on the same value, and a named R
  # list serializes to a JSON object, which is neither iterable nor an array.
  unname(sanitized)
}
#-------------------------------------------------------------------------------
#' Sanitize one startup track specification
#' @param track A named list, one igv.js track configuration.
#' @param depth An integer, how deep this track sits in a merged track.
#' @return The sanitized track, or NULL if it cannot be used.
#' @keywords igvShiny
.sanitizeTrack <- function(track, depth) {
  if (!is.list(track) || is.null(names(track)) || any(is.na(names(track))) ||
      any(names(track) == "")) {
    msg <- paste("Ignoring invalid entry in 'tracks': each entry must be",
                 "a named list.")
    warning(msg)
    return(NULL)
  }

  # igv.js dispatches on config.type.toLowerCase(), so "Merged" is a merged
  # track there and has to be one here too.
  type <- track[["type"]]
  merged <- is.character(type) && length(type) == 1L && !is.na(type) &&
    identical(tolower(type), "merged")

  # "tracks" is a key only a merged track may carry, so it is pulled out of
  # the allowlist pass rather than added to the allowlist: on any other track
  # it stays unsupported and is dropped with the usual warning.
  members <- NULL
  if (merged) {
    members <- track[["tracks"]]
    track[["tracks"]] <- NULL
  }

  invalidKeys <- setdiff(names(track), .validIgvTrackOptions)
  if (length(invalidKeys) > 0) {
    fmt <- paste("Ignoring invalid or unsupported track options in",
                 "'tracks': %s")
    warning(sprintf(fmt, toString(invalidKeys)))
    track[invalidKeys] <- NULL
  }

  if (merged) {
    if (depth >= .maxTrackNestingDepth) {
      fmt <- paste("Dropping merged track nested more than %d levels deep in",
                   "'tracks'.")
      warning(sprintf(fmt, .maxTrackNestingDepth))
      return(NULL)
    }
    # Members go through this same allowlist, so a merged track widens no
    # boundary; it only relaxes where the url has to sit.
    members <- .sanitizeTracks(members, depth = depth + 1L)
    if (length(members) == 0) {
      msg <- paste("Dropping merged track in 'tracks' with no usable member",
                   "track.")
      warning(msg)
      return(NULL)
    }
    track[["tracks"]] <- members
    return(track)
  }

  # A usable url is a single non-empty, non-NA character string. is.null()
  # alone let NA, "", character(0) and non-character values through to
  # igv.js as broken tracks.
  url <- track[["url"]]
  if (!is.character(url) || length(url) != 1L || is.na(url) ||
      !nzchar(url)) {
    msg <- paste("Dropping entry in 'tracks' with no valid 'url' after",
                 "removing unsupported options.")
    warning(msg)
    return(NULL)
  }
  track
}
#-------------------------------------------------------------------------------
#' Create an igvShiny instance
#'
#' @description This function is called in the server function of your shiny app
#'
#' @rdname igvShiny
#' @aliases igvShiny
#'
#' @import BiocGenerics
#' @import GenomicRanges
#' @import GenomeInfoDbData
#' @import shiny
#' @importFrom randomcoloR distinctColorPalette
#' @import httr
#' @importFrom htmlwidgets createWidget shinyWidgetOutput shinyRenderWidget
#' @importFrom futile.logger flog.debug
#'
#' @param genomeOptions a list with these fields: genomeName, initialLocus,
#' annotation, dataMode, fasta, fastaIndex, stockGenome, validated
#' @param width a character string, standard css notations,
#' either e.g., "1000px" or "95\%"
#' @param height a character string, needs to be an explicit pixel measure,
#' e.g., "800px"
#' @param elementId a character string, the html element id within which
#' igv is created
#' @param displayMode a character string, default "SQUISHED".
#' @param tracks a list of track specifications to be created and displayed
#' at startup. Each element is itself a named list of igv.js track options
#' (e.g. \code{name}, \code{type}, \code{format}, \code{url}), for example:
#' \code{list(list(name="genes", type="annotation", format="gff3",
#' url="https://.../genes.gff3"))}. Unrecognized keys are dropped with a
#' warning; see \code{.validIgvTrackOptions} for the full allowlist.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return the created widget
#'
#' @keywords igvShiny
#' @export
#'
igvShiny <- function(genomeOptions,
                     width = NULL,
                     height = NULL,
                     elementId = NULL,
                     displayMode = "squished",
                     tracks = list()) {
  stopifnot(
    sort(names(genomeOptions)) ==
      c(
        "annotation",
        "dataMode",
        "fasta",
        "fastaIndex",
        "genomeName",
        "initialLocus",
        "stockGenome",
        "validated"
      )
  )
  stopifnot(genomeOptions[["validated"]])

  if (!genomeOptions[["stockGenome"]] &&
        genomeOptions[["dataMode"]] == "localFiles") {
    directory.name <- .tracksDir()
    fasta.file <- genomeOptions[["fasta"]]
    fasta.indexFile <- genomeOptions[["fastaIndex"]]
    gff3.file <- genomeOptions[["annotation"]]
    destination <-
      file.path(directory.name, basename(fasta.file))
    file.copy(fasta.file, destination, overwrite = TRUE)
    destination <-
      file.path(directory.name, basename(fasta.indexFile))
    file.copy(fasta.indexFile, destination, overwrite = TRUE)
    if (!is.na(gff3.file)) {
      destination <- file.path(directory.name, basename(gff3.file))
      file.copy(gff3.file, destination, overwrite = TRUE)
      genomeOptions[["annotation"]] <-
        file.path(basename(directory.name), basename(gff3.file))
    }
    # now that they have been copied, store the new paths
    genomeOptions[["fasta"]] <-
      file.path(basename(directory.name), basename(fasta.file))
    genomeOptions[["fastaIndex"]] <-
      file.path(basename(directory.name), basename(fasta.indexFile))
  } # if custom genome, local files

  state[["requestedHeight"]] <- height

  flog.debug("---igvShiny ctor")
  flog.debug(sprintf("--initial track count: %d", length(tracks)))

  #send namespace info in case widget is being called from a module
  session <- shiny::getDefaultReactiveDomain()
  genomeOptions$displayMode <- displayMode
  genomeOptions$trackHeight <-
    100      # todo: make this an igvShiny ctor argument
  # outside a shiny session (script, vignette) there is no namespace to
  # prepend; the JS side concatenates moduleNS with the event name, so the
  # no-module case is the empty string
  genomeOptions$moduleNS <- if (is.null(session)) "" else session$ns("")
  genomeOptions$tracks <- .sanitizeTracks(tracks)

  htmlwidgets::createWidget(
    name = "igvShiny",
    genomeOptions,
    width = width,
    height = height,
    package = "igvShiny",
    elementId = elementId
  )

} # igvShiny constructor
#-------------------------------------------------------------------------------
#' create the UI for the widget
#'
#' @description This function is called in the ui function of your shiny app
#'
#' @rdname igvShinyOutput
#' @aliases igvShinyOutput
#'
#' @param outputId a character string, specifies the html element id
#' @param width a character string, standard css notations,
#' either e.g., "1000px" or "95\%", "100\%" by default
#' @param height a character string, needs to be an explicit pixel measure,
#' e.g., "800px", "400px" by default
#'
#' @return the created widget's html
#'
#' @examples
#' io <- igvShinyOutput("igvOut")
#'
#' @keywords igvShiny
#' @export
#'
igvShinyOutput <- function(outputId,
                           width = "100%",
                           height = NULL) {
  if ("requestedHeight" %in% ls(state)) {
    flog.debug("setting height from state")
    height <- state[["requestedHeight"]]
  }

  htmlwidgets::shinyWidgetOutput(outputId,
                                 "igvShiny",
                                 width,
                                 height,
                                 package = "igvShiny")
}

#-------------------------------------------------------------------------------
#' draw the igv genome browser element
#'
#' @description This function is called in the server function of your shiny app
#'
#' @rdname renderIgvShiny
#' @aliases renderIgvShiny
#'
#' @param expr an expression that generates an HTML widget
#' @param env  the environment in which to evaluate expr
#' @param quoted logical flag indicating if expr a quoted expression
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return an output or render function that enables the use of the widget
#' within Shiny applications
#'
#' @keywords igvShiny
#' @export
renderIgvShiny <- function(expr,
                           env = parent.frame(),
                           quoted = FALSE) {
  if (!quoted) {
    expr <- substitute(expr)
  } # force quoted

  x <- htmlwidgets::shinyRenderWidget(expr,
                                      igvShinyOutput,
                                      env,
                                      quoted = TRUE)
  flog.debug("--- leaving igvShiny.R, renderIgvShiny")
  return(x)

}

#-------------------------------------------------------------------------------
#' focus igv on a region
#'
#' @description zoom in or out to show the nominated region, by chromosome locus
#' or gene symbol
#'
#' @rdname showGenomicRegion
#' @aliases showGenomicRegion
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param region a character string, either e.g. "chr5:92,221,640-92,236,523"
#' or "MEF2C"
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @keywords igvShiny
#' @export
showGenomicRegion <- function(session, id, region) {
  message <- list(region = region, elementID = id)
  session$sendCustomMessage("showGenomicRegion", message)
} # showGenomicRegion

#-------------------------------------------------------------------------------
#' return the current igv region
#'
#' @description return the current region displayed by your igv instance
#'
#' @rdname showGenomicRegion
#' @aliases showGenomicRegion
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' a character string of format "chrom:start-end"
#'
#' @keywords igvShiny
#' @export

getGenomicRegion <- function(session, id) {
  message <- list(elementID = id)
  session$sendCustomMessage("getGenomicRegion", message)
} # gertGenomicRegion

#-------------------------------------------------------------------------------
#' remove tracks from the browser
#'
#' @description delete tracks on the browser
#'
#' @rdname removeTracksByName
#' @aliases removeTracksByName
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackNames a vector of character strings

#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords igvShiny
#' @export
removeTracksByName <- function(session, id, trackNames) {
  message <- list(trackNames = trackNames, elementID = id)
  lmsg <-
    sprintf("--- igvShiny sending message to js, removeTracksByName, %s",
            toString(trackNames))
  flog.debug(lmsg)
  session$sendCustomMessage("removeTracksByName", message)

} # removeTracksByName

#-------------------------------------------------------------------------------
#' remove only those tracks explicitly added by your app
#'
#' @description remove only those tracks explicitly added by your app.
#' stock tracks (i.e., #' Refseq Genes) remain
#'
#' @rdname removeUserAddedTracks
#' @aliases removeUserAddedTracks
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords igvShiny
#' @export

removeUserAddedTracks <- function(session, id) {

  removeTracksByName(session, id, state[["userAddedTracks"]])
  state[["userAddedTracks"]] <- list()

} # removeUserAddedTracks

#-------------------------------------------------------------------------------
#' load a bed track provided as a data.frame
#'
#' @description load a bed track provided as a data.frame
#'
#' @rdname loadBedTrack
#' @aliases loadBedTrack
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param tbl data.frame, with at least "chrom" "start" "end" columns
#' @param color character string, a legal CSS color, or "random",
#' "gray" by default
#' @param trackHeight an integer, 50 (pixels) by default
#' @param deleteTracksOfSameName logical, default TRUE
#' @param quiet logical, default TRUE, controls verbosity
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadBedTrack <-
  function(session,
           id,
           trackName,
           tbl,
           color = "",
           trackHeight = 50,
           deleteTracksOfSameName = TRUE,
           quiet = TRUE,
           trackConfig = list()) {
    if (color == "random")
      color <-
        randomColors[sample(seq_along(randomColors), 1)]

    if (!quiet) {
      flog.debug("--- igvShiny::loadBedTrack")

      flog.debug(sprintf("rows: %d  cols: %d", NROW(tbl), NCOL(tbl)))
    }

    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    if (colnames(tbl)[1] == "chrom")
      colnames(tbl)[1] <- "chr"

    if (all(colnames(tbl)[c(1, 2, 3)] != c("chr", "start", "end"))) {
      lmsg <- sprintf("found these colnames: %s",
                      toString(colnames(tbl)))
      lmsg2 <- sprintf("            required: %s",
                       toString(c("chr", "start", "end")))
      flog.debug(lmsg)
      flog.debug(lmsg2)
      stop("improper columns in bed track data.frame")
    }

    stopifnot(is(tbl$chr, "character"))
    stopifnot(is(tbl$start, "numeric"))
    stopifnot(is(tbl$end, "numeric"))
    new.order <- order(tbl$start, decreasing = FALSE)
    tbl <- tbl[new.order, ]

    temp.file <-
      .trackFile(session, ".bed")
    write.table(
      tbl,
      sep = "\t",
      row.names = FALSE,
      col.names = FALSE,
      quote = FALSE,
      file = temp.file
    )
    lmsg <- sprintf("--- igvShiny.R, loadBedTrack wrote %d,%d to %s",
                    NROW(tbl),
                    NCOL(tbl),
                    temp.file)
    flog.debug(lmsg)
    flog.debug(sprintf("exists? %s", file.exists(temp.file)))

    base.msg.to.igv <- list(
      elementID = id,
      trackName = trackName,
      bedFilepath = file.path("tracks", basename(temp.file)),
      color = color,
      trackHeight = trackHeight
    )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadBedTrackFromFile", msg.to.igv)

  } # loadBedTrack

#-------------------------------------------------------------------------------
#' load a bedgraph track from a URL
#'
#' @description load a bedgraph track provided as a data.frame
#'
#' @rdname loadBedGraphTrackFromURL
#' @aliases loadBedGraphTrackFromURL
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param url character
#' @param color character string, a legal CSS color, or "random",
#' "gray" by default
#' @param trackHeight an integer, 30 (pixels) by default
#' @param autoscale logical
#' @param min numeric, consulted when autoscale is FALSE
#' @param max numeric, consulted when autoscale is FALSE
#' @param quiet logical, default TRUE, controls verbosity
#' @param autoscaleGroup numeric(1) defaults to -1
#' @param deleteTracksOfSameName logical(1) defaults to TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadBedGraphTrackFromURL <-
  function(session,
           id,
           trackName,
           url,
           color = "gray",
           trackHeight = 30,
           autoscale = TRUE,
           min = 0,
           max = 1,
           autoscaleGroup = -1,
           deleteTracksOfSameName = TRUE,
           quiet = TRUE,
           trackConfig = list()) {
    message("---- loadBedGraphTrackFromURL")

    if (color == "random")
      color <-
        randomColors[sample(seq_along(randomColors), 1)]

    if (!quiet) {
      lmsg <- sprintf("--- igvShiny::loadBedGraphTrackFromURL: %s",
                      trackName)
      flog.debug(lmsg)
    }

    if (deleteTracksOfSameName) {
      lmsg <- sprintf(
        "--- loadBedGraphTrackFromURL, calling removeTracksByName: %s, %s",
        id,
        trackName
      )
      flog.debug(lmsg)
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        url = url,
        color = color,
        trackHeight = trackHeight,
        autoscale = autoscale,
        min = min,
        max = max,
        autoscaleGroup = autoscaleGroup
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)

    flog.debug("--- igvShiny.R loadBedGraphTrackFromURL, msg.to.igv: ")
    futile.logger::flog.info(jsonlite::toJSON(msg.to.igv))
    flog.debug("--- igvShiny.R loadBedGraphTrackFromURL, sendingCustomMessage")
    session$sendCustomMessage("loadBedGraphTrackFromURL", msg.to.igv)
    flog.debug("--- loadBedGraphTrackFromURL, after sendingCustomMessage")

  } # loadBedGraphTrackFromURL

#-------------------------------------------------------------------------------
#' load a scored genome annotation track provided as a data.frame
#'
#' @description load a genome annotation track provided as a data.frame
#'
#' @rdname loadGenomeAnnotationTrack
#' @aliases loadGenomeAnnotationTrack
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param tbl data.frame, with at least "chrom" "start" "end" "score" columns
#' @param color character string, a legal CSS color, or "random",
#' "gray" by default
#' @param trackHeight an integer, 30 (pixels) by default
#' @param autoscale logical
#' @param autoscaleGroup numeric(1) defaults to -1
#' @param min numeric, consulted when autoscale is FALSE
#' @param max numeric, consulted when autoscale is FALSE
#' @param deleteTracksOfSameName logical, default TRUE
#' @param quiet logical, default TRUE, controls verbosity
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadBedGraphTrack <-
  function(session,
           id,
           trackName,
           tbl,
           color = "gray",
           trackHeight = 30,
           autoscale,
           autoscaleGroup = -1,
           min = NA_real_,
           max = NA_real_,
           deleteTracksOfSameName = TRUE,
           quiet = TRUE,
           trackConfig = list()) {
    stopifnot(NCOL(tbl) >= 4)

    if (color == "random")
      color <-
        randomColors[sample(seq_along(randomColors), 1)]

    if (!quiet) {
      flog.debug("--- igvShiny::loadGenomeAnnotationTrack: %s",
                 trackName)
      flog.debug("    %d rows, %d columns", NROW(tbl), NCOL(tbl))
    }

    if (deleteTracksOfSameName) {
      flog.debug(
        "--- igvShiny.R loadBedGraphTrack, calling removeTracksByName: %s, %s",
        id,
        trackName
      )
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    if (colnames(tbl)[1] == "chrom")
      colnames(tbl)[1] <- "chr"

    colnames(tbl)[4] <- "value"

    if (all(colnames(tbl)[c(1, 2, 3)] != c("chr", "start", "end"))) {
      flog.debug("found these colnames: %s",
                 toString(colnames(tbl)[c(1, 2, 3)]))
      flog.debug("            required: %s",
                 toString(c("chr", "start", "end")))
      stop("improper columns in bed track data.frame")
    }

    stopifnot(is(tbl$chr, "character"))
    stopifnot(is(tbl$start, "numeric"))
    stopifnot(is(tbl$end, "numeric"))
    stopifnot(is(tbl$value, "numeric"))

    new.order <- order(tbl$start, decreasing = FALSE)
    tbl <- tbl[new.order, ]

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        tbl = jsonlite::toJSON(tbl),
        color = color,
        trackHeight = trackHeight,
        autoscale = autoscale,
        min = min,
        max = max,
        autoscaleGroup = autoscaleGroup
      )  # -1 means no grouping

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadBedGraphTrack", msg.to.igv)

  } # loadBedGraphTrack
#-------------------------------------------------------------------------------
#' load a seg track provided as a data.frame
#'
#' @description load a SEG track provided as a data.frame.  igv "displays
#' segmented data as a blue-to-red heatmap where the data range is
#' -1.5 to 1.5... The segmented data file format is the output of
#' the Circular Binary Segmentation algorithm (Olshen et al., 2004)".
#'
#' @rdname loadSEGTrack
#' @aliases loadSEGTrack
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param tbl data.frame, with at least "chrom" "start" "end" "score" columns
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export
loadSegTrack <-
  function(session,
           id,
           trackName,
           tbl,
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {
    flog.debug("--- entering loadSegTrack %s with %d rows",
               trackName,
               NROW(tbl))

    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        tbl = jsonlite::toJSON(tbl)
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    flog.debug("about to send loadSegTrack message")
    session$sendCustomMessage("loadSegTrack", msg.to.igv)

  } # loadSegTrack

#-------------------------------------------------------------------------------
#' load a VCF (variant) track provided as a Bioconductor
#' VariantAnnotation object
#'
#' @description load a VCF (variant) track provided as a Bioconductor
#' VariantAnnotation object
#'
#' @rdname loadVcfTrack
#' @aliases loadVcfTrack
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param vcfData VariantAnnotation object
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "demos", "local-data.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadVcfTrack <- function(session,
                         id,
                         trackName,
                         vcfData,
                         deleteTracksOfSameName = TRUE,
                         trackConfig = list()) {
  if (!requireNamespace("VariantAnnotation"))
    stop("install VariantAnnotation to use this function")

  flog.debug("======== igvShiny.R, loadVcfTrack")
  if (deleteTracksOfSameName) {
    removeTracksByName(session, id, trackName)
  }

  state[["userAddedTracks"]] <-
    unique(c(state[["userAddedTracks"]], trackName))
  temp.file <- .trackFile(session, ".vcf")
  lmsg <- sprintf("igvShiny::loadVcfTrack, about to write to file '%s'",
                  temp.file)
  flog.debug(lmsg)
  VariantAnnotation::writeVcf(vcfData, temp.file)
  lmsg2 <- sprintf("igvShiny::loadVcfTrack, file.exists(%s)? %s",
                   temp.file,
                   file.exists(temp.file))
  flog.debug(lmsg2)

  path <- file.path("tracks", basename(temp.file))
  base.msg.to.igv <-
    list(
      elementID = id,
      trackName = trackName,
      vcfDataFilepath = path
    )

  msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
  session$sendCustomMessage("loadVcfTrack", msg.to.igv)

} # loadVcfTrack

#-------------------------------------------------------------------------------
#' load a GWAS (genome-wide association study) track
#' provided as a data.frame
#'
#' @description load a GWAS (genome-wide association study) track
#' provided as a data.frame
#'
#' @rdname loadGwasTrack
#' @aliases loadGwasTrack
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param ymin numeric defaults to 0
#' @param ymax numeric defaults to 35
#' @param tbl.gwas data.frame, with at least "chrom" "start" "end" columns
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options. A table whose headers igv.js does not recognize needs its layout
#' spelled out, 1-based, as
#' \code{columns = list(chromosome = 3, position = 4, value = 10)};
#' \code{colorTable} takes a chromosome-to-color map, see
#' \code{\link{GWASTrack}}
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export
loadGwasTrack <- function(session,
                          id,
                          trackName,
                          tbl.gwas,
                          ymin = 0,
                          ymax = 35,
                          deleteTracksOfSameName = TRUE,
                          trackConfig = list()) {
  flog.debug("======== entering igvShiny::loadGwasTrack")

  if (deleteTracksOfSameName) {
    removeTracksByName(session, id, trackName)
  }

  state[["userAddedTracks"]] <-
    unique(c(state[["userAddedTracks"]], trackName))

  temp.file <-
    .trackFile(session, ".gwas")
  write.table(
    tbl.gwas,
    sep = "\t",
    row.names = FALSE,
    quote = FALSE,
    file = temp.file
  )
  lmsg <- sprintf(
    "--- igvShiny.R, loadGwasTrack wrote %d,%d to %s",
    NROW(tbl.gwas),
    NCOL(tbl.gwas),
    temp.file
  )
  flog.debug(lmsg)
  flog.debug(sprintf("exists? %s", file.exists(temp.file)))
  base.msg.to.igv <-
    list(
      elementID = id,
      trackName = trackName,
      gwasDataFilepath = file.path("tracks", basename(temp.file)),
      color = "red",
      trackHeight = 200,
      autoscale = FALSE,
      min = ymin,
      max = ymax
    )

  msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
  session$sendCustomMessage("loadGwasTrack", msg.to.igv)

} # loadGwasTrack

#-------------------------------------------------------------------------------
#' load a bam track which, with index, is served up by http
#'
#' @description load a remote bam track
#'
#' @rdname loadBamTrackFromURL
#' @aliases loadBamTrackFromURL
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param bamURL character string http url for the bam file,
#' typically very large
#' @param indexURL character string http url for the bam file index,
#' typically small
#' @param deleteTracksOfSameName logical, default TRUE
#' @param displayMode character string, possible values are "EXPANDED"(default),
#' "SQUISHED" or "COLLAPSED"
#' @param showAllBases logical, show all bases in the alignment, default FALSE
#' @param trackConfig a named list of additional igv.js track configuration
#' options. Alignment tracks read \code{sort}: \code{list(sort =
#' list(chr = "chr1", position = 155160540, option = "TAG", tag = "HP"))}
#' sorts the reads at that 1-based position by the HP tag. Any igv.js sort
#' option works (BASE, STRAND, INSERT_SIZE, ...); reads sort descending
#' unless \code{direction = "ASC"}.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadBamTrackFromURL <-
  function(session,
           id,
           trackName,
           bamURL,
           indexURL,
           deleteTracksOfSameName = TRUE,
           displayMode = "EXPANDED",
           showAllBases = FALSE,
           trackConfig = list()) {
    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))
    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        bam = bamURL,
        index = indexURL,
        displayMode = displayMode,
        showAllBases = showAllBases
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    flog.debug("--- about to send message, loadBamTrack")
    session$sendCustomMessage("loadBamTrackFromURL", msg.to.igv)

  } # loadBamTrackFromURL

#-------------------------------------------------------------------------------
#' load GenomicAlignments data as an igv.js alignment track
#'
#' @description load GenomicAlignments data  as an igv.js alignment track
#' @rdname loadBamTrackFromLocalData
#' @aliases loadBamTrackFromLocalData
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param data  GenomicAlignments object
#' @param deleteTracksOfSameName logical, default TRUE
#' @param displayMode character string, possible values are "EXPANDED"(default),
#' "SQUISHED" or "COLLAPSED"
#' @param trackConfig a named list of additional igv.js track configuration
#' options, \code{sort} among them; see \code{\link{loadBamTrackFromURL}}.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadBamTrackFromLocalData <-
  function(session,
           id,
           trackName,
           data,
           deleteTracksOfSameName = TRUE,
           displayMode = "EXPANDED",
           trackConfig = list()) {
    if (!requireNamespace("rtracklayer"))
      stop("install rtracklayer to use loadBamTrackFromLocalData")
    if (!requireNamespace("Rsamtools"))
      stop("install Rsamtools to use loadBamTrackFromLocalData")
    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    fpath <- .trackFile(session, ".bam")
    # rtracklayer indexes the bam it writes, next to it and under a name we
    # never see; it is served alongside, so it goes at session end too
    .registerTrackFile(session, paste0(fpath, ".bai"))

    lmsg <-
      sprintf("igvShiny::load bam from local data, about to write to file '%s'",
              fpath)
    flog.debug(lmsg)
    rtracklayer::export(data, fpath, format = "BAM")

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        bamDataFilepath = file.path("tracks", basename(fpath)),
        displayMode = displayMode
      )
    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadBamTrackFromLocalData", msg.to.igv)

  } # loadBamTrackFromLocalData

#-------------------------------------------------------------------------------
#' load a cram track which, with index, is served up by http
#'
#' @description load a remote cram track
#'
#' @rdname loadCramTrackFromURL
#' @aliases loadCramTrackFromURL
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param cramURL character string http url for the bam file,
#' typically very large
#' @param indexURL character string http url for the bam file index,
#' typically small
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options, \code{sort} among them; see \code{\link{loadBamTrackFromURL}}.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadCramTrackFromURL <-
  function(session,
           id,
           trackName,
           cramURL,
           indexURL,
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {
    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        cram = cramURL,
        index = indexURL
      )
    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadCramTrackFromURL", msg.to.igv)

  } # loadCramTrackFromURL

#-------------------------------------------------------------------------------
#' load a cram file sitting on the same machine as the shiny app
#'
#' @description load a local cram track. Unlike \code{loadBamTrackFromLocalData}
#' this loader takes file paths, not an R object: no bioconductor package parses
#' cram, so the file and its index are handed to igv.js untouched, through the
#' directory shiny serves as "tracks".
#'
#' @rdname loadCramTrackFromLocalData
#' @aliases loadCramTrackFromLocalData
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param cramFile character string, path to a cram file
#' @param indexFile character string, path to its crai index,
#' by default the cram file with ".crai" appended
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options, \code{sort} among them; see \code{\link{loadBamTrackFromURL}}.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadCramTrackFromLocalData <-
  function(session,
           id,
           trackName,
           cramFile,
           indexFile = paste0(cramFile, ".crai"),
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {
    checkmate::assert_file_exists(cramFile, access = "r")
    checkmate::assert_file_exists(indexFile, access = "r")
    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    cramPath <- .stageTrackFile(session, cramFile, ".cram")
    indexPath <- .stageTrackFile(session, indexFile, ".crai")
    flog.debug(sprintf("igvShiny::load local cram, serving '%s'", cramPath))

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        cram = cramPath,
        index = indexPath
      )
    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    # the payload is what the remote loader sends, only with app-relative urls,
    # so the browser side needs no handler of its own
    session$sendCustomMessage("loadCramTrackFromURL", msg.to.igv)

  } # loadCramTrackFromLocalData

#-------------------------------------------------------------------------------
#' load a GFF3 track which, with index, is served up by http
#'
#' @description load a remote GFF3 track
#'
#' @rdname loadGFF3TrackFromURL
#' @aliases loadGFF3TrackFromURL
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param trackHeight numeric defaults to 50
#' @param gff3URL character string http url for the bam file,
#' typically very large
#' @param indexURL character string http url for the bam file index,
#' typically small
#' @param color character #RGB or a recognized color name.  ignored if
#' colorTable and colorByAttribute provided
#' @param colorTable list, mapping a gff3 attribute, typically biotype,
#' to a color
#' @param colorByAttribute character, name of a gff3 attribute in column 9,
#' typically "biotype"
#' @param displayMode character,  "EXPANDED",  "SQUISHED" or "COLLAPSED"
#' @param visibilityWindow numeric, Maximum window size in base pairs
#' for which indexed annotations or variants are displayed
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "demos", "local-data.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadGFF3TrackFromURL <-
  function(session,
           id,
           trackName,
           gff3URL,
           indexURL,
           color = "gray",
           colorTable,
           colorByAttribute,
           displayMode,
           trackHeight = 50,
           visibilityWindow,
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {

    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        dataURL = gff3URL,
        indexURL = indexURL,
        color = color,
        colorTable = colorTable,
        colorByAttribute = colorByAttribute,
        displayMode = displayMode,
        trackHeight = trackHeight,
        visibilityWindow = visibilityWindow
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadGFF3TrackFromURL", msg.to.igv)

  } # loadGFF3TrackFromURL
#-------------------------------------------------------------------------------
#' load a GFF3 track defined by local data
#'
#' @description load a local GFF3 track file
#'
#' @rdname loadGFF3TrackFromLocalData
#' @aliases loadGFF3TrackFromLocalData
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param trackHeight numeric defaults to 50
#' @param tbl.gff3 data.frame  in standard 9-column GFF3 format
#' @param color character #RGB or a recognized color name.  ignored if
#' colorTable and colorByAttribute provided
#' @param colorTable list, mapping a gff3 attribute, typically biotype,
#' to a color
#' @param colorByAttribute character, name of a gff3 attribute in column 9,
#' typically "biotype"
#' @param displayMode character,  "EXPANDED",  "SQUISHED" or "COLLAPSED"
#' @param visibilityWindow numeric, Maximum window size in base pairs
#' for which indexed annotations or variants are displayed
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "demos", "local-data.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadGFF3TrackFromLocalData <-
  function(session,
           id,
           trackName,
           tbl.gff3,
           color = "gray",
           colorTable,
           colorByAttribute,
           displayMode,
           trackHeight = 50,
           visibilityWindow,
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {
    flog.debug("--- entering loadGFF3TrackFromLocalData")

    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    gff3.filePath <-
      .trackFile(session, ".gff3")
    write.table(
      tbl.gff3,
      sep = "\t",
      row.names = FALSE,
      quote = FALSE,
      file = gff3.filePath
    )
    lmsg <- sprintf(
      "--- igvShiny.R, loadGFF3TrackFromLocalData wrote %d,%d to %s",
      NROW(tbl.gff3),
      NCOL(tbl.gff3),
      gff3.filePath
    )
    flog.debug(lmsg)

    flog.debug(sprintf("exists? %s", file.exists(gff3.filePath)))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        filePath = file.path("tracks", basename(gff3.filePath)),
        color = color,
        colorTable = colorTable,
        colorByAttribute = colorByAttribute,
        displayMode = displayMode,
        trackHeight = trackHeight,
        visibilityWindow = visibilityWindow
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadGFF3TrackFromLocalData", msg.to.igv)

  } # loadGFF3TrackFromLocalData
#-------------------------------------------------------------------------------
#' load a splice junction track served up by http
#'
#' @description load splice junctions from a BED file reachable by URL, as
#' written by STAR (\code{SJ.out.tab} converted to BED). Six columns, with the
#' per-junction attributes packed into the name column as \code{key=value}
#' pairs separated by semicolons: \code{motif}, \code{uniquely_mapped},
#' \code{multi_mapped}, \code{maximum_spliced_alignment_overhang} and
#' \code{annotated_junction}. The track reads its filters and labels from
#' those attributes.
#'
#' igv.js draws junctions from a file of that shape only - it derives none of
#' them from a bam file, and it has no sashimi plot. Show this track above a
#' coverage track (bigWig, bedGraph) for the same sample to get the
#' arcs-over-coverage view sashimi plots are wanted for.
#'
#' @rdname loadSpliceJunctionTrackFromURL
#' @aliases loadSpliceJunctionTrackFromURL
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param url character string http url for the bed file of junctions
#' @param indexURL character string http url for a tabix index, needed only
#' for a bgzipped bed; "" by default, which loads the file whole
#' @param trackHeight an integer, 100 (pixels) by default
#' @param displayMode character, "COLLAPSED", "EXPANDED" or "SQUISHED"
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options. The junction ones are read straight off it:
#' \code{minUniquelyMappedReads}, \code{minTotalReads},
#' \code{maxFractionMultiMappedReads}, \code{minSplicedAlignmentOverhang},
#' \code{thicknessBasedOn}, \code{bounceHeightBasedOn}, \code{colorBy},
#' \code{labelWith}, \code{hideAnnotatedJunctions},
#' \code{hideUnannotatedJunctions}, \code{hideMotifs} among them.
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadSpliceJunctionTrackFromURL <-
  function(session,
           id,
           trackName,
           url,
           indexURL = "",
           trackHeight = 100,
           displayMode = "COLLAPSED",
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {
    # A usable index is a single non-empty, non-NA character string; anything
    # else means "no index". NULL is the trap: list() keeps it, shiny sends it
    # as JSON null, and the handler reads .length off it. That throws, so the
    # track never loads and R sees nothing at all - the failure shows up only
    # in the browser console.
    if (!is.character(indexURL) || length(indexURL) != 1L || is.na(indexURL)) {
      if (length(indexURL) > 0L) {
        warning("indexURL must be a single character string. Ignoring it.")
      }
      indexURL <- ""
    }

    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        url = url,
        indexURL = indexURL,
        trackHeight = trackHeight,
        displayMode = displayMode
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    session$sendCustomMessage("loadSpliceJunctionTrackFromURL", msg.to.igv)

  } # loadSpliceJunctionTrackFromURL
#-------------------------------------------------------------------------------
# The per-junction attributes SpliceJunctionTrack reads off the bed name
# column, in the spelling igv.js looks for.
.junctionAttributeColumns <-
  c("motif", "uniquely_mapped", "multi_mapped",
    "maximum_spliced_alignment_overhang", "annotated_junction")

#' Pack junction attributes into a bed name column
#' @param tbl A data.frame of junctions.
#' @return A character vector, one packed name field per row.
#' @keywords igvShiny
.packJunctionAttributes <- function(tbl) {
  columns <- intersect(.junctionAttributeColumns, colnames(tbl))
  if (length(columns) == 0) {
    return(rep(".", NROW(tbl)))
  }

  asAttribute <- function(values) {
    # igv.js compares annotated_junction against the strings "true" and
    # "false", and R would write a logical column as TRUE/FALSE, which matches
    # neither: the junction would be neither annotated nor unannotated and
    # both hide options would pass it through.
    if (is.logical(values)) {
      return(tolower(as.character(values)))
    }
    as.character(values)
  }

  packed <- lapply(columns, function(column) {
    paste0(column, "=", asAttribute(tbl[[column]]))
  })
  packed <- do.call(paste, c(packed, list(sep = ";")))

  # The bed decoder reads column 4 as attributes only when it holds a "=" and
  # a ";" past its first character. A single pair carries no ";", so it would
  # arrive as a plain track label and every junction filter would quietly find
  # nothing to read. A trailing ";" is dropped by the parser, so it costs
  # nothing and keeps one attribute working like five.
  if (length(columns) == 1L) {
    packed <- paste0(packed, ";")
  }
  packed
} # .packJunctionAttributes
#-------------------------------------------------------------------------------
#' load a splice junction track from a data.frame
#'
#' @description load splice junctions held in R. The table is written as the
#' six-column bed igv.js draws junctions from, and served from the same
#' directory as the other local-data tracks.
#'
#' Coordinates are taken as bed coordinates and written out unchanged, as the
#' other \code{*FromLocalData} loaders take theirs. STAR reports the first and
#' last intron base of a junction 1-based in \code{SJ.out.tab}, so a table read
#' straight from that file needs its start shifted by one first; a bed already
#' converted from it, which is what STAR wrappers and the igv.js test fixtures
#' ship, is passed in as is.
#'
#' Columns \code{chrom} (or \code{chr}), \code{start} and \code{end} are
#' required. \code{score} and \code{strand} are used when present. The
#' per-junction attributes \code{motif}, \code{uniquely_mapped},
#' \code{multi_mapped}, \code{maximum_spliced_alignment_overhang} and
#' \code{annotated_junction} are packed into the bed name column, which is
#' where the track reads its filters and labels from; supply the ones you
#' want to filter or label on. A \code{name} column, if you have already
#' packed it yourself, is written through untouched.
#'
#' Without \code{uniquely_mapped} or an explicit \code{score} the arcs all
#' draw at the same thickness: igv.js has nothing to size them by.
#'
#' @rdname loadSpliceJunctionTrackFromLocalData
#' @aliases loadSpliceJunctionTrackFromLocalData
#'
#' @param session an environment or list, provided and managed by shiny
#' @param id character string, the html element id of this widget instance
#' @param trackName character string
#' @param tbl data.frame, with at least "chrom" "start" "end" columns, in bed
#' coordinates: a 0-based start and an exclusive end
#' @param trackHeight an integer, 100 (pixels) by default
#' @param displayMode character, "COLLAPSED", "EXPANDED" or "SQUISHED"
#' @param deleteTracksOfSameName logical, default TRUE
#' @param trackConfig a named list of additional igv.js track configuration
#' options, the junction ones among them; see
#' \code{\link{loadSpliceJunctionTrackFromURL}}
#'
#' @examples
#' library(igvShiny)
#' demo_app_file <-
#'   system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
#' if (interactive()) {
#'   shiny::runApp(demo_app_file)
#' }
#'
#' @return
#' nothing
#'
#' @keywords track_loaders
#' @export

loadSpliceJunctionTrackFromLocalData <-
  function(session,
           id,
           trackName,
           tbl,
           trackHeight = 100,
           displayMode = "COLLAPSED",
           deleteTracksOfSameName = TRUE,
           trackConfig = list()) {
    stopifnot(is.data.frame(tbl))

    if ("chrom" %in% colnames(tbl)) {
      colnames(tbl)[colnames(tbl) == "chrom"] <- "chr"
    }
    missingColumns <- setdiff(c("chr", "start", "end"), colnames(tbl))
    if (length(missingColumns) > 0) {
      fmt <- "improper columns in splice junction data.frame, missing: %s"
      stop(sprintf(fmt, toString(missingColumns)))
    }
    stopifnot(is(tbl$chr, "character"))
    stopifnot(is(tbl$start, "numeric"))
    stopifnot(is(tbl$end, "numeric"))

    if (deleteTracksOfSameName) {
      removeTracksByName(session, id, trackName)
    }

    state[["userAddedTracks"]] <-
      unique(c(state[["userAddedTracks"]], trackName))

    name <- if ("name" %in% colnames(tbl)) {
      as.character(tbl$name)
    } else {
      .packJunctionAttributes(tbl)
    }
    # igv.js reads the bed score as the uniquely mapped spanning read count and
    # sizes the arcs by it, so a uniquely_mapped column stands in for a score
    # column when the caller gives one and not the other. 1000 is the score the
    # bed decoder itself falls back to.
    score <- if ("score" %in% colnames(tbl)) {
      tbl$score
    } else if ("uniquely_mapped" %in% colnames(tbl)) {
      tbl$uniquely_mapped
    } else {
      1000
    }
    strand <- if ("strand" %in% colnames(tbl)) as.character(tbl$strand) else "."

    tbl.bed <- data.frame(
      chr = tbl$chr,
      start = tbl$start,
      end = tbl$end,
      name = name,
      score = score,
      strand = strand,
      stringsAsFactors = FALSE
    )

    bed.filePath <- .trackFile(session, ".bed")
    write.table(
      tbl.bed,
      file = bed.filePath,
      sep = "\t",
      quote = FALSE,
      row.names = FALSE,
      col.names = FALSE
    )
    lmsg <- sprintf(
      "--- igvShiny.R, loadSpliceJunctionTrackFromLocalData wrote %d to %s",
      NROW(tbl.bed), bed.filePath
    )
    flog.debug(lmsg)

    base.msg.to.igv <-
      list(
        elementID = id,
        trackName = trackName,
        url = file.path("tracks", basename(bed.filePath)),
        indexURL = "",
        trackHeight = trackHeight,
        displayMode = displayMode
      )

    msg.to.igv <- .sanitizeAndMergeOptions(base.msg.to.igv, trackConfig)
    # The url handler takes it from here: an unindexed bed served by shiny is
    # the same job as an unindexed bed served by anyone else.
    session$sendCustomMessage("loadSpliceJunctionTrackFromURL", msg.to.igv)

  } # loadSpliceJunctionTrackFromLocalData
#-------------------------------------------------------------------------------

library(testthat)
library(igvShiny)

# The track loaders are thin wrappers: they validate and reshape their inputs,
# then hand a payload to igv.js over session$sendCustomMessage(). These tests
# pin down that payload - the R/JS contract is exactly what broke in #36, #105
# and #116, and until now nothing checked it outside a real browser.

ELEMENT_ID <- "igvShiny_0"

bed_tbl <- function() {
  data.frame(
    chr = c("chr1", "chr1", "chr1"),
    start = c(7432, 7561, 7300),
    end = c(7503, 7580, 7350),
    value = c(0.5, 0.9, 0.1),
    stringsAsFactors = FALSE
  )
}

# igv.js sorts an alignment track from this object at load time. Each loader
# builds its own payload, so a dropped trackConfig is a per-loader regression.
sort_by_hp <- function() {
  list(chr = "chr1", position = 155160540, option = "TAG", tag = "HP",
       direction = "ASC")
}

test_that("loadBedTrack sends a bed file path and honours colour and height", {
  session <- fake_session()
  loadBedTrack(session, ELEMENT_ID, "bed", bed_tbl(),
               color = "blue", trackHeight = 77)

  msg <- last_message(session, "loadBedTrackFromFile")
  expect_equal(msg$elementID, ELEMENT_ID)
  expect_equal(msg$trackName, "bed")
  expect_equal(msg$color, "blue")
  expect_equal(msg$trackHeight, 77)
  expect_match(msg$bedFilepath, "^tracks/.*\\.bed$")
  expect_true(file.exists(file.path(get_tracks_dir(), basename(msg$bedFilepath))))
})

test_that("loadBedTrack deletes same-named tracks first, and can be told not to", {
  session <- fake_session()
  loadBedTrack(session, ELEMENT_ID, "bed", bed_tbl())
  expect_length(sent_messages(session, "removeTracksByName"), 1L)

  session <- fake_session()
  loadBedTrack(session, ELEMENT_ID, "bed", bed_tbl(),
               deleteTracksOfSameName = FALSE)
  expect_length(sent_messages(session, "removeTracksByName"), 0L)
})

test_that("loadBedTrack rejects a data.frame without chr/start/end", {
  session <- fake_session()
  tbl <- data.frame(a = "chr1", b = 1, c = 2, stringsAsFactors = FALSE)
  expect_error(loadBedTrack(session, ELEMENT_ID, "bad", tbl),
               "improper columns")
})

test_that("loadBedGraphTrack passes the autoscale settings through", {
  session <- fake_session()
  loadBedGraphTrack(session, ELEMENT_ID, "bedgraph", bed_tbl(),
                    autoscale = FALSE, min = 0, max = 10,
                    autoscaleGroup = 2)

  msg <- last_message(session, "loadBedGraphTrack")
  expect_false(msg$autoscale)
  expect_equal(msg$min, 0)
  expect_equal(msg$max, 10)
  expect_equal(msg$autoscaleGroup, 2)
})

test_that("loadBedGraphTrackFromURL forwards the url and autoscaleGroup (#105)", {
  session <- fake_session()
  url <- "https://example.org/data.bedGraph"
  loadBedGraphTrackFromURL(session, ELEMENT_ID, "remote", url,
                           autoscale = TRUE, autoscaleGroup = 1)

  msg <- last_message(session, "loadBedGraphTrackFromURL")
  expect_equal(msg$url, url)
  expect_true(msg$autoscale)
  expect_equal(msg$autoscaleGroup, 1)
})

test_that("loadSegTrack serialises the table as JSON", {
  session <- fake_session()
  tbl <- data.frame(
    chr = "chr1", start = 100, end = 200, value = 0.3,
    stringsAsFactors = FALSE
  )
  loadSegTrack(session, ELEMENT_ID, "seg", tbl)

  msg <- last_message(session, "loadSegTrack")
  expect_equal(msg$trackName, "seg")
  expect_match(as.character(msg$tbl), "chr1")
})

test_that("loadGwasTrack writes the table and sets the y-axis limits", {
  session <- fake_session()
  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))
  loadGwasTrack(session, ELEMENT_ID, "gwas", tbl.gwas, ymin = 1, ymax = 20)

  msg <- last_message(session, "loadGwasTrack")
  expect_match(msg$gwasDataFilepath, "^tracks/.*\\.gwas$")
  expect_equal(msg$min, 1)
  expect_equal(msg$max, 20)
  expect_false(msg$autoscale)
})

test_that("loadGwasTrack carries a column mapping and color table through", {
  session <- fake_session()
  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))
  columns <- list(chromosome = 3, position = 4, value = 10)
  colors <- list("19" = "purple", "*" = "gray")
  # the documented way for a loadGwasTrack caller to reach what GWASTrack takes
  # as arguments, so the allowlist and the loader have to agree
  loadGwasTrack(session, ELEMENT_ID, "gwas", tbl.gwas,
                trackConfig = list(columns = columns, colorTable = colors))

  msg <- last_message(session, "loadGwasTrack")
  expect_equal(msg$columns, columns)
  expect_equal(msg$colorTable, colors)
})

test_that("loadBamTrackFromURL sends both the bam and the index url", {
  session <- fake_session()
  loadBamTrackFromURL(session, ELEMENT_ID, "bam",
                      bamURL = "https://example.org/x.bam",
                      indexURL = "https://example.org/x.bam.bai",
                      displayMode = "SQUISHED", showAllBases = TRUE)

  msg <- last_message(session, "loadBamTrackFromURL")
  expect_equal(msg$bam, "https://example.org/x.bam")
  expect_equal(msg$index, "https://example.org/x.bam.bai")
  expect_equal(msg$displayMode, "SQUISHED")
  expect_true(msg$showAllBases)
})

test_that("loadCramTrackFromURL sends both the cram and the index url", {
  session <- fake_session()
  loadCramTrackFromURL(session, ELEMENT_ID, "cram",
                       cramURL = "https://example.org/x.cram",
                       indexURL = "https://example.org/x.cram.crai")

  msg <- last_message(session, "loadCramTrackFromURL")
  expect_equal(msg$cram, "https://example.org/x.cram")
  expect_equal(msg$index, "https://example.org/x.cram.crai")
})

test_that("loadGFF3TrackFromURL forwards the colour mapping", {
  session <- fake_session()
  colorTable <- list(protein_coding = "red", lncRNA = "blue")
  loadGFF3TrackFromURL(session, ELEMENT_ID, "gff3",
                       gff3URL = "https://example.org/x.gff3",
                       indexURL = "https://example.org/x.gff3.tbi",
                       colorTable = colorTable,
                       colorByAttribute = "biotype",
                       displayMode = "EXPANDED",
                       visibilityWindow = 1000000)

  msg <- last_message(session, "loadGFF3TrackFromURL")
  expect_equal(msg$dataURL, "https://example.org/x.gff3")
  expect_equal(msg$indexURL, "https://example.org/x.gff3.tbi")
  expect_equal(msg$colorByAttribute, "biotype")
  expect_equal(msg$colorTable$protein_coding, "red")
  expect_equal(msg$visibilityWindow, 1000000)
})

test_that("loadGFF3TrackFromLocalData writes the file into the tracks dir", {
  session <- fake_session()
  f <- system.file(package = "igvShiny", "extdata", "GRCh38.94.NDUFS2.gff3")
  tbl.gff3 <- read.table(f, sep = "\t", header = FALSE, comment.char = "#",
                         quote = "", nrows = 50, stringsAsFactors = FALSE)
  loadGFF3TrackFromLocalData(session, ELEMENT_ID, "local gff3", tbl.gff3,
                             colorTable = list(exon = "red"),
                             colorByAttribute = "biotype",
                             displayMode = "EXPANDED",
                             visibilityWindow = 5000)

  msg <- last_message(session, "loadGFF3TrackFromLocalData")
  expect_match(msg$filePath, "^tracks/.*\\.gff3$")
  expect_true(file.exists(file.path(get_tracks_dir(), basename(msg$filePath))))
})

test_that("loadVcfTrack writes a vcf and points igv.js at it", {
  skip_if_not_installed("VariantAnnotation")
  session <- fake_session()
  f <- system.file(package = "igvShiny", "extdata", "chr19-cebpaRegion.vcf.gz")
  vcf <- VariantAnnotation::readVcf(f)
  loadVcfTrack(session, ELEMENT_ID, "vcf", vcf)

  msg <- last_message(session, "loadVcfTrack")
  expect_match(msg$vcfDataFilepath, "^tracks/.*\\.vcf$")
})

test_that("loadBamTrackFromLocalData exports the alignments to the tracks dir", {
  skip_if_not_installed("GenomicAlignments")
  skip_if_not_installed("Rsamtools")
  skip_if_not_installed("rtracklayer")
  session <- fake_session()
  f <- system.file(package = "igvShiny", "extdata", "tumor.bam")
  ga <- GenomicAlignments::readGAlignments(f, use.names = TRUE)
  loadBamTrackFromLocalData(session, ELEMENT_ID, "local bam", ga,
                            trackConfig = list(sort = sort_by_hp()))

  msg <- last_message(session, "loadBamTrackFromLocalData")
  expect_match(msg$bamDataFilepath, "^tracks/.*\\.bam$")
  expect_equal(msg$sort, sort_by_hp())
})

test_that("loadCramTrackFromLocalData serves the cram and its index (#102)", {
  session <- fake_session()
  cram <- tempfile(fileext = ".cram")
  writeBin(charToRaw("CRAM"), cram)
  writeBin(charToRaw("CRAI"), paste0(cram, ".crai"))
  loadCramTrackFromLocalData(session, ELEMENT_ID, "local cram", cram,
                             trackConfig = list(sort = sort_by_hp()))

  msg <- last_message(session, "loadCramTrackFromURL")
  expect_match(msg$cram, "^tracks/.*\\.cram$")
  expect_match(msg$index, "^tracks/.*\\.crai$")
  # whether the file was linked or copied, the served paths must reach the
  # bytes - and the index must not be a second copy of the cram
  served <- function(path) file.path(get_tracks_dir(), basename(path))
  expect_equal(readBin(served(msg$cram), "raw", 4L), charToRaw("CRAM"))
  expect_equal(readBin(served(msg$index), "raw", 4L), charToRaw("CRAI"))
  expect_equal(msg$sort, sort_by_hp())
})

test_that("loadCramTrackFromLocalData rejects a file that is not there (#102)", {
  session <- fake_session()
  cram <- tempfile(fileext = ".cram")
  writeBin(charToRaw("CRAM"), cram)

  expect_error(
    loadCramTrackFromLocalData(session, ELEMENT_ID, "gone",
                               tempfile(fileext = ".cram")))
  # a cram without its index is the likelier mistake, and igv.js needs both
  expect_error(
    loadCramTrackFromLocalData(session, ELEMENT_ID, "unindexed", cram))
  expect_length(sent_messages(session, "loadCramTrackFromURL"), 0L)
})

test_that("the alignment loaders pass a sort-by-tag object through (#104)", {
  session <- fake_session()
  loadBamTrackFromURL(session, ELEMENT_ID, "bam",
                      "https://example.org/a.bam", "https://example.org/a.bai",
                      trackConfig = list(sort = sort_by_hp()))
  expect_equal(last_message(session, "loadBamTrackFromURL")$sort, sort_by_hp())

  session <- fake_session()
  loadCramTrackFromURL(session, ELEMENT_ID, "cram",
                       "https://example.org/a.cram",
                       "https://example.org/a.crai",
                       trackConfig = list(sort = sort_by_hp()))
  expect_equal(last_message(session, "loadCramTrackFromURL")$sort, sort_by_hp())
})

test_that("loadSpliceJunctionTrackFromURL sends the url and the junction knobs (#103)", {
  session <- fake_session()
  loadSpliceJunctionTrackFromURL(
    session, ELEMENT_ID, "junctions",
    "https://example.org/sampleA.SJ.out.bed",
    trackHeight = 140,
    trackConfig = list(minUniquelyMappedReads = 2,
                       thicknessBasedOn = "numReads",
                       colorBy = "isAnnotatedJunction",
                       labelWith = "totalReadCount",
                       hideMotifs = list("non-canonical"))
  )

  msg <- last_message(session, "loadSpliceJunctionTrackFromURL")
  expect_equal(msg$elementID, ELEMENT_ID)
  expect_equal(msg$url, "https://example.org/sampleA.SJ.out.bed")
  expect_equal(msg$trackHeight, 140)
  expect_equal(msg$displayMode, "COLLAPSED")
  # an empty indexURL is what tells the JS side to set indexed = false; igv.js
  # otherwise guesses url + ".tbi"
  expect_equal(msg$indexURL, "")
  expect_equal(msg$minUniquelyMappedReads, 2)
  expect_equal(msg$thicknessBasedOn, "numReads")
  expect_equal(msg$colorBy, "isAnnotatedJunction")
  expect_equal(msg$labelWith, "totalReadCount")
  expect_equal(msg$hideMotifs, list("non-canonical"))
})

test_that("loadSpliceJunctionTrackFromURL carries a tabix index when given one", {
  session <- fake_session()
  loadSpliceJunctionTrackFromURL(
    session, ELEMENT_ID, "junctions",
    "https://example.org/sampleA.SJ.out.bed.gz",
    indexURL = "https://example.org/sampleA.SJ.out.bed.gz.tbi",
    deleteTracksOfSameName = FALSE
  )

  msg <- last_message(session, "loadSpliceJunctionTrackFromURL")
  expect_equal(msg$indexURL, "https://example.org/sampleA.SJ.out.bed.gz.tbi")
  expect_length(sent_messages(session, "removeTracksByName"), 0L)
})

test_that("loadSpliceJunctionTrackFromURL treats an unusable indexURL as no index", {
  # NULL survives list() and reaches the browser as JSON null, where the
  # handler's message.indexURL.length throws and the track silently never
  # loads. Normalising in R keeps that off the wire.
  session <- fake_session()
  loadSpliceJunctionTrackFromURL(session, ELEMENT_ID, "j",
                                 "https://example.org/a.bed", indexURL = NULL)
  expect_equal(last_message(session, "loadSpliceJunctionTrackFromURL")$indexURL, "")

  session <- fake_session()
  expect_warning(
    loadSpliceJunctionTrackFromURL(session, ELEMENT_ID, "j",
                                   "https://example.org/a.bed",
                                   indexURL = c("a.tbi", "b.tbi")),
    "single character string"
  )
  expect_equal(last_message(session, "loadSpliceJunctionTrackFromURL")$indexURL, "")

  session <- fake_session()
  expect_warning(
    loadSpliceJunctionTrackFromURL(session, ELEMENT_ID, "j",
                                   "https://example.org/a.bed", indexURL = NA),
    "single character string"
  )
  expect_equal(last_message(session, "loadSpliceJunctionTrackFromURL")$indexURL, "")
})

test_that("the label options from the bundled example are not igv.js 3.x options", {
  # spliceJunctionTrack.html still sets labelUniqueReadCount and friends; the
  # library dropped them for labelWith/labelWithInParen, so allowlisting them
  # would have promised users something that draws nothing
  session <- fake_session()
  expect_warning(
    loadSpliceJunctionTrackFromURL(
      session, ELEMENT_ID, "junctions", "https://example.org/a.bed",
      trackConfig = list(labelUniqueReadCount = TRUE)
    ),
    "invalid or unsupported"
  )
  expect_null(last_message(session, "loadSpliceJunctionTrackFromURL")$labelUniqueReadCount)
})

junction_tbl <- function() {
  data.frame(
    chrom = c("U13369.1", "U13369.1"),
    start = c(7276, 7276),
    end = c(8225, 9540),
    motif = c("GT/AG", "CT/AC"),
    uniquely_mapped = c(117, 3),
    multi_mapped = c(6, 9),
    maximum_spliced_alignment_overhang = c(37, 12),
    annotated_junction = c(TRUE, FALSE),
    strand = c("+", "-"),
    stringsAsFactors = FALSE
  )
}

read_staged_bed <- function(msg) {
  path <- file.path(get_tracks_dir(), basename(msg$url))
  expect_true(file.exists(path))
  read.table(path, sep = "\t", header = FALSE, quote = "",
             stringsAsFactors = FALSE)
}

test_that("loadSpliceJunctionTrackFromLocalData serves a bed through the url handler (#103)", {
  session <- fake_session()
  loadSpliceJunctionTrackFromLocalData(session, ELEMENT_ID, "junctions",
                                       junction_tbl(), trackHeight = 140)

  msg <- last_message(session, "loadSpliceJunctionTrackFromURL")
  expect_equal(msg$elementID, ELEMENT_ID)
  expect_equal(msg$trackName, "junctions")
  expect_match(msg$url, "^tracks/.*\\.bed$")
  expect_equal(msg$indexURL, "")
  expect_equal(msg$trackHeight, 140)

  bed <- read_staged_bed(msg)
  expect_equal(NROW(bed), 2L)
  expect_equal(bed[[1]], c("U13369.1", "U13369.1"))
  expect_equal(bed[[2]], c(7276, 7276))
  expect_equal(bed[[6]], c("+", "-"))
})

test_that("the junction attributes reach the bed name column igv.js parses (#103)", {
  session <- fake_session()
  loadSpliceJunctionTrackFromLocalData(session, ELEMENT_ID, "j",
                                       junction_tbl())
  bed <- read_staged_bed(last_message(session, "loadSpliceJunctionTrackFromURL"))

  expect_equal(
    bed[[4]][1],
    paste0("motif=GT/AG;uniquely_mapped=117;multi_mapped=6;",
           "maximum_spliced_alignment_overhang=37;annotated_junction=true")
  )
  # igv.js compares annotated_junction against "true"/"false"; R's own TRUE
  # would match neither and both hide options would let the junction through
  expect_match(bed[[4]][2], "annotated_junction=false", fixed = TRUE)
  # score stands in for the uniquely mapped count, which is what sizes the arcs
  expect_equal(bed[[5]], c(117, 3))
})

test_that("a single junction attribute still parses as an attribute (#103)", {
  # the bed decoder reads column 4 as attributes only when it holds both "="
  # and ";", so one pair on its own would arrive as a plain label
  session <- fake_session()
  tbl <- junction_tbl()[, c("chrom", "start", "end", "motif")]
  loadSpliceJunctionTrackFromLocalData(session, ELEMENT_ID, "j", tbl)
  bed <- read_staged_bed(last_message(session, "loadSpliceJunctionTrackFromURL"))

  expect_equal(bed[[4]], c("motif=GT/AG;", "motif=CT/AC;"))
  expect_equal(bed[[5]], c(1000, 1000))
  expect_equal(bed[[6]], c(".", "."))
})

test_that("a packed name column is written through untouched (#103)", {
  session <- fake_session()
  tbl <- data.frame(chrom = "chr1", start = 100, end = 200,
                    name = "motif=GT/AG;uniquely_mapped=9", score = 9,
                    stringsAsFactors = FALSE)
  loadSpliceJunctionTrackFromLocalData(session, ELEMENT_ID, "j", tbl)
  bed <- read_staged_bed(last_message(session, "loadSpliceJunctionTrackFromURL"))

  expect_equal(bed[[4]], "motif=GT/AG;uniquely_mapped=9")
  expect_equal(bed[[5]], 9)
})

test_that("loadSpliceJunctionTrackFromLocalData rejects a table without chrom/start/end", {
  session <- fake_session()
  tbl <- data.frame(a = "chr1", b = 1, c = 2, stringsAsFactors = FALSE)
  expect_error(loadSpliceJunctionTrackFromLocalData(session, ELEMENT_ID, "j", tbl),
               "improper columns")
})

test_that("the navigation and removal helpers send their own messages", {
  session <- fake_session()
  showGenomicRegion(session, ELEMENT_ID, "chr1:1-1000")
  expect_equal(last_message(session, "showGenomicRegion")$region, "chr1:1-1000")

  getGenomicRegion(session, ELEMENT_ID)
  expect_equal(last_message(session, "getGenomicRegion")$elementID, ELEMENT_ID)

  removeTracksByName(session, ELEMENT_ID, c("a", "b"))
  expect_equal(last_message(session, "removeTracksByName")$trackNames, c("a", "b"))

  removeUserAddedTracks(session, ELEMENT_ID)
  expect_gt(length(sent_messages(session, "removeTracksByName")), 1L)
})

library(testthat)
library(igvShiny)

# The widget constructor itself: building an htmlwidget needs no browser, so the
# genome plumbing, the startup `tracks` sanitisation (#36) and the Shiny
# output/render wrappers can all be checked here.
#
# igvShiny() reads the module namespace off the current reactive domain
# (`shiny::getDefaultReactiveDomain()$ns("")`); a MockShinySession provides one
# without starting an app. Outside a session the namespace is "" (#128).

with_domain <- function(code) {
  shiny::withReactiveDomain(shiny::MockShinySession$new(), code)
}

test_that("igvShiny builds a widget for a stock genome", {
  opts <- parseAndValidateGenomeSpec(genomeName = "hg38",
                                     initialLocus = "MEF2C")
  widget <- with_domain(igvShiny(opts))

  expect_s3_class(widget, "htmlwidget")
  expect_equal(widget$x$genomeName, "hg38")
  expect_equal(widget$x$initialLocus, "MEF2C")
})

test_that("igvShiny sanitises the startup tracks and keeps the valid ones (#36)", {
  opts <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "all")
  tracks <- list(
    list(name = "genes", type = "annotation", format = "gff3",
         url = "https://example.org/genes.gff3"),
    list(name = "dropped", type = "annotation")          # no url
  )
  expect_warning(widget <- with_domain(igvShiny(opts, tracks = tracks)),
                 "no valid 'url'")

  expect_length(widget$x$tracks, 1L)
  expect_equal(widget$x$tracks[[1]]$name, "genes")
})

test_that("igvShiny defaults to no startup tracks", {
  opts <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "all")
  widget <- with_domain(igvShiny(opts))
  expect_length(widget$x$tracks, 0L)
})

test_that("igvShiny refuses a genome spec that was not validated", {
  opts <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "all")
  opts$validated <- FALSE
  expect_error(with_domain(igvShiny(opts)))

  expect_error(with_domain(igvShiny(list(genomeName = "hg38"))))
})

test_that("igvShiny copies local genome files into the tracks directory", {
  data.dir <- system.file(package = "igvShiny", "extdata")
  opts <- parseAndValidateGenomeSpec(
    genomeName = "ribo",
    initialLocus = "all",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = file.path(data.dir, "ribosomal-RNA-gene.fasta"),
    fastaIndex = file.path(data.dir, "ribosomal-RNA-gene.fasta.fai"),
    genomeAnnotation = file.path(data.dir, "ribosomal-RNA-gene.gff3")
  )
  widget <- with_domain(igvShiny(opts))

  # the paths handed to igv.js are rewritten to the served "tracks" directory
  expect_match(widget$x$fasta, "^tracks/")
  expect_match(widget$x$fastaIndex, "^tracks/")
  expect_true(file.exists(file.path(get_tracks_dir(),
                                    basename(widget$x$fasta))))
})

test_that("igvShiny builds outside a reactive domain (#128)", {
  expect_null(shiny::getDefaultReactiveDomain())

  opts <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "all")
  widget <- igvShiny(opts)

  expect_s3_class(widget, "htmlwidget")
  # the JS side concatenates moduleNS with the event name, so no module means ""
  expect_identical(widget$x$moduleNS, "")
})

test_that("igvShinyOutput and renderIgvShiny return Shiny bindings", {
  out <- igvShinyOutput("igv")
  expect_s3_class(out, "shiny.tag.list")

  opts <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "all")
  renderer <- renderIgvShiny(with_domain(igvShiny(opts)))
  expect_type(renderer, "closure")
})

library(testthat)
library(shinytest2)
library(igvShiny)

# Helper function for clicking a UI element and checking the resulting HTML
.click_and_check <- function(app, button_id, expected_html_label, selector = "#igvShiny_0", timeout = 30) {
    app$click(button_id)
    # igv.js 3.x renders the browser inside a shadow root (attachShadow), so the
    # track DOM is not part of the element's light-DOM HTML and get_html() cannot
    # see it; read the shadow root's innerHTML instead. Poll rather than sleep a
    # fixed time: some tracks (remote bigWig, GFF3 with per-feature colouring)
    # render their label several seconds after the click.
    js <- sprintf("document.querySelector('%s').shadowRoot.innerHTML", selector)
    deadline <- Sys.time() + timeout
    found <- FALSE
    repeat {
        if (grepl(expected_html_label, app$get_js(js), fixed = TRUE)) {
            found <- TRUE
            break
        }
        if (Sys.time() > deadline) {
            break
        }
        Sys.sleep(0.5)
    }
    expect_true(found)
}

test_that("igvShinyDemo loads tracks correctly", {
    # Increase timeout for potentially slow-loading remote resources
    options(chromote.timeout = 120)

    sf <- system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
    app <- AppDriver$new(
        app_dir = shiny::shinyAppFile(sf),
        name = "igv-shiny-demo-app",
        height = 695,
        width = 1235,
        load_timeout = 1e+6,
        timeout = 1e+6
    )

    # Wait for the app and genome to fully load
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    .click_and_check(app, "addBedGraphTrackButton", 'title="wig/bedGraph/local"')
    # the remote ENCODE bigWig has a test of its own below, see the note there
    .click_and_check(app, "addBamViaHttpButton", 'title="NA19240.bam"')
    .click_and_check(app, "addCramViaHttpButton", 'title="CRAM"')

    app$stop()
})

test_that("igvShinyDemo loads a remote bigWig", {
    # igv.js reports "Error accessing resource: ...ENCFF356YES.bigWig status: 0"
    # for this track - a network/CORS refusal by the browser, not a slow render
    # and not a dead host (curl gets a 307 to S3 with permissive CORS). Skipped
    # first, so no app is started while the fetch cannot succeed; drop the skip
    # to re-check whether the block is gone.
    skip("remote ENCODE bigWig fetch is refused by the browser (status 0)")

    options(chromote.timeout = 120)
    sf <- system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
    app <- AppDriver$new(
        app_dir = shiny::shinyAppFile(sf),
        name = "igv-shiny-demo-app-bigwig",
        height = 695,
        width = 1235,
        load_timeout = 1e+6,
        timeout = 1e+6
    )
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    .click_and_check(app, "addBedGraphTrackFromURLButton", 'title="bedGraph/remote"')

    app$stop()
})

test_that("the gff3 demo loads its GFF3 tracks correctly", {
    options(chromote.timeout = 120)

    sf <- system.file(package = "igvShiny", "demos", "gff3.R")
    app <- AppDriver$new(
      app_dir = shiny::shinyAppFile(sf),
      name = "test_app",
      height = 695,
      width = 1235,
      load_timeout = 1e+6,
      timeout = 1e+6
    )
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    .click_and_check(app, "addRemoteGFF3TrackButton", 'title="url gff3"')
    .click_and_check(app, "addRemoteGFF3TrackButtonWithBiotypeColors", 'title="url gff3 (colors)"')
    # this one used to 404: the file was written to a moved tracks directory
    # while shiny kept serving the one registered at load time. .tracksDir()
    # re-points the resource path before each write, so it renders again.
    .click_and_check(app, "addLocalGFF3TrackButtonWithBiotypeColors", 'title="local gff3 (colors)"')

    app$stop()
})

test_that("the junctions demo renders junctions held in a data.frame", {
    options(chromote.timeout = 120)

    sf <- system.file(package = "igvShiny", "demos", "junctions.R")
    app <- AppDriver$new(
      app_dir = shiny::shinyAppFile(sf),
      name = "test_app",
      height = 695,
      width = 1235,
      load_timeout = 1e+6,
      timeout = 1e+6
    )
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    # written as bed, served from the tracks directory, no network in the
    # path at all (#103)
    .click_and_check(app, "addLocalJunctionsButton", 'title="sampleA junctions"')

    app$stop()
})

test_that("two widgets at the same locus both report it", {
    # #126: the handler compared the new locus against window.chromLocString,
    # one global for the whole page, so whichever widget fired first claimed the
    # value and the other one's currentGenomicRegion event was dropped as a
    # duplicate. Each widget now compares against its own last locus.
    options(chromote.timeout = 120)

    port <- local_server()
    app_file <- tempfile(fileext = ".R")
    writeLines(sprintf('
        library(shiny)
        library(igvShiny)

        riboSpec <- function()
            parseAndValidateGenomeSpec(
                genomeName = "ribo",
                initialLocus = "all",
                stockGenome = FALSE,
                dataMode = "http",
                fasta = "%s",
                fastaIndex = "%s",
                genomeAnnotation = "%s"
            )

        ui <- fluidPage(
            actionButton("moveBothButton", "Move both"),
            verbatimTextOutput("regionOne"),
            verbatimTextOutput("regionTwo"),
            igvShinyOutput("igvShiny_0"),
            igvShinyOutput("igvShiny_1")
        )

        server <- function(input, output, session) {
            output$igvShiny_0 <- renderIgvShiny(igvShiny(riboSpec()))
            output$igvShiny_1 <- renderIgvShiny(igvShiny(riboSpec()))

            observeEvent(input$moveBothButton, {
                showGenomicRegion(session, "igvShiny_0", "U13369.1:7,276-8,225")
                showGenomicRegion(session, "igvShiny_1", "U13369.1:7,276-8,225")
            })
            observeEvent(input[["currentGenomicRegion.igvShiny_0"]], {
                output$regionOne <- renderText({
                    input[["currentGenomicRegion.igvShiny_0"]]
                })
            })
            observeEvent(input[["currentGenomicRegion.igvShiny_1"]], {
                output$regionTwo <- renderText({
                    input[["currentGenomicRegion.igvShiny_1"]]
                })
            })
        }

        shinyApp(ui = ui, server = server)',
        local_url(port, "ribosomal-RNA-gene.fasta"),
        local_url(port, "ribosomal-RNA-gene.fasta.fai"),
        local_url(port, "ribosomal-RNA-gene.gff3")
    ), app_file)

    app <- AppDriver$new(
        app_dir = shiny::shinyAppFile(app_file),
        name = "igv-shiny-two-widgets-region",
        height = 695,
        width = 1235,
        load_timeout = 1e+6,
        timeout = 1e+6
    )
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    app$click("moveBothButton")

    # both widgets move to the same region, so on the old JS the second one is
    # silenced; poll until both outputs are filled or the deadline passes.
    # An output whose renderText has not run yet reads back as NULL.
    .region <- function(name) {
        value <- app$get_value(output = name)
        if (is.null(value)) "" else value
    }

    deadline <- Sys.time() + 30
    regions <- c("", "")
    repeat {
        regions <- c(.region("regionOne"), .region("regionTwo"))
        if (all(nzchar(regions))) {
            break
        }
        if (Sys.time() > deadline) {
            break
        }
        Sys.sleep(0.5)
    }

    # each widget reports the chromosome it was sent to; the end is left out of
    # the assertion because igv.js fits the range to the viewport
    expect_true(nzchar(regions[1]))
    expect_true(nzchar(regions[2]))
    expect_true(all(startsWith(regions, "U13369.1:")))

    # the reported start matches the one sent, rather than trailing it by a
    # base: the frame start is 0-based and showGenomicRegion reads 1-based, so
    # emitting it raw made every round trip creep one to the left (#126)
    starts <- as.integer(sub("^[^:]+:([0-9]+)-.*$", "\\1", regions))
    expect_equal(starts, c(7276L, 7276L))

    app$stop()
})

test_that("a splice junction track renders from a url (#103)", {
    # #103 asked for sashimi plots. igv.js has no such feature - it has a
    # junction track drawn from a bed file - so what has to be proven here is
    # that the type string, the bed and the unindexed flag reach igv.js and
    # produce a track. The bed is our own fixture on 127.0.0.1: the upstream
    # example files sat on s3.amazonaws.com/igv.org.demo and 404 today.
    options(chromote.timeout = 120)

    port <- local_server()
    app_file <- tempfile(fileext = ".R")
    writeLines(sprintf('
        library(shiny)
        library(igvShiny)

        ui <- fluidPage(
            actionButton("addJunctionsButton", "Add junctions"),
            igvShinyOutput("igvShiny_0")
        )

        server <- function(input, output, session) {
            output$igvShiny_0 <- renderIgvShiny({
                igvShiny(parseAndValidateGenomeSpec(
                    genomeName = "ribo",
                    initialLocus = "U13369.1:6,000-13,000",
                    stockGenome = FALSE,
                    dataMode = "http",
                    fasta = "%s",
                    fastaIndex = "%s",
                    genomeAnnotation = "%s"
                ))
            })
            observeEvent(input$addJunctionsButton, {
                loadSpliceJunctionTrackFromURL(
                    session,
                    id = "igvShiny_0",
                    trackName = "junctions",
                    url = "%s",
                    trackConfig = list(minUniquelyMappedReads = 1,
                                       colorBy = "isAnnotatedJunction")
                )
            })
        }

        shinyApp(ui = ui, server = server)',
        local_url(port, "ribosomal-RNA-gene.fasta"),
        local_url(port, "ribosomal-RNA-gene.fasta.fai"),
        local_url(port, "ribosomal-RNA-gene.gff3"),
        local_url(port, "ribosomal-RNA-gene.SJ.out.bed")
    ), app_file)

    app <- AppDriver$new(
        app_dir = shiny::shinyAppFile(app_file),
        name = "igv-shiny-splice-junctions",
        height = 695,
        width = 1235,
        load_timeout = 1e+6,
        timeout = 1e+6
    )
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    .click_and_check(app, "addJunctionsButton", 'title="junctions"')

    app$stop()
})

test_that("getGenomicRegion replies to a module named anything but 'igv'", {
    # #134: the reply event name was built as
    # "igv-currentGenomicRegion." + elementID.replace("igv-", ""), so only a
    # module whose id happened to be "igv" (as every demo uses) received it.
    # The app below names the module "browser" and fails on the old JS.
    options(chromote.timeout = 120)

    port <- local_server()
    app_file <- tempfile(fileext = ".R")
    writeLines(sprintf('
        library(shiny)
        library(igvShiny)

        igvUI <- function(id) {
            ns <- NS(id)
            fluidPage(
                actionButton(ns("getRegionButton"), "Get Region"),
                verbatimTextOutput(ns("regionDisplay")),
                igvShinyOutput(ns("igvShiny_0"))
            )
        }

        igvServer <- function(id) {
            moduleServer(id, function(input, output, session) {
                ns <- session$ns
                output$igvShiny_0 <- renderIgvShiny({
                    igvShiny(parseAndValidateGenomeSpec(
                        genomeName = "ribo",
                        initialLocus = "all",
                        stockGenome = FALSE,
                        dataMode = "http",
                        fasta = "%s",
                        fastaIndex = "%s",
                        genomeAnnotation = "%s"
                    ))
                })
                observeEvent(input$getRegionButton, {
                    getGenomicRegion(session, id = ns("igvShiny_0"))
                })
                observeEvent(input[["currentGenomicRegion.igvShiny_0"]], {
                    output$regionDisplay <- renderText({
                        input[["currentGenomicRegion.igvShiny_0"]]
                    })
                })
            })
        }

        shinyApp(
            ui = fluidPage(igvUI("browser")),
            server = function(input, output, session) igvServer("browser")
        )',
        local_url(port, "ribosomal-RNA-gene.fasta"),
        local_url(port, "ribosomal-RNA-gene.fasta.fai"),
        local_url(port, "ribosomal-RNA-gene.gff3")
    ), app_file)

    app <- AppDriver$new(
        app_dir = shiny::shinyAppFile(app_file),
        name = "igv-shiny-module-region",
        height = 695,
        width = 1235,
        load_timeout = 1e+6,
        timeout = 1e+6
    )
    app$wait_for_value(input = "igvReady")
    Sys.sleep(2)

    app$click("browser-getRegionButton")

    # the reply travels browser -> shiny as an input event, so poll rather than
    # read once; an empty display after the deadline is the #134 failure
    deadline <- Sys.time() + 30
    region <- ""
    repeat {
        region <- app$get_value(output = "browser-regionDisplay")
        if (!is.null(region) && nzchar(region)) {
            break
        }
        if (Sys.time() > deadline) {
            break
        }
        Sys.sleep(0.5)
    }
    expect_equal(region, "all")

    app$stop()
})

# igvShiny — GFF3 annotation tracks, from a local table and from a URL.
# colorTable and colorByAttribute have no defaults: omitting them fails inside
# the observer, where nothing surfaces the error to the browser.

library(shiny)
library(bslib)
library(igvShiny)

gff3.colors <- list(processed_transcript = "blue", protein_coding = "darkgreen",
                    retained_intron = "brown", nonsense_mediated_decay = "orange",
                    miRNA = "darkred", default = "black")
gff3.url <- "https://s3.amazonaws.com/igv.org.genomes/hg38/Homo_sapiens.GRCh38.94.chr.gff3.gz"

ui <- page_sidebar(
  title = "igvShiny — GFF3 tracks",
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  fillable = TRUE,
  sidebar = sidebar(
    width = 280,
    actionButton("addLocalGFF3TrackButtonWithBiotypeColors", "GFF3 local (colors)",
                 class = "w-100 mb-2"),
    actionButton("addRemoteGFF3TrackButton", "GFF3 from URL", class = "w-100 mb-2"),
    actionButton("addRemoteGFF3TrackButtonWithBiotypeColors", "GFF3 from URL (colors)",
                 class = "w-100 mb-2")
  ),
  card(full_screen = TRUE,
       card_body(class = "p-0", igvShinyOutput("igvShiny_0", height = "100%")))
)

server <- function(input, output, session) {
  observeEvent(input$addLocalGFF3TrackButtonWithBiotypeColors, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:161,172,691-161,241,018")
    f <- system.file(package = "igvShiny", "extdata", "GRCh38.94.NDUFS2.gff3")
    loadGFF3TrackFromLocalData(session, id = "igvShiny_0", trackName = "local gff3 (colors)",
                               tbl.gff3 = read.table(f, sep = "\t", as.is = TRUE, header = TRUE),
                               color = "brown", colorTable = gff3.colors, colorByAttribute = "biotype",
                               displayMode = "EXPANDED", trackHeight = 200, visibilityWindow = 80000)
  })
  observeEvent(input$addRemoteGFF3TrackButton, {
    showGenomicRegion(session, id = "igvShiny_0", "NDUFS2")
    loadGFF3TrackFromURL(session, id = "igvShiny_0", trackName = "url gff3",
                         gff3URL = gff3.url, indexURL = sprintf("%s.tbi", gff3.url),
                         color = "brown", colorTable = list(), colorByAttribute = NA_character_,
                         displayMode = "EXPANDED", trackHeight = 200, visibilityWindow = 50000)
  })
  observeEvent(input$addRemoteGFF3TrackButtonWithBiotypeColors, {
    showGenomicRegion(session, id = "igvShiny_0", "chr1:161,172,691-161,241,018")
    loadGFF3TrackFromURL(session, id = "igvShiny_0", trackName = "url gff3 (colors)",
                         gff3URL = gff3.url, indexURL = sprintf("%s.tbi", gff3.url),
                         color = "brown", colorTable = gff3.colors, colorByAttribute = "biotype",
                         displayMode = "EXPANDED", trackHeight = 200, visibilityWindow = 80000)
  })

  output$igvShiny_0 <- renderIgvShiny({
    igvShiny(parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "NDUFS2"))
  })
}

shinyApp(ui = ui, server = server)

library(shiny)
library(igvShiny)
library(htmlwidgets)
# library(biomaRt)

#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
# print(dim(tbl.gwas))
printf <- function(...) print(noquote(sprintf(...)))
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(

  sidebarLayout(
     sidebarPanel(
        textInput("roi", label="", placeholder="gene or chrN:start-end"),
        actionButton("searchButton", "Search"),
        br(), br(), br(),
        #actionButton("addLocalVCFTrackButton", "Add 1kg VCF (local file)"),
        actionButton("addRemoteGFF3TrackButton", "Add Remote GFF3 Track (AWS)"),
        br(),
        actionButton("addRemoteGFF3TrackButtonWithBiotypeColors", "GFF3 Track (AWS) - colors"),
        br(),
        actionButton("addLocalGFF3TrackButton", "GFF3 Track (local)"),
        br(),
        actionButton("addLocalGFF3TrackButtonWithBiotypeColors", "GFF3 Track (local) - colors"),

        actionButton("removeUserTracksButton", "Remove User Tracks"),
        br(), br(),
        actionButton("getChromLocButton", "Get Region"),
        actionButton("clearChromLocButton", "Clear Region Readout"),
        div(style="background-color: white; width: 200px; height:30px; padding-left: 5px;
                   margin-top: 10px; border: 1px solid blue;",
            htmlOutput("chromLocDisplay")),
        hr(),
        width=2
        ),
     mainPanel(
        igvShinyOutput('igvShiny_0'),
        # igvShinyOutput('igvShiny_1'),
        width=10
        )
     ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {

   observeEvent(input$searchButton, {
      searchString = isolate(input$roi)
      printf("--- search: %s", searchString)
      if(nchar(searchString) > 0)
        showGenomicRegion(session, id="igvShiny_0", searchString)
      })

    observeEvent(input$addRemoteGFF3TrackButton, {
       showGenomicRegion(session, id="igvShiny_0", "NDUFS2")
       url <- "https://s3.amazonaws.com/igv.org.genomes/hg38/Homo_sapiens.GRCh38.94.chr.gff3.gz"
       indexURL <- "https://s3.amazonaws.com/igv.org.genomes/hg38/Homo_sapiens.GRCh38.94.chr.gff3.gz.tbi"
       printf("--- about to call loadGFF3TrackFromURL")
       loadGFF3TrackFromURL(session,
                            id="igvShiny_0",
                            trackName="url gff3",
                            gff3URL=url,
                            indexURL=indexURL,
                            color="brown",
                            colorTable=list(),
                            colorByAttribute=NA_character_,
                            displayMode="EXPANDED",
                            trackHeight=200,
                            visibilityWindow=50000,
                            deleteTracksOfSameName=TRUE)
       })


    observeEvent(input$addRemoteGFF3TrackButtonWithBiotypeColors, {
       showGenomicRegion(session, id="igvShiny_0", "chr1:161,172,691-161,241,018")
       url <- "https://s3.amazonaws.com/igv.org.genomes/hg38/Homo_sapiens.GRCh38.94.chr.gff3.gz"
       indexURL <- "https://s3.amazonaws.com/igv.org.genomes/hg38/Homo_sapiens.GRCh38.94.chr.gff3.gz.tbi"
       printf("--- about to call loadGFF3TrackFromURL")
       color.table <- list(processed_transcript="blue",
                           protein_coding="darkgreen",
                           retained_intron="brown",
                           nonsense_mediated_decay="orange",
                           miRNA="darkred",
                           default="black")

       colorByAttribute <- "biotype"
       loadGFF3TrackFromURL(session,
                            id="igvShiny_0",
                            trackName="url gff3 (colors)",
                            gff3URL=url,
                            indexURL=indexURL,
                            color="brown",
                            colorTable=color.table,
                            colorByAttribute=colorByAttribute,
                            displayMode="EXPANDED",
                            trackHeight=200,
                            visibilityWindow=80000,
                            deleteTracksOfSameName=TRUE)
       })


    observeEvent(input$addLocalGFF3TrackButton, {
       showGenomicRegion(session, id="igvShiny_0", "chr1:161,172,691-161,241,018")
       full.path <- system.file(package="igvShiny", "extdata", "GRCh38.94.NDUFS2.gff3")
       tbl.gff3 <- read.table(full.path, sep="\t", as.is=TRUE, header=TRUE)
       printf("--- about to call loadGFF3rackFromLocalData, dim: %d, %d", nrow(tbl.gff3), ncol(tbl.gff3))

       loadGFF3TrackFromLocalData(session,
                                  id="igvShiny_0",
                                  trackName="local gff3",
                                  tbl.gff3=tbl.gff3,
                                  colorTable=list(),
                                  colorByAttribute=NA_character_,
                                  color="magenta",
                                  displayMode="EXPANDED",
                                  trackHeight=200,
                                  visibilityWindow=80000,
                                  deleteTracksOfSameName=TRUE)
       }) # addLocalGFF3TrackButtonWithBiotypeColors


    observeEvent(input$addLocalGFF3TrackButtonWithBiotypeColors, {
       showGenomicRegion(session, id="igvShiny_0", "chr1:161,172,691-161,241,018")
       full.path <- system.file(package="igvShiny", "extdata", "GRCh38.94.NDUFS2.gff3")
       tbl.gff3 <- read.table(full.path, sep="\t", as.is=TRUE, header=TRUE)
       printf("--- about to call loadGFF3rackFromLocalData, dim: %d, %d", nrow(tbl.gff3), ncol(tbl.gff3))
       color.table <- list(processed_transcript="blue",
                           protein_coding="darkgreen",
                           retained_intron="brown",
                           nonsense_mediated_decay="orange",
                           miRNA="darkred",
                           default="black")

       colorByAttribute <- "biotype"
       loadGFF3TrackFromLocalData(session,
                                  id="igvShiny_0",
                                  trackName="local gff3 (colors)",
                                  tbl.gff3=tbl.gff3,
                                  color="brown",
                                  colorTable=color.table,
                                  colorByAttribute=colorByAttribute,
                                  displayMode="EXPANDED",
                                  trackHeight=200,
                                  visibilityWindow=80000,
                                  deleteTracksOfSameName=TRUE)
       }) # addLocalGFF3TrackButtonWithBiotypeColors


   observeEvent(input$removeUserTracksButton, {
      printf("---- removeUserTracks")
      removeUserAddedTracks(session, id="igvShiny_0")
      })


   observeEvent(input[["igv-trackClick"]], {
       printf("--- igv-trackClick event")
       x <- input[["igv-trackClick"]]
       attribute.name.positions <- grep("name", names(x))
       attribute.value.positions <- grep("value", names(x))
       attribute.names <- as.character(x)[attribute.name.positions]
       attribute.values <- as.character(x)[attribute.value.positions]
       tbl <- data.frame(name=attribute.names,
                         value=attribute.values,
                         stringsAsFactors=FALSE)
       dialogContent <- renderTable(tbl)
       html <- HTML(dialogContent())
       showModal(modalDialog(html, easyClose=TRUE))
       })

   observeEvent(input$getChromLocButton, {
      # printf("--- getChromLoc event")
      # sends message to igv.js in browser; currentGenomicRegion.<id> event sent back
      # see below for how that can be captured and displayed
      getGenomicRegion(session, id="igvShiny_0")
      })

   observeEvent(input$clearChromLocButton, {
      output$chromLocDisplay <- renderText({" "})
      })

   observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
      newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
      #printf("new chromLocString: %s", newLoc)
      output$chromLocDisplay <- renderText({newLoc})
      })

   genomes <- c("hg38", "hg19", "mm10", "tair10", "rhos")
   loci <- c("chr5:88,466,402-89,135,305", "MEF2C", "Mef2c", "1:7,432,931-7,440,395", "NC_007494.2:370,757-378,078")
   i <- 2

   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <-
       parseAndValidateGenomeSpec(
         genomeName = "hg38",
         initialLocus = loci[2]
       )
     igvShiny(genomeOptions)
     })

} # server
#----------------------------------------------------------------------------------------------------
deploy <-function()
{
   require(rsconnect)
   #rsconnect::setAccountInfo(name='hoodlab',
   #                          token='41E779ABC50F6A98036C95AEEA1A92F7',
   #                          secret='')
   setRepositories(addURLs=c(BioCsoft="https://bioconductor.org/packages/3.12/bioc",
                             BioCann="https://bioconductor.org/packages/3.12/data/annotation",
                             BioCexp="https://bioconductor.org/packages/3.12/data/experiment",
                             BioC="https://bioconductor.org/packages/3.12/bioc",
                             CRAN="https://cran.microsoft.com"),
                   graphics=FALSE)

   deployApp(account="hoodlab",
              appName="igvShinyDemo-VCF",
              appTitle="igvShiny Demo VCF",
              appFiles=c("igvShinyDemo-withVCF.R", "tracks/placeholder"),
              appPrimaryDoc="igvShinyDemo-withVCF.R"
              )

} # deploy
#------------------------------------------------------------------------------------------------------------------------
if(grepl("hagfish", Sys.info()[["nodename"]]) & !interactive()){
   runApp(shinyApp(ui, server), port=6867)
   } else {
   shinyApp(ui, server)
   }

library(igvShiny)
library(cyjShiny)
#----------------------------------------------------------------------------------------------------
tbl.nodes <- data.frame(id=c("A", "B", "C"),
                        type=c("kinase", "TF", "glycoprotein"),
                        lfc=c(-3, 1, 1),
                        count=c(0, 0, 0),
                        stringsAsFactors=FALSE)

tbl.edges <- data.frame(source=c("A", "B", "C"),
                        target=c("B", "C", "A"),
                        interaction=c("phosphorylates", "synthetic lethal", "unknown"),
                        stringsAsFactors=FALSE)

graph.json <- toJSON(dataFramesToJSON(tbl.edges, tbl.nodes), auto_unbox=TRUE)
#----------------------------------------------------------------------------------------------------
ui = shinyUI(

   fluidPage(
     div(id="igvBox",
         igvShinyOutput('igvShiny')
         ),
     #igvShinyOutput('igvShiny_0'),
     #cyjShinyOutput('cyjShiny_0'),
     div(id="cyjBox",
         cyjShinyOutput('cyjShiny')
         ),
     width=10
     )

) # ui
#----------------------------------------------------------------------------------------------------
server = function(input, output, session)
{
   output$igvShiny <- renderIgvShiny({
      genomeOptions <- parseAndValidateGenomeSpec(genomeName="hg38",  initialLocus="NDUFS2")
      igvShiny(genomeOptions)
      })

   #output$igvShiny_0 <- renderIgvShiny({
   #  options <- list(genomeName="hg38", initialLocus="NDUFS2")
   #   igvShiny(options)
   #   })

   output$cyjShiny <- renderCyjShiny({
     cyjShiny(graph=graph.json, layoutName="cola")
     })

   #output$cyjShiny_0 <- renderCyjShiny({
#      cyjShiny(graph=graph.json, layoutName="random")
#      })


} # server
#----------------------------------------------------------------------------------------------------
runApp(shinyApp(ui=ui, server=server), port=9997)

# igvShiny — GWAS tracks, both ways of building one. loadGwasTrack() takes a
# data.frame and guesses the columns; the GWASTrack S4 class takes a data.frame
# or a URL and is told the columns, which is what non-standard headers need.

library(shiny)
library(bslib)
library(igvShiny)

f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
tbl.gwas <- get(load(f))

# some p-values in this remote table are zero, which -log10 turns into Inf, so
# it is displayed on a fixed scale rather than autoscaled
url.gwasTrack <- GWASTrack("remote url gwas",
                           "https://gladki.pl/igvShiny/gwas_sample.tsv.gz",
                           chrom.col = 3, pos.col = 4, pval.col = 10,
                           trackHeight = 100,
                           autoscale = FALSE, minY = 0, maxY = 300)

tbl.gwasTrack <- GWASTrack("data.frame gwas", tbl.gwas,
                           chrom.col = 3, pos.col = 4, pval.col = 10,
                           trackHeight = 100)
ui <- page_sidebar(
  title = "igvShiny — GWAS tracks",
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  fillable = TRUE,
  sidebar = sidebar(
    width = 280,
    actionButton("loadSimpleGwasTrackButton", "loadGwasTrack(data.frame)", class = "w-100 mb-2"),
    actionButton("loadLocalTableGwasTrackButton", "GWASTrack(data.frame)", class = "w-100 mb-2"),
    actionButton("loadRemoteUrlGwasTrackButton", "GWASTrack(url)", class = "w-100 mb-2"),
    hr(),
    actionButton("showApoeButton", "APOE region", class = "btn-outline-dark w-100 mb-2"),
    actionButton("showAllButton", "Whole genome", class = "btn-outline-dark w-100")
  ),
  card(full_screen = TRUE, card_body(class = "p-0",
                                     igvShinyOutput("igvShiny_0", height = "100%")))
)

server <- function(input, output, session) {
  observeEvent(input$loadSimpleGwasTrackButton,
               loadGwasTrack(session, id = "igvShiny_0", trackName = "gwas",
                             tbl.gwas = tbl.gwas, deleteTracksOfSameName = FALSE))
  observeEvent(input$loadLocalTableGwasTrackButton,
               display(tbl.gwasTrack, session, id = "igvShiny_0"))
  observeEvent(input$loadRemoteUrlGwasTrackButton,
               display(url.gwasTrack, session, id = "igvShiny_0"))
  observeEvent(input$showApoeButton,
               showGenomicRegion(session, id = "igvShiny_0",
                                 "chr19:45,303,720-45,463,047"))
  observeEvent(input$showAllButton,
               showGenomicRegion(session, id = "igvShiny_0", "all"))

  output$igvShiny_0 <- renderIgvShiny({
    igvShiny(parseAndValidateGenomeSpec(genomeName = "hg19", initialLocus = "all"))
  })
}

shinyApp(ui = ui, server = server)

#' get_tracks_dir
#
#' Get the directory where tracks are stored. The directory can be defined
#' with environmental variable.
#' If not defined the default is a directory called "tracks" in the temp
#' directory.
#
#' We need a local directory to write files - for instance,
#' a vcf file representing a genomic region of interest.
#' We then tell shiny about that directory, so that shiny's built-in http server
#' can serve up files we write there, ultimately consumed by igv.js
#'
#' @param env_var The name of the environmental variable to use.
#'
#' @return string with the path to the tracks directory.
#'
#' @examples
#' gtd <- get_tracks_dir(env_var = "TRACKS_DIR")
#'
#' @keywords utils
#' @export
get_tracks_dir <- function(env_var = "TRACKS_DIR") {
  checkmate::assert_string(env_var)
  default_dir <- file.path(tempdir(), "tracks")
  tracks_dir <- Sys.getenv(env_var, default_dir)

  checkmate::assert_access(dirname(tracks_dir), "rw")

  if (!dir.exists(tracks_dir)) {
    dir.create(tracks_dir, recursive = TRUE)
  }
  tracks_dir
} # get_tracks_dir
#-------------------------------------------------------------------------------
#' Get the tracks directory and make sure shiny serves that very directory
#'
#' @description The "tracks" resource path is registered once, when the package
#' is loaded. The default directory lives under \code{tempdir()}, which is not
#' guaranteed to stay put for the life of the process: when it moves, files are
#' written to the new directory while shiny still serves the old one, and igv.js
#' gets a 404 for a file that exists on disk. Re-registering the path before
#' each write keeps the served directory and the written directory the same.
#'
#' @return string with the path to the tracks directory.
#'
#' @keywords internal
.tracksDir <- function() {
  tracks_dir <- get_tracks_dir()
  # resourcePaths() is an atomic vector, so [["tracks"]] errors when the path is
  # not registered yet - which is the case the first time this runs (.onLoad)
  registered <- shiny::resourcePaths()
  served <- if ("tracks" %in% names(registered)) {
    registered[["tracks"]]
  } else {
    NULL
  }

  # normalizePath so that /var/... and its /private/var/... realpath (macOS)
  # do not look like two different directories
  canonical <- function(path) {
    normalizePath(path, winslash = "/", mustWork = FALSE)
  }

  if (is.null(served) || !identical(canonical(served), canonical(tracks_dir))) {
    if (!is.null(served)) {
      shiny::removeResourcePath("tracks")
    }
    shiny::addResourcePath("tracks", tracks_dir)
  }

  tracks_dir
} # .tracksDir
#-------------------------------------------------------------------------------
#' Name a file in the served tracks directory, and have it removed with the
#' session that asked for it
#'
#' @description Loaders write a file per track load and nothing ever removes
#' it. In an interactive session that is harmless - the default directory sits
#' under \code{tempdir()} and goes with the process - but a deployed app keeps
#' one R process across many user sessions, and \code{TRACKS_DIR} may point
#' outside \code{tempdir()} altogether, where nothing removes the files at all.
#' Alignment loads make it concrete: a bam export or a cram copy is gigabytes.
#'
#' The paths handed out to one session are collected in that session's
#' \code{userData}, and the first call registers a single
#' \code{session$onSessionEnded()} hook to unlink the set. Assignment goes
#' through a local binding to the \code{userData} environment on purpose:
#' \code{session$userData$x <- value} would call the assignment method of a
#' module's session proxy, which refuses to be written to.
#'
#' @param session a shiny session object
#' @param ext character string, the extension of the served file
#'
#' @return string with the path to the file
#'
#' @keywords internal
.trackFile <- function(session, ext) {
  .registerTrackFile(session, tempfile(tmpdir = .tracksDir(), fileext = ext))
} # .trackFile
#-------------------------------------------------------------------------------
#' Have a path removed when the session ends
#'
#' @description The companion of \code{.trackFile}, for the files a loader does
#' not name itself: \code{rtracklayer::export(format = "BAM")} writes an index
#' next to the bam it was given, and that index is served and leaks the same
#' way. The path need not exist yet, or ever - \code{unlink()} does not mind.
#'
#' @param session a shiny session object
#' @param path character string, the path to remove when the session ends
#'
#' @return the path, unchanged
#'
#' @keywords internal
.registerTrackFile <- function(session, path) {
  # a stub session (some downstream tests) has nothing to hang a hook on, and
  # a leaked file beats a loader that errors
  if (!is.function(session$onSessionEnded)) {
    return(path)
  }

  registry <- session$userData$igvShinyTrackFiles
  if (is.null(registry)) {
    registry <- new.env(parent = emptyenv())
    registry$paths <- character(0)
    user_data <- session$userData
    user_data$igvShinyTrackFiles <- registry
    session$onSessionEnded(function() unlink(registry$paths))
  }
  registry$paths <- c(registry$paths, path)

  path
} # .registerTrackFile
#-------------------------------------------------------------------------------
#' Make a file on disk reachable by igv.js
#'
#' @description Files igv.js reads by url have to sit in the directory shiny
#' serves as "tracks". Alignment files run to gigabytes, so the file is linked
#' rather than copied where the filesystem allows it (windows, and any mount
#' refusing links, falls back to a copy). The name is randomized: two loaders
#' may be handed same-named files from different directories.
#'
#' @param session a shiny session object; the staged file is removed when that
#' session ends
#' @param path character string, an existing file
#' @param ext character string, the extension of the served file
#'
#' @return string with the path to the file, relative to the shiny app
#'
#' @keywords internal
.stageTrackFile <- function(session, path, ext) {
  dest <- .trackFile(session, ext)
  ok <- suppressWarnings(file.symlink(normalizePath(path), dest))
  if (!ok) {
    ok <- file.copy(path, dest)
  }
  if (!ok) {
    stop(sprintf("igvShiny: could not stage '%s' for serving", path))
  }
  file.path("tracks", basename(dest))
} # .stageTrackFile

library(testthat)
library(igvShiny)

test_that("get_tracks_dir returns a valid character string", {
    # The checkmate test is good, so we can keep it inside an expect_true call
    expect_true(checkmate::test_string(get_tracks_dir()))
})

test_that(".tracksDir re-points the shiny resource path when it moves (#132)", {
  # simulate the tracks directory moving: point the resource path somewhere else
  # and check that the next .tracksDir() call brings it back in step
  moved <- file.path(tempdir(), "tracks-stale")
  dir.create(moved, showWarnings = FALSE, recursive = TRUE)
  shiny::removeResourcePath("tracks")
  shiny::addResourcePath("tracks", moved)

  dir <- igvShiny:::.tracksDir()

  served <- shiny::resourcePaths()[["tracks"]]
  expect_identical(normalizePath(served, winslash = "/", mustWork = FALSE),
                   normalizePath(dir, winslash = "/", mustWork = FALSE))
  # a file written into the returned directory is one shiny actually serves
  f <- tempfile(tmpdir = dir, fileext = ".bed")
  writeLines("chr1\t1\t2", f)
  expect_true(file.exists(file.path(served, basename(f))))
})

test_that("track files written for a session go when the session does (#152)", {
  session <- fake_session()
  tbl <- data.frame(chr = "chr1", start = 100, end = 200,
                    stringsAsFactors = FALSE)
  loadBedTrack(session, "igv", "one", tbl)
  loadBedTrack(session, "igv", "two", tbl)
  written <- session$userData$igvShinyTrackFiles$paths

  # two loads, two files, and one hook to remove both of them
  expect_length(written, 2L)
  expect_true(all(file.exists(written)))
  expect_length(session$endedCallbacks, 1L)

  end_session(session)
  expect_false(any(file.exists(written)))
})

test_that("a session without an onSessionEnded hook still gets a path (#152)", {
  # a stub session is no reason for a loader to error, here or downstream
  stub <- new.env(parent = emptyenv())
  path <- igvShiny:::.trackFile(stub, ".bed")

  expect_true(checkmate::test_string(path))
  expect_null(stub$userData)
})

library(igvShiny)

#----------------------------------------------------------------------------------------------------
f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
# print(dim(tbl.gwas))
printf <- function(...) print(noquote(sprintf(...)))
genomeSpec <- parseAndValidateGenomeSpec(genomeName="danRer11",  initialLocus="all")
stock.genomes <- sort(get_css_genomes())
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(
  sidebarLayout(
    sidebarPanel(
      selectInput("genomeChooser", "Choose stock igv genome:", stock.genomes, selected = "danRer11"),
      actionButton("getChromLocButton", "Get Region"),
      actionButton("clearChromLocButton", "Clear Region"),
      div(style = "background-color: white; width: 140px; height:20px; padding-left: 5px;
                   margin-top: 30px; border: 1px solid blue; font-size: 10px;",
          htmlOutput("chromLocDisplay")),
      hr(),
      width = 3
    ),
    mainPanel(
      igvShinyOutput('igvShiny_0'),
      width = 9
    )
  ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {

  observeEvent(input$genomeChooser, ignoreInit=TRUE, {
     newGenome <- input$genomeChooser
     printf("new genome: %s", newGenome)
     genomeSpec <- parseAndValidateGenomeSpec(genomeName=newGenome,  initialLocus="all")
     output$igvShiny_0 <- renderIgvShiny(
       igvShiny(genomeSpec)
       )
     })

  observeEvent(input$searchButton, {
    printf("--- search")
    searchString = isolate(input$roi)
    if(nchar(searchString) > 0)
      showGenomicRegion(session, id = "igvShiny_0", searchString)
  })

  observeEvent(input$addBamLocalFileButton, {
    printf("---- addBamLocalFileButton")
    bamFile <- system.file(package = "igvShiny", "extdata", "A_2_A24_02_01_01.nanopore.minimap.sorted.bam")
    x <- readGAlignments(bamFile)
    loadBamTrackFromLocalData(session, id = "igvShiny_0", trackName = "A_2_A24_02_01_01.nanopore.minimap.sorted.bam",
                              data = x, displayMode = "squished")
  })

  observeEvent(input$removeUserTracksButton, {
    printf("---- removeUserTracks")
    removeUserAddedTracks(session, id = "igvShiny_0")
  })


  observeEvent(input$trackClick, {
    printf("--- trackclick event")
    x <- input$trackClick
    print(x)
  })

  observeEvent(input[["igv-trackClick"]], {
    printf("--- igv-trackClick event")
    x <- input[["igv-trackClick"]]
    print(x)
  })

  observeEvent(input$getChromLocButton, {
    # printf("--- getChromLoc event")
    # sends message to igv.js in browser; currentGenomicRegion.<id> event sent back
    # see below for how that can be captured and displayed
    getGenomicRegion(session, id = "igvShiny_0")
  })

  observeEvent(input$clearChromLocButton, {
    output$chromLocDisplay <- renderText({" "})
  })

  observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
    newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
    #printf("new chromLocString: %s", newLoc)
    output$chromLocDisplay <- renderText({newLoc})
  })

  ##ü TODO add fasta and bam files to inst
  output$igvShiny_0 <- renderIgvShiny(
    igvShiny(genomeSpec)
    )


} # server
#------------------------------------------------------------------------------------------------------------------------
runApp(shinyApp(ui, server), port=6868)

# A minimal stand-in for the Shiny session object.
#
# Every track loader ends in `session$sendCustomMessage(type, message)`, so a
# fake session that records those calls lets the loaders be tested without a
# browser. That matters for coverage as well as speed: the shinytest2 tests run
# the app in a separate process, which covr cannot instrument, so everything
# exercised only through them counts as uncovered.

fake_session <- function() {
  self <- new.env(parent = emptyenv())
  self$messages <- list()
  self$sendCustomMessage <- function(type, message) {
    self$messages[[length(self$messages) + 1L]] <-
      list(type = type, message = message)
    invisible(NULL)
  }
  self
}

# The messages recorded so far, optionally narrowed to one message type.
sent_messages <- function(session, type = NULL) {
  if (is.null(type)) {
    return(session$messages)
  }
  Filter(function(m) identical(m$type, type), session$messages)
}

# The payload of the last message of `type`; fails the test if none was sent,
# which gives a clearer report than indexing into an empty list.
last_message <- function(session, type) {
  msgs <- sent_messages(session, type)
  testthat::expect_gt(length(msgs), 0L)
  msgs[[length(msgs)]]$message
}

library(testthat)
library(igvShiny)

# display() and show() are the two GWASTrack methods with no coverage: display()
# is the S4 counterpart of the loadTrack functions and sends the igv.js message,
# show() is the print method BiocCheck asks S4 classes to provide.

gwas_from_data_frame <- function() {
  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))
  GWASTrack("gwas", tbl.gwas, chrom.col = 3, pos.col = 4, pval.col = 10,
            trackHeight = 100)
}

test_that("display() sends the track and rewrites a local url into tracks/", {
  session <- fake_session()
  track <- gwas_from_data_frame()
  display(track, session, id = "igvShiny_0")

  msg <- last_message(session, "loadGwasTrackFlexibleSource")
  expect_equal(msg$trackName, "gwas")
  expect_equal(msg$dataMode, "local.url")
  # local.url data is rewritten to the shorthand resource path that the Shiny
  # web server exposes, not left as the absolute temp-dir path
  expect_match(msg$dataUrl, "^tracks/")
  expect_equal(msg$trackHeight, 100)
})

test_that("display() sends the column mapping igv.js needs", {
  session <- fake_session()
  display(gwas_from_data_frame(), session, id = "igvShiny_0")

  msg <- last_message(session, "loadGwasTrackFlexibleSource")
  # igv.js reads these 1-based, exactly as the constructor takes them; without
  # them it falls back to guessing the columns from the header names
  expect_equal(msg$columns, list(chromosome = 3, position = 4, value = 10))
})

test_that("display() sends a chromosome color map only when there is one", {
  session <- fake_session()
  display(gwas_from_data_frame(), session, id = "igvShiny_0")
  msg <- last_message(session, "loadGwasTrackFlexibleSource")
  # an empty colorTable would replace the igv.js palette with nothing
  expect_null(msg$colorTable)

  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))
  colors <- list("1" = "red", "*" = "gray")
  track <- GWASTrack("gwas", tbl.gwas, chrom.col = 3, pos.col = 4,
                     pval.col = 10, chromosomeColorMap = colors)
  session <- fake_session()
  display(track, session, id = "igvShiny_0")

  msg <- last_message(session, "loadGwasTrackFlexibleSource")
  expect_equal(msg$colorTable, colors)
})

test_that("display() removes same-named tracks unless told otherwise", {
  session <- fake_session()
  display(gwas_from_data_frame(), session, id = "igvShiny_0")
  expect_length(sent_messages(session, "removeTracksByName"), 1L)

  session <- fake_session()
  display(gwas_from_data_frame(), session, id = "igvShiny_0",
          deleteTracksOfSameName = FALSE)
  expect_length(sent_messages(session, "removeTracksByName"), 0L)
})

test_that("display() keeps a remote url untouched", {
  port <- local_server()
  url <- local_url(port, "gwas-5k.tsv.gz")
  track <- GWASTrack("remote gwas", url, chrom.col = 3, pos.col = 4,
                     pval.col = 10)
  session <- fake_session()
  display(track, session, id = "igvShiny_0")

  msg <- last_message(session, "loadGwasTrackFlexibleSource")
  expect_equal(msg$dataMode, "remote.url")
  expect_equal(msg$dataUrl, url)
})

test_that("show() prints the track summary", {
  out <- capture.output(show(gwas_from_data_frame()))
  expect_match(paste(out, collapse = "\n"), "GWASTrack object")
  expect_match(paste(out, collapse = "\n"), "trackName:  gwas")
  expect_match(paste(out, collapse = "\n"), "chrom=3")
  # the color line is printed only for a track that carries a map
  expect_false(any(grepl("colors:", out)))

  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))
  track <- GWASTrack("gwas", tbl.gwas, chrom.col = 3, pos.col = 4,
                     pval.col = 10,
                     chromosomeColorMap = list("1" = "red", "*" = "gray"))
  expect_match(paste(capture.output(show(track)), collapse = "\n"),
               "colors:     2 chromosomes")
})

library(shiny)
library(igvShiny)
library(GenomicAlignments)
library(later)
library(htmlwidgets)

#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
printf <- function(...) print(noquote(sprintf(...)))
#----------------------------------------------------------------------------------------------------
tbl.bed5 <- data.frame(chr=c("1","1", "1"),
                       start=c(7432951, 7437000, 7438000),
                       end=  c(7436000, 7437500, 7440000),
                       value=c(-2.239, 3.0, 0.5),
                       sampleID=c("sample1", "sample2", "sample3"),
                       stringsAsFactors=FALSE)

base.loc <- 161200000
tbl.bed9 <- data.frame(chrom=rep("chr1", 3),
                       start=c(base.loc, base.loc+100, base.loc + 250),
                       end=c(base.loc + 50, base.loc+120, base.loc+290),
                       name=c("red", "green", "blue"),
                       score=round(runif(3), 2),
                       strand=c("+", "-", "+"),
                       thickStart=c(base.loc+10, base.loc+110, base.loc+260),
                       thickEnd=c(base.loc+20, base.loc+130, base.loc+250),
                       itemRgb=c("255,0,128", "0,255,0", "0,0,255"),
                       stringsAsFactors=FALSE)

wig.size <- 100
values.100 <- runif(n=wig.size, min=-1, max=1)
starts.100 <- seq(from=7432951, to=7432951+(wig.size-1))
ends.100   <- starts.100 + 1
tbl.wig <- data.frame(chr=rep("1", wig.size),
                      start=starts.100,
                      end=ends.100,
                      value=values.100,
                      stringsAsFactors=FALSE)
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(
  
  sidebarLayout(
    sidebarPanel(
      actionButton("searchButton", "Search"),
      textInput("roi", label=""),
      h5("One simple data.frame, three igv formats:"),
      actionButton("addBedTrackButton", "Add as Bed"),
      actionButton("addBedGraphTrackButton", "Add as BedGraph"),
      actionButton("addBedGraphWithAltColorTrackButton", "Add as BedGraph (with AltColor)"),
      actionButton("addBedGraphTrackFromURLButton", "Add BedGraph from URL"),
      # actionButton("addAutoscaledGroupBedGraphTrackButton", "Add Autoscaled Group BedGraphs"),
      # data immediate seg track apparently abandoned with igv.js 2.10.4 or before
      #actionButton("addSegTrackButton", "Add as SEG"),
      br(),
      actionButton("addBed9TrackButton", "bed9 track"),
      actionButton("addGwasTrackButton", "Add GWAS Track"),
      actionButton("addBamViaHttpButton", "BAM from URL"),
      actionButton("addBamLocalFileButton", "BAM local data"),
      actionButton("addCramViaHttpButton", "CRAM from URL"),
      actionButton("removeUserTracksButton", "Remove User Tracks"),
      actionButton("getChromLocButton", "Get Region"),
      actionButton("clearChromLocButton", "Clear Region"),
      div(style="background-color: white; width: 200px; height:30px; padding-left: 5px;
                   margin-top: 10px; border: 1px solid blue;",
          htmlOutput("chromLocDisplay")),
      hr(),
      width=2
    ),
    mainPanel(
      igvShinyOutput('igvShiny_0'),
      # igvShinyOutput('igvShiny_1'),
      width=10
    )
  ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {
  
  observeEvent(input$searchButton, {
    printf("--- search")
    searchString = isolate(input$roi)
    if(nchar(searchString) > 0)
      showGenomicRegion(session, id="igvShiny_0", searchString)
  })
  
  observeEvent(input$addBedTrackButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
    loadBedTrack(session, id="igvShiny_0", trackName="bed5", tbl=tbl.bed5);
  })
  
  observeEvent(input$addBed9TrackButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:161,199,757-161,201,277")
    loadBedTrack(session, id="igvShiny_0", trackName="bed9", tbl=tbl.bed9)
  })
  
  observeEvent(input$addBedGraphTrackButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig/bedGraph/local", tbl=tbl.bed5,
                      color="blue", autoscale=TRUE)
  })
  
  observeEvent(input$addBedGraphWithAltColorTrackButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig/bedGraph/local/alt", tbl=tbl.bed5,
                      color="blue", trackConfig = list(altColor = "red"), autoscale=TRUE)
  })
  
  observeEvent(input$addAutoscaledGroupBedGraphTrackButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:7,432,868-7,433,167")
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig1a", tbl=tbl.wig, color="blue",
                      autoscale=TRUE, autoscaleGroup=1)
    tbl.wig1b <- tbl.wig
    tbl.wig1b$value <-tbl.wig1b$value * 10
    loadBedGraphTrack(session, id="igvShiny_0", trackName="wig1b", tbl=tbl.wig1b, color="brown",
                      autoscale=TRUE, autoscaleGroup=1)
  })
  
  #
  
  observeEvent(input$addBedGraphTrackFromURLButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:154,946,914-155,080,475")
    url <- "https://www.encodeproject.org/files/ENCFF356YES/@@download/ENCFF356YES.bigWig"
    loadBedGraphTrackFromURL(session, id="igvShiny_0", trackName="bedGraph/remote",
                             url=url, color="brown",
                             trackHeight=50, autoscale=TRUE)
  })
  
  
  observeEvent(input$addSegTrackButton, {
    showGenomicRegion(session, id="igvShiny_0", "chr1:7,426,231-7,453,241")
    loadSegTrack(session, id="igvShiny_0", trackName="seg", tbl=tbl.bed5)
  })
  
  observeEvent(input$addGwasTrackButton, {
    printf("---- addGWASTrack")
    printf("current working directory: %s", getwd())
    showGenomicRegion(session, id="igvShiny_0", "chr19:45,248,108-45,564,645")
    loadGwasTrack(session, id="igvShiny_0", trackName="gwas", tbl=tbl.gwas, deleteTracksOfSameName=FALSE)
  })
  
  observeEvent(input$addBamViaHttpButton, {
    printf("---- addBamViaHttpTrack")
    showGenomicRegion(session, id="igvShiny_0", "chr5:88,733,959-88,761,606")
    base.url <- "https://1000genomes.s3.amazonaws.com/phase3/data/HG02450/alignment"
    url <- sprintf("%s/%s", base.url, "HG02450.mapped.ILLUMINA.bwa.ACB.low_coverage.20120522.bam")
    indexURL <- sprintf("%s/%s", base.url, "HG02450.mapped.ILLUMINA.bwa.ACB.low_coverage.20120522.bam.bai")
    loadBamTrackFromURL(session, id="igvShiny_0",trackName="1kg.bam", bamURL=url, indexURL=indexURL)
  })
  
  observeEvent(input$addBamLocalFileButton, {
    printf("---- addBamLocalFileButton")
    showGenomicRegion(session, id="igvShiny_0", "chr21:10,397,614-10,423,341")
    bamFile <- system.file(package="igvShiny", "extdata", "tumor.bam")
    x <- readGAlignments(bamFile, param = Rsamtools::ScanBamParam(what="seq"))
    loadBamTrackFromLocalData(session, id="igvShiny_0", trackName="tumor.bam", data=x)
  })
  
  observeEvent(input$addCramViaHttpButton, {
    printf("---- addCramViaHttpTrack")
    showGenomicRegion(session, id="igvShiny_0", "chr5:88,733,959-88,761,606")
    base.url <- "https://s3.amazonaws.com/1000genomes"
    url <- sprintf("%s/%s", base.url, "1000G_2504_high_coverage/additional_698_related/data/ERR3989250/HG04160.final.cram")
    indexURL <- sprintf("%s.%s", url, "crai")
    loadCramTrackFromURL(session, id="igvShiny_0",trackName="CRAM", cramURL=url, indexURL=indexURL)
  })
  
  observeEvent(input$removeUserTracksButton, {
    printf("---- removeUserTracks")
    removeUserAddedTracks(session, id="igvShiny_0")
  })
  
  observeEvent(input$igvReady, {
    printf("--- igvReady")
    containerID <- input$igvReady
    printf("igv ready, %s", containerID)
    # loadBedTrack(session, id=containerID, trackName="bed5 loaded on ready", tbl=tbl.bed5, color="red");
  })
  
  observeEvent(input$trackClick, {
    printf("--- trackclick event")
    x <- input$trackClick
    print(x)
  })
  
  observeEvent(input[["igv-trackClick"]], {
    printf("--- igv-trackClick event")
    x <- input[["igv-trackClick"]]
    print(x)
    attribute.name.positions <- grep("name", names(x))
    attribute.value.positions <- grep("value", names(x))
    attribute.names <- as.character(x)[attribute.name.positions]
    attribute.values <- as.character(x)[attribute.value.positions]
    tbl <- data.frame(name=attribute.names,
                      value=attribute.values,
                      stringsAsFactors=FALSE)
    dialogContent <- renderTable(tbl)
    html <- HTML(dialogContent())
    showModal(modalDialog(html, easyClose=TRUE))
  })
  
  observeEvent(input$getChromLocButton, {
    # printf("--- getChromLoc event")
    # sends message to igv.js in browser; currentGenomicRegion.<id> event sent back
    # see below for how that can be captured and displayed
    getGenomicRegion(session, id="igvShiny_0")
  })
  
  observeEvent(input$clearChromLocButton, {
    output$chromLocDisplay <- renderText({" "})
  })
  
  observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
    newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
    #printf("new chromLocString: %s", newLoc)
    output$chromLocDisplay <- renderText({newLoc})
  })
  
  genomes <- c("hg38", "hg19", "mm10", "tair10", "rhos")
  loci <- c("chr5:88,466,402-89,135,305",  "chr1:7,426,231-7,453,241", "MEF2C", "Mef2c",
            "1:7,432,931-7,440,395", "NC_007494.2:370,757-378,078",
            "chr1:6,575,383-8,304,088")
  
  output$igvShiny_0 <- renderIgvShiny({
    cat("--- starting renderIgvShiny\n");
    genomeOptions <- parseAndValidateGenomeSpec(genomeName="hg38",  initialLocus=loci[7])
    x <- igvShiny(genomeOptions,
                  displayMode="SQUISHED",
                  tracks=list()
    )
    cat("--- ending renderIgvShiny\n");
    return(x)
  })
  
} # server
#----------------------------------------------------------------------------------------------------
deploy <-function()
{
  repos <- options("repos")[[1]]
  stopifnot(sort(names(repos)) == c("BioCann", "BioCsoft", "CRAN"))
  stopifnot(repos$BioCann=="https://bioconductor.org/packages/3.13/data/annotation")
  stopifnot(repos$BioCsoft=="https://bioconductor.org/packages/3.13/bioc")
  stopifnot(repos$CRAN=="https://cran.microsoft.com")
  require(devtools)
  
  # jim hester suggests, with reference
  # Setting R_REMOTES_NO_ERRORS_FROM_WARNINGS="false" will cause warning
  # messages during calls to install.packages() to become errors. Often warning
  # messages are caused by dependencies failing to install.
  Sys.setenv("R_REMOTES_NO_ERRORS_FROM_WARNINGS" = "true")
  
  install_github("paul-shannon/igvShiny", force=TRUE)
  
  require(rsconnect)
  
  deployApp(account="hoodlab",
            appName="igvShinyDemo",
            appTitle="igvShiny Demo",
            appFiles=c("igvShinyDemo.R", "tracks/file14c6569b08f1.bam"),
            appPrimaryDoc="igvShinyDemo.R"
  )
  
} # deploy
#------------------------------------------------------------------------------------------------------------------------
# shinyApp(ui = ui, server = server)
if(grepl("hagfish", Sys.info()[["nodename"]]) & !interactive()){
  runApp(shinyApp(ui, server), port=6867)
} else {
  shinyApp(ui, server)
}

CAS_GENOMES <-
  c("hg38", "hg19", "mm10", "tair10", "custom", "dm6", "sacCer3")

BASIC_GENOMES <-
  c("hg38",
    "hg19",
    "mm10",
    "tair10",
    "rhos",
    "custom",
    "dm6",
    "sacCer3")

randomColors <- randomcoloR::distinctColorPalette(24)

state <- new.env(parent = emptyenv())
state[["userAddedTracks"]] <- list()

library(shiny)
library(igvShiny)
library(GenomicAlignments)
library(htmlwidgets)
library(VariantAnnotation)
library(biomaRt)

#----------------------------------------------------------------------------------------------------
f <- system.file(package="igvShiny", "extdata", "gwas.RData")
stopifnot(file.exists(f))
tbl.gwas <- get(load(f))
# print(dim(tbl.gwas))
printf <- function(...) print(noquote(sprintf(...)))
#----------------------------------------------------------------------------------------------------
tbl.bed <- data.frame(chr=c("1","1", "1"),
                      start=c(7432951, 7437000, 7438000),
                      end=  c(7436000, 7437500, 7440000),
                      value=c(-2.239, 3.0, 0.5),
                      sampleID=c("sample1", "sample2", "sample3"),
                      stringsAsFactors=FALSE)
#----------------------------------------------------------------------------------------------------
ui = shinyUI(fluidPage(

  sidebarLayout(
     sidebarPanel(
        textInput("roi", label="", placeholder="gene or chrN:start-end"),
        actionButton("searchButton", "Search"),
        br(), br(), br(),
        actionButton("addLocalVCFTrackButton", "Add 1kg VCF (local file)"),
        actionButton("addRemoteVCFTrackButton", "Add 1kg VCF (AWS)"),
        br(), br(),  br(),
        actionButton("removeUserTracksButton", "Remove User Tracks"),
        br(), br(),
        actionButton("getChromLocButton", "Get Region"),
        actionButton("clearChromLocButton", "Clear Region Readout"),
        div(style="background-color: white; width: 200px; height:30px; padding-left: 5px;
                   margin-top: 10px; border: 1px solid blue;",
            htmlOutput("chromLocDisplay")),
        hr(),
        width=2
        ),
     mainPanel(
        igvShinyOutput('igvShiny_0'),
        # igvShinyOutput('igvShiny_1'),
        width=10
        )
     ) # sidebarLayout
))
#----------------------------------------------------------------------------------------------------
server = function(input, output, session) {

   observeEvent(input$searchButton, {
      searchString = isolate(input$roi)
      printf("--- search: %s", searchString)
      if(nchar(searchString) > 0)
        showGenomicRegion(session, id="igvShiny_0", searchString)
      })

    observeEvent(input$addLocalVCFTrackButton, {
       f <- system.file(package="igvShiny", "extdata", "chr19-cebpaRegion.vcf.gz")
       file.exists(f) # [1] TRUE
       vcf <- readVcf(f, "hg38")
       showGenomicRegion(session, id="igvShiny_0", sprintf("chr19:%d-%d", 33299144-1000, 33300107+1000))
       loadVcfTrack(session, id="igvShiny_0", trackName="vcf", vcf)
       })

    observeEvent(input$addRemoteVCFTrackButton, {
       rng <- GRanges(seqnames="22", ranges=IRanges(start=33298112+1000, end=33298112+2000))
         # tabix index file must also be there
       showGenomicRegion(session, id="igvShiny_0", sprintf("chr22:%d-%d", 33298112+1000, end=33298112+2000))
       url <-  "https://s3.amazonaws.com/1000genomes/release/20130502/ALL.chr22.phase3_shapeit2_mvncall_integrated_v5a.20130502.genotypes.vcf.gz";
       vcf <- readVcf(url, "hg38", rng)
       loadVcfTrack(session, id="igvShiny_0", trackName="vcf", vcf)
       })


   observeEvent(input$removeUserTracksButton, {
      printf("---- removeUserTracks")
      removeUserAddedTracks(session, id="igvShiny_0")
      })


  # observeEvent(input$trackClick, {
  #     printf("--- trackclick event")
  #     x <- input$trackClick
  #     print(x)
  #     })

   observeEvent(input[["igv-trackClick"]], {
       printf("--- igv-trackClick event")
       x <- input[["igv-trackClick"]]
       attribute.name.positions <- grep("name", names(x))
       attribute.value.positions <- grep("value", names(x))
       attribute.names <- as.character(x)[attribute.name.positions]
       attribute.values <- as.character(x)[attribute.value.positions]
       tbl <- data.frame(name=attribute.names,
                         value=attribute.values,
                         stringsAsFactors=FALSE)
       dialogContent <- renderTable(tbl)
       html <- HTML(dialogContent())
       showModal(modalDialog(html, easyClose=TRUE))
       })

   observeEvent(input$getChromLocButton, {
      # printf("--- getChromLoc event")
      # sends message to igv.js in browser; currentGenomicRegion.<id> event sent back
      # see below for how that can be captured and displayed
      getGenomicRegion(session, id="igvShiny_0")
      })

   observeEvent(input$clearChromLocButton, {
      output$chromLocDisplay <- renderText({" "})
      })

   observeEvent(input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]], {
      newLoc <- input[[sprintf("currentGenomicRegion.%s", "igvShiny_0")]]
      #printf("new chromLocString: %s", newLoc)
      output$chromLocDisplay <- renderText({newLoc})
      })

   genomes <- c("hg38", "hg19", "mm10", "tair10", "rhos")
   loci <- c("chr5:88,466,402-89,135,305", "MEF2C", "Mef2c", "1:7,432,931-7,440,395", "NC_007494.2:370,757-378,078")
   i <- 2

   output$igvShiny_0 <- renderIgvShiny({
     genomeOptions <- parseAndValidateGenomeSpec(genomeName="hg38",  initialLocus=loci[i])
     igvShiny(genomeOptions)
     })

} # server
#----------------------------------------------------------------------------------------------------
deploy <-function()
{
   require(rsconnect)
   #rsconnect::setAccountInfo(name='hoodlab',
   #                          token='41E779ABC50F6A98036C95AEEA1A92F7',
   #                          secret='')
   setRepositories(addURLs=c(BioCsoft="https://bioconductor.org/packages/3.12/bioc",
                             BioCann="https://bioconductor.org/packages/3.12/data/annotation",
                             BioCexp="https://bioconductor.org/packages/3.12/data/experiment",
                             BioC="https://bioconductor.org/packages/3.12/bioc",
                             CRAN="https://cran.microsoft.com"),
                   graphics=FALSE)

   deployApp(account="hoodlab",
              appName="igvShinyDemo-VCF",
              appTitle="igvShiny Demo VCF",
              appFiles=c("igvShinyDemo-withVCF.R", "tracks/placeholder"),
              appPrimaryDoc="igvShinyDemo-withVCF.R"
              )

} # deploy
#------------------------------------------------------------------------------------------------------------------------
if(grepl("hagfish", Sys.info()[["nodename"]]) & !interactive()){
   runApp(shinyApp(ui, server), port=6867)
   } else {
   shinyApp(ui, server)
   }

# igvShiny — inside a Shiny module. The widget takes its namespace from the
# session, so igvShinyOutput(ns("igv")) is all the UI needs. The catch is on the
# server side: loaders address the browser by its HTML element id, the
# namespaced ns("igv") and never "igv", while the region event comes back on the
# unnamespaced name, since it is named after the element inside the module.

library(shiny)
library(bslib)
library(igvShiny)

tbl.bed <- data.frame(chr = c("1", "1", "1"),
                      start = c(7432951, 7437000, 7438000),
                      end   = c(7436000, 7437500, 7440000),
                      value = c(-2.239, 3.0, 0.5),
                      sampleID = c("sample1", "sample2", "sample3"),
                      stringsAsFactors = FALSE)

igvModuleUI <- function(id) {
  ns <- NS(id)
  card(
    full_screen = TRUE,
    card_header(sprintf("module id: %s", id)),
    card_body(
      actionButton(ns("addTrack"), "Add BED track", class = "mb-2"),
      div(class = "small text-muted font-monospace", textOutput(ns("region"))),
      igvShinyOutput(ns("igv"), height = "400px")
    )
  )
}

igvModuleServer <- function(id, genomeOptions, tbl) {
  moduleServer(id, function(input, output, session) {
    ns <- session$ns

    output$igv <- renderIgvShiny(igvShiny(genomeOptions))

    observeEvent(input$addTrack, {
      loadBedTrack(session, id = ns("igv"), trackName = "regions", tbl = tbl,
                   color = "green")
    })

    # the region input keeps its unnamespaced name inside the module
    output$region <- renderText(input[["currentGenomicRegion.igv"]])
  })
}

ui <- page_fillable(
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  igvModuleUI("browserA")
)

server <- function(input, output, session) {
  options <- parseAndValidateGenomeSpec(genomeName = "hg38",
                                        initialLocus = "chr1:7,426,231-7,453,241")
  igvModuleServer("browserA", options, tbl.bed)
}

shinyApp(ui = ui, server = server)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
    )

## ----install, eval=FALSE------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("igvShiny")

## ----load-package, results="hide", message=FALSE, warning=FALSE---------------
library(igvShiny)

## ----genome-spec--------------------------------------------------------------
genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = "hg38",
    initialLocus = "NDUFS2"
    )
str(genomeOptions)

## ----minimal-app, eval=FALSE--------------------------------------------------
# library(shiny)
# library(igvShiny)
# 
# genomeOptions <- parseAndValidateGenomeSpec(
#     genomeName = "hg38",
#     initialLocus = "NDUFS2"
#     )
# 
# ui <- fluidPage(
#     titlePanel("igvShiny"),
#     igvShinyOutput("igv")
#     )
# 
# server <- function(input, output, session) {
#     output$igv <- renderIgvShiny({
#         igvShiny(genomeOptions)
#         })
#     }
# 
# shinyApp(ui, server)

## ----bed-track, eval=FALSE----------------------------------------------------
# tbl <- data.frame(
#     chrom = c("chr1", "chr1", "chr1"),
#     start = c(7432000, 7437000, 7443000),
#     end   = c(7436000, 7442000, 7447000),
#     value = c(0.2, 0.9, 0.4),
#     stringsAsFactors = FALSE
#     )
# 
# server <- function(input, output, session) {
#     output$igv <- renderIgvShiny({
#         igvShiny(genomeOptions)
#         })
# 
#     observeEvent(input$addTrack, {
#         loadBedTrack(
#             session,
#             id = "igv",
#             trackName = "my regions",
#             tbl = tbl,
#             color = "darkblue"
#             )
#         })
#     }

## ----url-track, eval=FALSE----------------------------------------------------
# gff3 <- paste0(
#     "https://s3.amazonaws.com/igv.org.genomes/hg38/",
#     "Homo_sapiens.GRCh38.94.chr.gff3.gz"
#     )
# 
# loadGFF3TrackFromURL(
#     session,
#     id = "igv",
#     trackName = "genes",
#     gff3URL = gff3,
#     indexURL = paste0(gff3, ".tbi"),
#     color = "darkgreen",
#     trackHeight = 100
#     )

## ----startup-tracks, eval=FALSE-----------------------------------------------
# output$igv <- renderIgvShiny({
#     igvShiny(
#         genomeOptions,
#         tracks = list(
#             list(
#                 name = "genes",
#                 type = "annotation",
#                 format = "gff3",
#                 url = paste0(
#                     "https://s3.amazonaws.com/igv.org.genomes/hg38/",
#                     "Homo_sapiens.GRCh38.94.chr.gff3.gz"
#                     ),
#                 indexed = FALSE
#                 )
#             )
#         )
#     })

## ----merged-track, eval=FALSE-------------------------------------------------
# sample <- paste0(
#     "https://raw.githubusercontent.com/igvteam/igv-data/main/data/test/",
#     "splice_junctions/splice_junction_track_test_cases_sampleA.",
#     "chr15-92835700-93031800"
#     )
# 
# output$igv <- renderIgvShiny({
#     igvShiny(
#         genomeOptions,
#         tracks = list(
#             list(
#                 name = "sampleA",
#                 type = "merged",
#                 height = 180,
#                 tracks = list(
#                     list(
#                         type = "wig",
#                         format = "bigwig",
#                         url = paste0(sample, ".bigWig"),
#                         color = "rgb(190,190,190)"
#                         ),
#                     list(
#                         type = "junction",
#                         format = "bed",
#                         url = paste0(sample, ".SJ.out.bed.gz"),
#                         indexURL = paste0(sample, ".SJ.out.bed.gz.tbi"),
#                         colorBy = "motif"
#                         )
#                     )
#                 )
#             )
#         )
#     })

## ----navigate, eval=FALSE-----------------------------------------------------
# observeEvent(input$goto, {
#     showGenomicRegion(
#         session,
#         id = "igv",
#         region = "chr5:88,700,000-88,800,000"
#         )
#     })

## ----read-region, eval=FALSE--------------------------------------------------
# observeEvent(input$whereAmI, {
#     getGenomicRegion(session, id = "igv")
#     })
# 
# observeEvent(input[["currentGenomicRegion.igv"]], {
#     message("now showing: ", input[["currentGenomicRegion.igv"]])
#     })

## ----module, eval=FALSE-------------------------------------------------------
# igvModuleUI <- function(id) {
#     ns <- NS(id)
#     tagList(
#         actionButton(ns("addTrack"), "Add track"),
#         igvShinyOutput(ns("igv"))
#         )
#     }
# 
# igvModuleServer <- function(id, genomeOptions, tbl) {
#     moduleServer(id, function(input, output, session) {
#         ns <- session$ns
# 
#         output$igv <- renderIgvShiny({
#             igvShiny(genomeOptions)
#             })
# 
#         observeEvent(input$addTrack, {
#             loadBedTrack(
#                 session,
#                 id = ns("igv"),
#                 trackName = "regions",
#                 tbl = tbl
#                 )
#             })
# 
#         # the region input keeps its unnamespaced name inside the module
#         observeEvent(input[["currentGenomicRegion.igv"]], {
#             message("now showing: ", input[["currentGenomicRegion.igv"]])
#             })
#         })
#     }

## ----custom-genome------------------------------------------------------------
data_directory <- system.file(package = "igvShiny", "extdata")

customOptions <- parseAndValidateGenomeSpec(
    genomeName = "ribosomal RNA gene",
    initialLocus = "U13369.1:7,276-8,225",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = file.path(data_directory, "ribosomal-RNA-gene.fasta"),
    fastaIndex = file.path(data_directory, "ribosomal-RNA-gene.fasta.fai"),
    genomeAnnotation = file.path(data_directory, "ribosomal-RNA-gene.gff3")
    )
customOptions[c("genomeName", "dataMode", "stockGenome")]

## ----custom-genome-http, eval=FALSE-------------------------------------------
# base_url <- "https://gladki.pl/igvr/testFiles"
# 
# remoteOptions <- parseAndValidateGenomeSpec(
#     genomeName = "ribosomal RNA gene",
#     initialLocus = "U13369.1:7,276-8,225",
#     stockGenome = FALSE,
#     dataMode = "http",
#     fasta = paste0(base_url, "/ribosomal-RNA-gene.fasta"),
#     fastaIndex = paste0(base_url, "/ribosomal-RNA-gene.fasta.fai"),
#     genomeAnnotation = paste0(base_url, "/ribosomal-RNA-gene.gff3")
#     )
# remoteOptions[c("genomeName", "dataMode", "stockGenome")]

## ----session-info-------------------------------------------------------------
sessionInfo()

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
    )

## ----load-package, results="hide", message=FALSE, warning=FALSE---------------
library(igvShiny)
futile.logger::flog.threshold(futile.logger::WARN)

## ----stub-session, include = FALSE--------------------------------------------
# The loaders are one-way messages to a browser: they take a shiny session and
# send to it, and there is no browser here. A session that prints instead of
# sending lets every example below run, and shows what igv.js receives.
session <- list(sendCustomMessage = function(type, message) {
    flatten <- function(v) paste(v, collapse = ",")
    fields <- vapply(message, flatten, character(1))
    cat(type, "\n", paste0("  ", names(fields), ": ", fields, collapse = "\n"),
        "\n", sep = "")
    })

## ----trackconfig--------------------------------------------------------------
tbl <- data.frame(
    chrom = "chr1",
    start = 7432000,
    end = 7436000
    )

loadBedTrack(
    session,
    id = "igv",
    trackName = "my regions",
    tbl = tbl,
    trackConfig = list(displayMode = "EXPANDED", maxRows = 20)
    )

## ----rejected-key-------------------------------------------------------------
genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = "hg38",
    initialLocus = "NDUFS2"
    )

widget <- igvShiny(
    genomeOptions,
    tracks = list(
        list(
            name = "genes",
            type = "annotation",
            url = "https://example.org/genes.bed",
            onClick = "alert('hi')"
            )
        )
    )

names(widget$x$tracks[[1]])

## ----wig----------------------------------------------------------------------
coverage <- data.frame(
    chrom = "chr1",
    start = c(7432000, 7437000),
    end = c(7436000, 7442000),
    value = c(0.2, 0.9)
    )

loadBedGraphTrack(
    session, id = "igv", trackName = "coverage", tbl = coverage,
    autoscale = TRUE,
    trackConfig = list(
        graphType = "points",
        autoscaleGroup = "sampleA",
        logScale = FALSE
        )
    )

## ----gff3---------------------------------------------------------------------
gff3.url <- paste0(
    "https://s3.amazonaws.com/igv.org.genomes/hg38/",
    "Homo_sapiens.GRCh38.94.chr.gff3.gz"
    )

loadGFF3TrackFromURL(
    session, id = "igv", trackName = "genes",
    gff3URL = gff3.url, indexURL = paste0(gff3.url, ".tbi"),
    colorByAttribute = "biotype",
    displayMode = "EXPANDED",
    visibilityWindow = 1000000,
    colorTable = list(
        protein_coding = "darkgreen",
        processed_transcript = "blue",
        default = "black"
        )
    )

## ----bam-sort-----------------------------------------------------------------
bam.url <- "https://1000genomes.s3.amazonaws.com/phase3/NA12878.bam"

loadBamTrackFromURL(
    session, id = "igv", trackName = "reads",
    bamURL = bam.url, indexURL = paste0(bam.url, ".bai"),
    trackConfig = list(
        sort = list(
            option = "TAG", tag = "HP",
            chr = "chr8", position = 128750000
            )
        )
    )

## ----gwas---------------------------------------------------------------------
gwas <- data.frame(
    CHR = "chr1",
    BP = c(1000, 2000),
    SNP = c("rs1", "rs2"),
    P = c(1e-8, 1e-5)
    )

loadGwasTrack(
    session, id = "igv", trackName = "gwas", tbl.gwas = gwas,
    trackConfig = list(
        columns = list(chromosome = 12, position = 13, value = 28)
        )
    )

## ----junctions----------------------------------------------------------------
junctions.url <- paste0(
    "https://raw.githubusercontent.com/igvteam/igv-data/main/data/test/",
    "splice_junctions/sampleA.SJ.out.bed.gz"
    )

loadSpliceJunctionTrackFromURL(
    session, id = "igv", trackName = "sampleA junctions",
    url = junctions.url, indexURL = paste0(junctions.url, ".tbi"),
    trackHeight = 150,
    trackConfig = list(
        colorBy = "motif",
        labelWith = "uniquelyMapped",
        minUniquelyMappedReads = 5,
        hideAnnotatedJunctions = FALSE
        )
    )

## ----startup------------------------------------------------------------------
gff3.url <- paste0(
    "https://s3.amazonaws.com/igv.org.genomes/hg38/",
    "Homo_sapiens.GRCh38.94.chr.gff3.gz"
    )

startupTrack <- list(
    name = "genes",
    type = "annotation",
    format = "gff3",
    url = gff3.url,
    indexed = FALSE,
    displayMode = "EXPANDED"
    )

widget <- igvShiny(genomeOptions, tracks = list(startupTrack))
str(widget$x$tracks[[1]])

## ----startup-app, eval=FALSE--------------------------------------------------
# output$igv <- renderIgvShiny({
#     igvShiny(genomeOptions, tracks = list(startupTrack))
#     })

## ----session-info-------------------------------------------------------------
sessionInfo()

#' @title get_basic_genomes
#' @description a helper function for basic genomes,
#' obtains the genome codes (e.g. 'hg38')
#'
#' @rdname get_basic_genomes
#' @aliases get_basic_genomes
#'
#' @return an list of short genome codes, e.g., "hg38", "dm6"
#'
#' @examples
#' bs <- get_basic_genomes()
#'
#' @keywords utils
#' @export
#'
get_basic_genomes <- function() {
  BASIC_GENOMES

} # get_basic_genomes
#-------------------------------------------------------------------------------
#' @title get_cas_genomes
#' @description a helper function for common always available stock genomes,
#' obtains the genome codes (e.g. 'hg38')
#'
#' @rdname get_cas_genomes
#' @aliases get_cas_genomes
#'
#' @return an list of short genome codes, e.g., "hg38", "dm6"
#'
#' @examples
#' cas <- get_cas_genomes()
#'
#' @keywords utils
#' @export
#'
get_cas_genomes <- function() {
  CAS_GENOMES

} # get_cas_genomes
#-------------------------------------------------------------------------------
#' @title get_css_genomes
#' @description a helper function for mostly internal use,
#' obtains the genome codes (e.g. 'hg38') supported by igv.js
#'
#' @rdname get_css_genomes
#' @aliases get_css_genomes
#' @param test logical(1) defaults to FALSE
#'
#' @return an list of short genome codes, e.g., "hg38", "dm6", "tair10"
#'
#' @examples
#' css <- get_css_genomes(test = TRUE)
#'
#' @keywords utils
#' @export
#'
get_css_genomes <- function(test = FALSE) {
  if (test)
    return(get_basic_genomes())

  # The registry the bundled igv.js 3.x reads (DEFAULT_GENOMES_URL). The 2.x
  # file, genomes.json, still exists and lists ids 3.x cannot resolve by name.
  current.genomes.file <- "https://igv.org/genomes/genomes3.json"

  if (httr::http_error(current.genomes.file))
    return(get_basic_genomes())

  current.genomes.raw <-
    readLines(current.genomes.file, warn = FALSE, skipNul = TRUE)
  tbl.genomes <- jsonlite::fromJSON(current.genomes.raw)
  tbl.genomes$id

} # get_css_genomes
#-------------------------------------------------------------------------------
#' @title parseAndValidateGenomeSpec
#' @description a helper function for internal use by the igvShiny constructor,
#' but possible also of use to those building an igvShiny app,
#' to test their genome specification for validity
#'
#' @rdname  parseAndValidateGenomeSpec
#' @aliases parseAndValidateGenomeSpec
#'
#' @param genomeName character usually one short code of a supported ("stock")
#' genome (e.g., "hg38") or for a user-supplied custom genome,
#' the name you wish to use
#' @param initialLocus character default "all", otherwise "chrN:start-end"
#' or a recognized gene symbol
#' @param stockGenome logical default TRUE
#' @param dataMode character either "stock", "localFile" or "http"
#' @param fasta character when supplying a custom (non-stock) genome,
#' either a file path or a URL
#' @param fastaIndex character when supplying a custom (non-stock) genome,
#' either a file path or a URL,
#' essential for all but the very small custom genomes.
#' @param genomeAnnotation character when supplying a custom (non-stock) genome,
#' a file path or URL pointing to a genome annotation file in
#' a gff3 format
#'
#' @examples
#' genomeSpec <-
#'   parseAndValidateGenomeSpec("hg38", "APOE")  # the simplest case
#' # a custom genome, read from the files shipped with the package: every
#' # resource is checked for existence, and "http" would check it over the
#' # network on each R CMD check
#' base.dir <-
#'   system.file(package = "igvShiny", "extdata", "sarsGenome")
#' fasta.file <-
#'   file.path(base.dir, "Sars_cov_2.ASM985889v3.dna.toplevel.fa")
#' fastaIndex.file <- paste0(fasta.file, ".fai")
#' annotation.file <-
#'   file.path(base.dir, "Sars_cov_2.ASM985889v3.101.gff3")
#' custom.genome.title <- "SARS-CoV-2"
#' genomeOptions <-
#'   parseAndValidateGenomeSpec(
#'     genomeName = custom.genome.title,
#'     initialLocus = "all",
#'     stockGenome = FALSE,
#'     dataMode = "localFiles",
#'     fasta = fasta.file,
#'     fastaIndex = fastaIndex.file,
#'     genomeAnnotation = annotation.file
#'   )
#'
#' @seealso [get_css_genomes()] for stock genomes we support.
#'
#' @return an options list directly usable by igvApp.js, and thus igv.js
#' @keywords igvShiny
#' @export
#'
parseAndValidateGenomeSpec <- function(genomeName,
                                       initialLocus = "all",
                                       stockGenome = TRUE,
                                       dataMode = NA,
                                       fasta = NA,
                                       fastaIndex = NA,
                                       genomeAnnotation = NA) {
  options <- list()
  options[["stockGenome"]] <- stockGenome
  options[["dataMode"]] <- dataMode
  options[["validated"]] <- FALSE


  #--------------------------------------------------
  # first: is this a stock genome?  if so, we need
  # only check if the genomeName is recognized
  #--------------------------------------------------

  if (stockGenome) {
    if (!genomeName %in% get_cas_genomes()) {
      supported.stock.genomes <- get_css_genomes()
      if (!genomeName %in% supported.stock.genomes) {
        s.1 <-
          sprintf("Your genome '%s' is not currently supported",
                  genomeName)
        s.2 <-
          sprintf("Currently supported: %s",
                  paste(supported.stock.genomes, collapse = ","))
        msg <- sprintf("%s\n%s", s.1, s.2)
        stop(msg)
      }
    } # if not common & always available
    options[["genomeName"]] <- genomeName
    options[["initialLocus"]] <- initialLocus
    options[["fasta"]] <- NA
    options[["fastaIndex"]] <- NA
    options[["annotation"]] <- NA
    options[["validated"]] <- TRUE
  }# stockGenome requested

  if (!stockGenome) {
    stopifnot(!is.na(dataMode))
    stopifnot(!is.na(fasta))
    stopifnot(!is.na(fastaIndex))
    # genomeAnnotation is optional

    # "direct" for an in-memory R data structure, deferred
    recognized.modes <- c("localFiles", "http")
    if (!dataMode %in% recognized.modes) {
      msg <-
        sprintf(
          "dataMode '%s' should be one of %s",
          dataMode,
          paste(recognized.modes, collapse = ",")
        )
      stop(msg)
    }
    #---------------------------------------------------------------------
    # dataMode determines how to check for the existence of each resource
    #---------------------------------------------------------------------

    exists.function <- switch(dataMode,
                              "localFiles" = file.exists,
                              "http" = function(x, ...) {
                                !(httr::http_error(x, ...))
                              })
    stopifnot(exists.function(fasta))
    stopifnot(exists.function(fastaIndex))
    if (!is.na(genomeAnnotation))
      stopifnot(exists.function(genomeAnnotation))

    options[["genomeName"]]  <- genomeName
    options[["fasta"]] <- fasta
    options[["fastaIndex"]] <- fastaIndex
    options[["initialLocus"]] <- initialLocus
    options[["annotation"]] <- genomeAnnotation
    options[["validated"]] <- TRUE
  } # if !stockGenome

  return(options)

} # parseAndValidateGenomeSpec
#-------------------------------------------------------------------------------

library(testthat)
library(igvShiny)

test_that("Supported genomes can be retrieved correctly", {
  cg.minimal <- get_css_genomes(test = TRUE)
  expect_equal(cg.minimal,
               c(
                 "hg38",
                 "hg19",
                 "mm10",
                 "tair10",
                 "rhos",
                 "custom",
                 "dm6",
                 "sacCer3"
               ))

  cg <- get_css_genomes()

  # With igv.org unreachable get_css_genomes() falls back to the eight ids
  # above, so the two assertions below would fail on someone else's outage and
  # read as a code regression
  skip_if(identical(cg, get_basic_genomes()),
          "genome registry unreachable, fell back to the built-in list")

  expect_true(length(cg) > 30)

  # the list has to come from the registry the bundled igv.js reads, so that
  # every id it reports resolves by bare name; mm10, danRer11 and rn6 are the
  # genomes #107 was about
  expect_true(all(c("hg38", "mm10", "danRer11", "rn6") %in% cg))
})

test_that("Cached (caas) genomes are retrieved quickly", {
  t1 <- system.time(caasg <- get_cas_genomes())
  expect_lt(t1[["elapsed"]], 0.1)
  expect_true(all(
    c("hg38", "hg19", "mm10", "tair10", "custom", "dm6", "sacCer3") %in% caasg
  ))

  t2 <- system.time(cssg  <- get_css_genomes())
  expect_true(t2[["elapsed"]] > 0.01)

  # Parsing a cached genome spec should be fast
  t3 <- system.time(
    parseAndValidateGenomeSpec(
      genomeName = "hg38",
      initialLocus = "NDUFS2",
      stockGenome = TRUE,
      dataMode = "stock"
    )
  )
  expect_lt(t3[["elapsed"]], 0.1)

  # Parsing a non-cached genome spec is slower
  t4 <- system.time(
    parseAndValidateGenomeSpec(
      genomeName = "macFas5",
      initialLocus = "all",
      stockGenome = TRUE,
      dataMode = "stock"
    )
  )
  expect_true(t4[["elapsed"]] > 0.01)
})

test_that("Parsing and validation of stock genome specs works", {
  options <- parseAndValidateGenomeSpec(
    genomeName = "hg38",
    initialLocus = "NDUFS2",
    stockGenome = TRUE,
    dataMode = "stock",
    fasta = NA,
    fastaIndex = NA,
    genomeAnnotation = NA
  )

  expect_equal(
    sort(names(options)),
    c(
      "annotation",
      "dataMode",
      "fasta",
      "fastaIndex",
      "genomeName",
      "initialLocus",
      "stockGenome",
      "validated"
    )
  )
  expect_equal(options[["genomeName"]], "hg38")
  expect_true(options[["validated"]])
  expect_true(options[["stockGenome"]])
  expect_true(all(is.na(options[c("fasta", "fastaIndex", "annotation")])))

  # Using a non-existent genome name should fail
  expect_error(suppressWarnings(
    parseAndValidateGenomeSpec(genomeName = "fubar99", initialLocus = "all")
  ))
})

test_that("Parsing and validation of custom HTTP genome specs works", {
  # Served from 127.0.0.1 rather than gladki.pl: the http code path is still
  # exercised for real, but CI no longer depends on a third-party host.
  port <- local_server()
  fasta.file <- local_url(port, "ribosomal-RNA-gene.fasta")
  fastaIndex.file <- local_url(port, "ribosomal-RNA-gene.fasta.fai")
  annotation.file <- local_url(port, "ribosomal-RNA-gene.gff3")

  options <- parseAndValidateGenomeSpec(
    genomeName = "ribo",
    initialLocus = "all",
    stockGenome = FALSE,
    dataMode = "http",
    fasta = fasta.file,
    fastaIndex = fastaIndex.file,
    genomeAnnotation = annotation.file
  )
  expect_true(options$validated)
  expect_false(options$stockGenome)
  expect_equal(options$dataMode, "http")

  # A bogus fasta file URL should fail
  expect_error({
    fasta.file.bogus <- sprintf("%s-bogus", fasta.file)
    parseAndValidateGenomeSpec(
      genomeName = "ribo-willFail",
      initialLocus = "all",
      stockGenome = FALSE,
      dataMode = "http",
      fasta = fasta.file.bogus,
      fastaIndex = fastaIndex.file,
      genomeAnnotation = annotation.file
    )
  })
})

test_that("Parsing and validation of custom local file genome specs works", {
  data.directory <- system.file(package = "igvShiny", "extdata")
  fasta.file <- file.path(data.directory, "ribosomal-RNA-gene.fasta")
  fastaIndex.file <- file.path(data.directory, "ribosomal-RNA-gene.fasta.fai")
  annotation.file <- file.path(data.directory, "ribosomal-RNA-gene.gff3")

  expect_true(file.exists(fasta.file))
  expect_true(file.exists(fastaIndex.file))
  expect_true(file.exists(annotation.file))

  options <- parseAndValidateGenomeSpec(
    genomeName = "ribosome local files",
    initialLocus = "all",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = fasta.file,
    fastaIndex = fastaIndex.file,
    genomeAnnotation = annotation.file
  )
  expect_true(options$validated)
  expect_false(options$stockGenome)
  expect_equal(options$dataMode, "localFiles")

  # A bogus file path should fail
  expect_error({
    fasta.file.bogus <- sprintf("%s-bogus", fasta.file)
    parseAndValidateGenomeSpec(
      genomeName = "ribo-willFail",
      initialLocus = "all",
      stockGenome = FALSE,
      dataMode = "localFiles",
      fasta = fasta.file.bogus,
      fastaIndex = fastaIndex.file,
      genomeAnnotation = annotation.file
    )
  })
})

test_that("Parsing SARS genome with GFF3 from local files works", {
  data.directory <- system.file(package = "igvShiny", "extdata", "sarsGenome")
  fasta.file <- file.path(data.directory, "Sars_cov_2.ASM985889v3.dna.toplevel.fa")
  fastaIndex.file <- file.path(data.directory,
                               "Sars_cov_2.ASM985889v3.dna.toplevel.fa.fai")
  annotation.file <- file.path(data.directory, "Sars_cov_2.ASM985889v3.101.gff3")

  expect_true(file.exists(fasta.file))

  title <- "SARS-CoV-2"
  options <- parseAndValidateGenomeSpec(
    genomeName = title,
    initialLocus = "all",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = fasta.file,
    fastaIndex = fastaIndex.file,
    genomeAnnotation = annotation.file
  )
  expect_true(options$validated)
  expect_equal(options$genomeName, title)
  expect_false(options$stockGenome)
})

# igvShiny — the smallest app there is: a genome, a widget, nothing else.

library(shiny)
library(igvShiny)

options <- parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "NDUFS2")

ui <- fluidPage(igvShinyOutput("igvShiny"))

server <- function(input, output, session) {
  output$igvShiny <- renderIgvShiny(igvShiny(options))
}

shinyApp(ui = ui, server = server)

# igvShiny — two browsers on one page. Each widget is addressed by its own
# element id and reports its own region: the currentGenomicRegion.<id> event
# carries the id it came from, so two widgets on the same locus both report
# it (#126).

library(shiny)
library(bslib)
library(igvShiny)

tbl.bed <- data.frame(chr = c("1", "1", "1"),
                      start = c(7432951, 7437000, 7438000),
                      end   = c(7436000, 7437500, 7440000),
                      value = c(-2.239, 3.0, 0.5),
                      sampleID = c("sample1", "sample2", "sample3"),
                      stringsAsFactors = FALSE)

browserCard <- function(id, label) {
  card(card_header(label,
                   tags$span(class = "small text-muted font-monospace ms-2",
                             textOutput(sprintf("region_%s", id), inline = TRUE))),
       card_body(class = "p-0", igvShinyOutput(id, height = "320px")))
}

ui <- page_fillable(
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  layout_columns(
    col_widths = 12,
    div(class = "d-flex gap-2",
        actionButton("addTracks", "Add a BED track to both", class = "btn-primary"),
        actionButton("moveBoth", "Send both to MEF2C", class = "btn-outline-primary")),
    browserCard("igvShiny_0", "browser 0 — hg38"),
    browserCard("igvShiny_1", "browser 1 — hg38")
  )
)

server <- function(input, output, session) {
  ids <- c("igvShiny_0", "igvShiny_1")

  observeEvent(input$addTracks, {
    for (id in ids)
      loadBedTrack(session, id = id, trackName = "bed", tbl = tbl.bed, color = "green")
  })

  observeEvent(input$moveBoth, {
    for (id in ids) showGenomicRegion(session, id = id, "MEF2C")
  })
  for (id in ids) local({
    this.id <- id
    output[[sprintf("region_%s", this.id)]] <- renderText(
      input[[sprintf("currentGenomicRegion.%s", this.id)]])
  })

  options <- parseAndValidateGenomeSpec(genomeName = "hg38",
                                        initialLocus = "chr1:7,426,231-7,453,241")
  output$igvShiny_0 <- renderIgvShiny(igvShiny(options))
  output$igvShiny_1 <- renderIgvShiny(igvShiny(options))
}

shinyApp(ui = ui, server = server)

# A static HTTP server over inst/extdata, for the code paths that take a URL.
#
# These paths used to be tested against https://gladki.pl, which made the test
# suite depend on a third-party host: CI went red whenever that host was slow or
# throttled the runner IPs, with no change in the package. Serving the same
# fixture files from 127.0.0.1 keeps the real http:// code path (httr::HEAD,
# httr::http_error) under test while making it offline-safe.
#
# httpuv comes in with shiny, so this adds no dependency.

local_server_start <- function() {
  dir <- system.file(package = "igvShiny", "extdata")
  port <- httpuv::randomPort()
  server <- httpuv::startServer(
    "127.0.0.1", port,
    # the R-side checks (httr::HEAD) do not care about CORS, but a browser
    # test does: the shiny app and this server sit on different ports, so
    # igv.js fetching a genome from here is a cross-origin request and is
    # refused without this header - silently, as a promise that never settles
    list(staticPaths = list("/" = httpuv::staticPath(
      dir,
      headers = list("Access-Control-Allow-Origin" = "*")
    )))
  )
  list(server = server, port = port)
}

# Serves for the lifetime of the calling test (or test file, when called from
# a helper), then shuts down - `withr::defer` is what testthat uses internally
# for this kind of scoped cleanup.
local_server <- function(env = parent.frame()) {
  srv <- local_server_start()
  withr::defer(srv$server$stop(), envir = env)
  srv$port
}

local_url <- function(port, filename) {
  sprintf("http://127.0.0.1:%d/%s", port, filename)
}

library(testthat)
library(igvShiny)

test_that("GWASTrack constructor works with a data.frame", {
  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))

  gwasTrack <- GWASTrack(
    "data.frame gwas",
    tbl.gwas,
    chrom.col = 3,
    pos.col = 4,
    pval.col = 10,
    trackHeight = 100
  )
  url <- getUrl(gwasTrack)
  expect_true(grepl("tracks.*\\.gwas", url))
})

test_that("GWASTrack constructor works with a remote URL", {
  # Local static server instead of gladki.pl - same http code path (the
  # constructor runs httr::http_error on the url), no third-party dependency.
  port <- local_server()
  url <- local_url(port, "gwas-5k.tsv.gz")
  gwasTrack <- GWASTrack(
    "remote url gwas",
    url,
    chrom.col = 3,
    pos.col = 4,
    pval.col = 10,
    trackHeight = 100
  )
  url.retrieved <- getUrl(gwasTrack)
  expect_equal(url, url.retrieved)
})

test_that("GWASTrack constructor rejects impossible column numbers", {
  f <- system.file(package = "igvShiny", "extdata", "gwas.RData")
  tbl.gwas <- get(load(f))

  expect_error(
    GWASTrack("bad col", tbl.gwas, chrom.col = 3, pos.col = 4, pval.col = 0),
    "at least 1"
  )
  expect_error(
    GWASTrack("bad col", tbl.gwas, chrom.col = c(3, 4), pos.col = 4,
              pval.col = 10),
    "single column number"
  )
  # a column past the end of the table would otherwise yield an empty track
  expect_error(
    GWASTrack("bad col", tbl.gwas, chrom.col = 3, pos.col = 4,
              pval.col = NCOL(tbl.gwas) + 1),
    sprintf("beyond the %d columns", NCOL(tbl.gwas))
  )
  # all three share one check, so the message has to name the guilty one
  expect_error(
    GWASTrack("bad col", tbl.gwas, chrom.col = NCOL(tbl.gwas) + 1, pos.col = 4,
              pval.col = 10),
    "chrom.col beyond"
  )
})

sanitizeColorMap <- igvShiny:::.sanitizeChromosomeColorMap

test_that("chromosomeColorMap keeps only named, single-string colors", {
  expect_equal(sanitizeColorMap(NULL), list())
  expect_equal(sanitizeColorMap(list()), list())
  expect_equal(sanitizeColorMap(c("1" = "red")), list("1" = "red"))
  expect_error(sanitizeColorMap(list("red", "blue")),
               "needs a chromosome name")
  expect_error(sanitizeColorMap(list("1" = "red", "blue")),
               "needs a chromosome name")
  expect_error(sanitizeColorMap(list("1" = c("red", "blue"))),
               "single strings")
  expect_error(sanitizeColorMap(list("1" = 42)), "single strings")
  expect_error(sanitizeColorMap(42), "named list or character")
  # JSON keeps one value per key, so a repeated chromosome loses a color
  expect_error(sanitizeColorMap(list("1" = "red", "1" = "blue")),
               "one chromosome twice")
})

test_that("GWASTrack constructor fails with illegal arguments", {
  # A url the server answers with 404, so the http-error path is tested
  # without reaching out to a real external host.
  port <- local_server()
  expect_error(
    GWASTrack(
      "bogus url",
      data = local_url(port, "does-not-exist.gwas"),
      chrom.col = 3,
      pos.col = 4,
      pval.col = 10,
      color = "darkgreen",
      trackHeight = 100
    )
  )

  expect_error(
    GWASTrack(
      "bogus data type",
      data = 42,
      chrom.col = 3,
      pos.col = 4,
      pval.col = 10,
      color = "darkgreen",
      trackHeight = 100
    )
  )
})

# igvShiny — splice junctions held in R, written as the bed igv.js reads.
#
# One row per junction, attributes in their own columns; the loader packs them
# into the bed name field igv.js filters and labels on. Coordinates are bed
# coordinates already, read off the converted bed igvteam ships as a test case
# (data/test/splice_junctions in igvteam/igv-data).

library(shiny)
library(bslib)
library(igvShiny)

junctions <- data.frame(
  chrom = "chr15",
  start = c(92883186, 92883186, 92883778, 92885584, 92886254, 92892072),
  end = c(92883584, 92885514, 92885514, 92886168, 92892003, 92897723),
  motif = "GT/AG",
  uniquely_mapped = c(56, 222, 95, 555, 415, 1202),
  multi_mapped = c(2, 9, 10, 11, 13, 10),
  maximum_spliced_alignment_overhang = c(35, 37, 38, 37, 37, 38),
  annotated_junction = TRUE,
  strand = "+",
  stringsAsFactors = FALSE
)

ui <- page_sidebar(
  title = "igvShiny — splice junctions",
  theme = bs_theme(version = 5, primary = "#2c6faa"),
  fillable = TRUE,
  sidebar = sidebar(
    width = 280,
    actionButton("addLocalJunctionsButton", "Splice junctions (data.frame)",
                 class = "w-100 mb-2"),
    actionButton("removeUserTracksButton", "Remove user tracks",
                 class = "btn-outline-danger w-100")
  ),
  card(full_screen = TRUE,
       card_body(class = "p-0", igvShinyOutput("igvShiny_0", height = "100%")))
)

server <- function(input, output, session) {
  observeEvent(input$addLocalJunctionsButton, {
    showGenomicRegion(session, id = "igvShiny_0", "chr15:92,880,000-92,900,000")
    loadSpliceJunctionTrackFromLocalData(
      session, id = "igvShiny_0", trackName = "sampleA junctions",
      tbl = junctions, trackHeight = 150,
      trackConfig = list(colorBy = "motif", labelWith = "uniqueReadCount"))
  })

  observeEvent(input$removeUserTracksButton,
               removeUserAddedTracks(session, id = "igvShiny_0"))

  output$igvShiny_0 <- renderIgvShiny({
    igvShiny(parseAndValidateGenomeSpec(genomeName = "hg38", initialLocus = "NDUFS2"))
  })
}

shinyApp(ui = ui, server = server)

test_that("every allowlisted track option is one igv.js reads", {
  # A name igv.js does not know passes .sanitizeAndMergeOptions, reaches the
  # browser and is ignored there, so nothing warns on either side. The bundled
  # build is a sufficient oracle: a minifier renames local variables but cannot
  # rename property names, which are reached by string. So a whole-identifier
  # hit in the minified file means igv.js reads that name off an object.
  libDir <- system.file("htmlwidgets", "lib", package = "igvShiny")
  bundled <- Sys.glob(file.path(libDir, "igv-*.min.js"))
  expect_length(bundled, 1L)

  # Read the bytes rather than the lines. readLines() has to decide on an
  # encoding, and on a minified bundle that decision goes differently per
  # platform: on Windows it came back short enough that every option looked
  # missing, which reads as 60-odd dead options rather than as a bad read.
  size <- file.size(bundled)
  src <- rawToChar(readBin(bundled, "raw", n = size))
  expect_equal(nchar(src, type = "bytes"), size)

  # Substring matching is not enough: "negColor" sits inside "negColorScale"
  # and "autoScaleGroup" inside the internal "autoScaleGroupColorHash", so both
  # would look alive while being read by nothing.
  reads <- function(option) {
    pattern <- sprintf("(?<![A-Za-z0-9_$])%s(?![A-Za-z0-9_$])", option)
    grepl(pattern, src, perl = TRUE, useBytes = TRUE)
  }

  options <- igvShiny:::.validIgvTrackOptions
  found <- vapply(options, reads, logical(1))
  expect_equal(options[!found], character(0))

  # The check above passes just as well when the search is broken and finds
  # everything, so pin some names igv.js 3.8.4 does not read. These are the
  # ones #159 removed, and they are what the search has to be able to reject.
  removed <- c("negColor", "posColor", "autoScaleGroup", "barChart",
               "noStroke", "noFill", "showLabels", "fontStyle", "fontWeight",
               "colorByAttribute", "hideEmptyTracks", "maxPanelHeight",
               "separateBam")
  expect_equal(removed[vapply(removed, reads, logical(1))], character(0))
})

library(testthat)
library(igvShiny)

# Edge cases in the two input sanitizers, flagged by review on #127. Both are
# pure functions, so they need no Shiny session.

sanitizeAndMergeOptions <- igvShiny:::.sanitizeAndMergeOptions
sanitizeTracks <- igvShiny:::.sanitizeTracks

test_that("a NA-named trackConfig warns and is ignored, not an error", {
  base <- list(elementID = "igvShiny_0")
  bad <- list(1)
  names(bad) <- NA_character_
  # before the fix this raised "missing value where TRUE/FALSE needed"
  expect_warning(out <- sanitizeAndMergeOptions(base, bad),
                 "must be a named list")
  expect_equal(out, base)
})

test_that("a startup track url must be a non-empty scalar string", {
  for (bad_url in list(NA_character_, "", character(0))) {
    expect_warning(out <- sanitizeTracks(list(list(name = "t", url = bad_url))),
                   "no valid 'url'")
    expect_length(out, 0L)
  }
})

test_that("a non-character startup track url is rejected", {
  expect_warning(out <- sanitizeTracks(list(list(name = "t", url = 42))),
                 "no valid 'url'")
  expect_length(out, 0L)
})

test_that("a multi-valued startup track url is rejected", {
  expect_warning(
    out <- sanitizeTracks(list(list(name = "t",
                                    url = c("https://a.bed", "https://b.bed")))),
    "no valid 'url'"
  )
  expect_length(out, 0L)
})

test_that("a startup track with no url field is dropped", {
  expect_warning(out <- sanitizeTracks(list(list(name = "t"))),
                 "no valid 'url'")
  expect_length(out, 0L)
})

test_that("a valid scalar url is still kept", {
  out <- sanitizeTracks(list(list(name = "t",
                                  url = "https://example.org/x.bed")))
  expect_length(out, 1L)
  expect_equal(out[[1]]$url, "https://example.org/x.bed")
})

# A merged track holds its members instead of a url of its own: that is how
# igv.js draws junction arcs over a coverage track in one panel (#103).

merged_track <- function(...) {
  list(type = "merged", name = "sashimi-ish",
       tracks = list(list(type = "wig", url = "https://example.org/a.bw"),
                     list(type = "junction", format = "bed",
                          url = "https://example.org/a.bed")),
       ...)
}

test_that("a merged track keeps its members and needs no url of its own", {
  out <- sanitizeTracks(list(merged_track()))
  expect_length(out, 1L)
  expect_length(out[[1]]$tracks, 2L)
  expect_equal(out[[1]]$tracks[[2]]$url, "https://example.org/a.bed")
})

test_that("member tracks are sanitized against the same allowlist", {
  track <- merged_track()
  track$tracks[[1]]$rmdir <- "/tmp"
  expect_warning(out <- sanitizeTracks(list(track)),
                 "invalid or unsupported")
  expect_null(out[[1]]$tracks[[1]]$rmdir)
  expect_equal(out[[1]]$tracks[[1]]$url, "https://example.org/a.bw")
})

test_that("a merged track left with no usable member is dropped", {
  track <- merged_track()
  track$tracks <- list(list(type = "wig", url = NA_character_))
  # two warnings: the member goes first, and the empty parent follows it
  warnings <- capture_warnings(out <- sanitizeTracks(list(track)))
  expect_match(warnings, "no valid 'url'", all = FALSE)
  expect_match(warnings, "no usable member", all = FALSE)
  expect_length(out, 0L)

  expect_warning(out <- sanitizeTracks(list(list(type = "merged", name = "m"))),
                 "no usable member")
  expect_length(out, 0L)
})

test_that("the merged type is matched the way igv.js matches it", {
  # igv.js dispatches on config.type.toLowerCase(). The field is overwritten
  # rather than passed through ..., which would append a second `type` that
  # track[["type"]] never reaches, and the assertion would hold either way.
  track <- merged_track()
  track$type <- "Merged"
  out <- sanitizeTracks(list(track))
  expect_length(out, 1L)
  expect_length(out[[1]]$tracks, 2L)
})

test_that("'tracks' stays unsupported on anything but a merged track", {
  expect_warning(
    out <- sanitizeTracks(list(list(type = "wig",
                                    url = "https://example.org/a.bw",
                                    tracks = list(list(url = "x"))))),
    "invalid or unsupported"
  )
  expect_length(out, 1L)
  expect_null(out[[1]]$tracks)
})

test_that("merged nesting is capped", {
  deep <- merged_track()
  for (i in seq_len(3)) {
    deep <- list(type = "merged", tracks = list(deep))
  }
  expect_warning(out <- sanitizeTracks(list(deep)), "nested more than")
  expect_length(out, 0L)
})

test_that("sanitized tracks carry no names, so they serialize as a JSON array", {
  # igv.js walks a merged track with for...of and calls .every() on the same
  # value; a named R list would reach it as an object and break both
  track <- merged_track()
  names(track$tracks) <- c("coverage", "junctions")
  out <- sanitizeTracks(list(one = track))
  expect_null(names(out))
  expect_null(names(out[[1]]$tracks))
})

library(testthat)
library(igvShiny)

# trackConfig is user-supplied and goes straight into an igv.js track object, so
# the allowlist in .sanitizeAndMergeOptions is a security boundary as much as a
# convenience. .sanitizeTracks does the same job for the startup `tracks`
# argument (the regression fixed in #36). Neither had any test.

sanitizeAndMergeOptions <- igvShiny:::.sanitizeAndMergeOptions
sanitizeTracks <- igvShiny:::.sanitizeTracks

test_that("an empty or absent trackConfig leaves the base options alone", {
  base <- list(elementID = "igvShiny_0", trackName = "t")
  expect_equal(sanitizeAndMergeOptions(base, list()), base)
  expect_equal(sanitizeAndMergeOptions(base, NULL), base)
})

test_that("valid igv.js options are merged in", {
  base <- list(elementID = "igvShiny_0", trackName = "t")
  merged <- sanitizeAndMergeOptions(base, list(height = 120, autoscale = TRUE))
  expect_equal(merged$height, 120)
  expect_true(merged$autoscale)
  expect_equal(merged$trackName, "t")
})

test_that("a gwas column mapping reaches igv.js through trackConfig", {
  base <- list(elementID = "igvShiny_0", trackName = "t")
  columns <- list(chromosome = 3, position = 4, value = 10)
  merged <- sanitizeAndMergeOptions(base, list(columns = columns))
  expect_equal(merged$columns, columns)
})

test_that("options outside the allowlist are dropped with a warning", {
  base <- list(elementID = "igvShiny_0")
  expect_warning(
    merged <- sanitizeAndMergeOptions(base, list(onClick = "alert(1)")),
    "invalid or unsupported"
  )
  expect_null(merged$onClick)
})

test_that("user options never override the function's own arguments", {
  base <- list(elementID = "igvShiny_0", trackName = "mine", color = "blue")
  expect_warning(
    merged <- sanitizeAndMergeOptions(base, list(trackName = "theirs")),
    "conflict with function arguments"
  )
  expect_equal(merged$trackName, "mine")
})

test_that("a trackConfig that is not a named list is refused", {
  base <- list(elementID = "igvShiny_0")
  expect_warning(merged <- sanitizeAndMergeOptions(base, list(1, 2)),
                 "must be a named list")
  expect_equal(merged, base)
})

test_that("autoscaleGroup survives as both number and string (#105)", {
  base <- list(elementID = "igvShiny_0")
  expect_equal(sanitizeAndMergeOptions(base, list(autoscaleGroup = 3))$autoscaleGroup, 3)
  expect_equal(sanitizeAndMergeOptions(base, list(autoscaleGroup = "grp"))$autoscaleGroup, "grp")
})

test_that("startup tracks keep their valid keys and lose the invalid ones", {
  tracks <- list(list(name = "genes", type = "annotation",
                      url = "https://example.org/g.gff3", bogusKey = 1))
  expect_warning(out <- sanitizeTracks(tracks), "invalid or unsupported")
  expect_length(out, 1L)
  expect_null(out[[1]]$bogusKey)
  expect_equal(out[[1]]$name, "genes")
})

test_that("a startup track without a url is dropped", {
  expect_warning(out <- sanitizeTracks(list(list(name = "no url"))),
                 "no valid 'url'")
  expect_length(out, 0L)
})

test_that("an unnamed startup track entry is dropped", {
  expect_warning(out <- sanitizeTracks(list(list("a", "b"))),
                 "must be a named list")
  expect_length(out, 0L)
})

test_that("an empty tracks argument yields an empty list, not NULL (#36)", {
  expect_equal(sanitizeTracks(list()), list())
  expect_equal(sanitizeTracks(NULL), list())
})

test_that("the human stock genomes keep their assets off hgdownload", {
  # Requesting hg19 or hg38 by bare id hands igv.js the genomes3.json entry,
  # where sequence, cytobands and the RefSeq annotation all sit on
  # hgdownload.soe.ucsc.edu: a host that throttles to tens of seconds per range
  # request, serving an unindexed whole-genome annotation on top of it. That is
  # a browser which looks hung on startup, so igvShiny.js pins its own
  # reference for the two of them. A revert to the bare id is silent - it costs
  # startup time, never correctness - so assert the pin here.
  binding <- system.file("htmlwidgets", "igvShiny.js", package = "igvShiny")
  expect_true(file.exists(binding))
  js <- paste(readLines(binding, warn = FALSE), collapse = "\n")

  pinned <- sub(".*function pinnedReference\\(", "", js)
  pinned <- sub("\\n\\}.*", "", pinned)

  expect_match(pinned, "hg19:", fixed = TRUE)
  expect_match(pinned, "hg38:", fixed = TRUE)
  expect_false(grepl("hgdownload.soe.ucsc.edu", pinned, fixed = TRUE))

  # the annotation has to be the tabix-indexed build: the plain one is
  # whole-genome, and igv.js reads all of it before drawing the first gene
  expect_equal(lengths(regmatches(pinned, gregexpr("indexURL", pinned)))[1], 1L)
  expect_match(pinned, "indexed: true", fixed = TRUE)

  # twoBitURL has to be spelled out for both, and this is the assertion that
  # earns its keep: igv.js merges the registry entry by id, so a reference
  # naming only a fasta keeps the registry's twoBitURL and the sequence comes
  # from UCSC after all. getSequence() then never resolves and every alignment
  # track spins forever, while the gene track still draws - so the symptom
  # points nowhere near the cause
  twoBit <- regmatches(pinned, gregexpr("twoBitURL: \"[^\"]+\"", pinned))[[1]]
  expect_length(twoBit, 2L)
  expect_true(all(grepl("https://igv.org/genomes/data/", twoBit, fixed = TRUE)))
})

---
title: "igvShiny: a wrapper of IGV in the Shiny apps"
author:
- name: Arkadiusz Gladki
package: igvShiny
output:
  BiocStyle::html_document:
    toc: false
vignette: >
  %\VignetteIndexEntry{igvShiny Overview}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
    )
```

# igvShiny

The `igvShiny` package is a wrapper of Integrative Genomics Viewer (IGV). 
It comprises an htmlwidget version of IGV. 
It can be used as a module in Shiny apps. 

# Installation

```{r,eval=FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("igvShiny")
```

# Loading the package

```{r,include=TRUE,results="hide",message=FALSE,warning=FALSE}
library(igvShiny)
```

# Running minimal Shiny app

Running the minimal Shiny app with igvShiny is as simple as:

* providing genome details via `parseAndValidateGenomeSpec`
* using `igvShinyOutput` as the UI function
* using `renderIgvShiny` as the server function

The he minimal Shiny app make look like: 

```{r,eval=FALSE}
library(igvShiny)
options <- parseAndValidateGenomeSpec(genomeName="hg38",  initialLocus="NDUFS2")
if (interactive()) {
    ui <- shinyUI(fluidPage(igvShinyOutput('igvShiny'), width = 10))
    server <-
        function(input, output, session) {
            output$igvShiny <- renderIgvShiny({
                igvShiny(options)
            })
            }
    runApp(shinyApp(ui = ui, server = server))
    }
```

# Providing genome details

Multiple genomes are currently supported by IGV:
[link](https://s3.amazonaws.com/igv.org.genomes/genomes.json). 
In igvShiny  this set of genomes is called stock genomes.
One can select any stock genome easily by running `parseAndValidateGenomeSpec`
with single `genomeName` value properly assigned.
For example to use the most popular mouse genome one need to run:

```{r,echo=TRUE}
igvShiny::parseAndValidateGenomeSpec("mm10")
```

The list of available stock genomes in igvShiny can be found with:
```{r,echo=TRUE}
igvShiny::get_css_genomes()
```

See also demo app for stock genomes when one can select genome of interest 
and familiarize with the basic functionalities provided via igvShiny:
```{r,echo=TRUE}
library(igvShiny)
demo_app_file <- system.file(package= "igvShiny", "demos", "stockGenomesDemo.R")
    if (interactive()) {
        shiny::runApp(demo_app_file)
        }
```

It's also possible to use custom genomes (i.e. non-stock genomes). 
In such cases  one has to provide data for: 
`FASTA file`, `FASTA index file` and `genome annotation file`.
The files can provided as local paths (`dataMode` = `localFiles` or via URLs
(`dataMode = http`). See below the examples for both cases.

```{r,echo=TRUE}
library(igvShiny)

# defining custom genome with data provided via URLs
base_url <- "https://gladki.pl/igvr/testFiles"
title <- "ribo remote"
fasta_file <- sprintf("%s/%s", base_url, "ribosomal-RNA-gene.fasta")
fastaIndex_file <- sprintf("%s/%s", base_url, "ribosomal-RNA-gene.fasta.fai")
annotation_file <- sprintf("%s/%s", base_url, "ribosomal-RNA-gene.gff3")
locus <- "U13369.1:7,276-8,225"
genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = title,
    initialLocus = locus,
    stockGenome = FALSE,
    dataMode = "http",
    fasta = fasta_file,
    fastaIndex = fastaIndex_file,
    genomeAnnotation = annotation_file
    )
genomeOptions

# defining custom genome with data provided with local files
data_directory <- system.file(package = "igvShiny", "extdata")
fasta_file      <- file.path(data_directory, "ribosomal-RNA-gene.fasta")
fastaIndex_file <- file.path(data_directory, "ribosomal-RNA-gene.fasta.fai")
annotation_file <- file.path(data_directory, "ribosomal-RNA-gene.gff3")
genomeOptions2 <- parseAndValidateGenomeSpec(
    genomeName = "ribo local",
    initialLocus = "U13369.1:7,276-8,225",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = fasta_file,
    fastaIndex = fastaIndex_file,
    genomeAnnotation = annotation_file
    )
genomeOptions2
```

See also demo apps for custom genomes with data provided via URLs:
```{r,echo=TRUE}
library(igvShiny)
demo_app_file <- system.file(
    package = "igvShiny", 
    "demos", 
    "igvShinyDemo-customGenome-http.R"
    )
if (interactive()) {
    shiny::runApp(demo_app_file)
    }
```
as well as data provided via local files:
```{r,echo=TRUE}
library(igvShiny)
demo_app_file <-
    system.file(
        package = "igvShiny", 
        "demos", 
        "igvShinyDemo-customGenome-localFiles.R")
if (interactive()) {
    shiny::runApp(demo_app_file)
    }
```


# Main functionalities

In principle igvShiny provides the same functionalities that one 
can find in igv.js. In summary following actions are currently possible 
(wrapped as the R helpers):

* load tracks with the following formats/data types:
    + Bed
    + BedGraph
    + Seg
    + GWAS
    + Bam
    + Cram
    + Vcf
    + GFF3
* showGenomicRegion (zoom in or out to show the nominated region, 
by chromosome locus or gene symbol)
* getGenomicRegion (return the current IGV region)
* removeUserAddedTracks (remove tracks from the browser, 
added during the session by the user)

See also demo app to familiarize with the basic functionalities 
provided by igvShiny:
```{r,echo=TRUE}
library(igvShiny)
demo_app_file <- system.file(package= "igvShiny", "demos", "igvShinyDemo.R")
if (interactive()) {
    shiny::runApp(demo_app_file)
    }
```

# Session Info

```{r}
sessionInfo()
```

---
title: "Getting started with igvShiny"
author:
- name: Arkadiusz Gladki
package: igvShiny
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Getting started with igvShiny}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
    )
```

# Who this is for

You have a Shiny app and some genomic data — intervals, alignments, variants,
association statistics — and you want to look at it in a genome browser without
leaving R. This vignette takes you from installation to a working app with a
track loaded from your own `data.frame`.

It assumes you have written a Shiny app before. It does not assume you know
[igv.js](https://github.com/igvteam/igv.js), the JavaScript browser `igvShiny`
wraps.

For the complete API — every track type and the options each one accepts — see
the *Track options reference* vignette. For a live app you can click through
before installing anything, see the
[demo on Posit Connect Cloud][live-demo].

[live-demo]: https://gladkia-igvshiny-demo.share.connect.posit.cloud

# Install

```{r install, eval=FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("igvShiny")
```

```{r load-package, results="hide", message=FALSE, warning=FALSE}
library(igvShiny)
```

# The three moving parts

An `igvShiny` browser is an htmlwidget, so it follows the usual Shiny pattern of
an output function in the UI and a render function on the server. What is
specific to this package is the third part: the browser has to be told which
genome to display *before* it is created.

| Part | Function | Where |
|------|----------|-------|
| Genome specification | `parseAndValidateGenomeSpec()` | top of the app |
| Placeholder in the layout | `igvShinyOutput()` | UI |
| The widget itself | `renderIgvShiny()` + `igvShiny()` | server |

`parseAndValidateGenomeSpec()` returns a validated list. Building it separately
means a bad genome name fails immediately, with a readable message, rather than
producing an empty browser in the page:

```{r genome-spec}
genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = "hg38",
    initialLocus = "NDUFS2"
    )
str(genomeOptions)
```

`initialLocus` accepts a gene symbol, a `chrom:start-end` string, or `"all"` for
the whole-genome view.

# A complete app

Everything above assembled — this is the smallest useful `igvShiny` app:

```{r minimal-app, eval=FALSE}
library(shiny)
library(igvShiny)

genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = "hg38",
    initialLocus = "NDUFS2"
    )

ui <- fluidPage(
    titlePanel("igvShiny"),
    igvShinyOutput("igv")
    )

server <- function(input, output, session) {
    output$igv <- renderIgvShiny({
        igvShiny(genomeOptions)
        })
    }

shinyApp(ui, server)
```

Two notes on sizing. `igvShinyOutput()` takes `height` as an explicit pixel
measure (`"800px"`); percentages do not work, because the browser needs a
concrete height to lay its panels out. And the widget only draws once its
container is visible, so if you put it behind a `tabsetPanel()` tab, expect it
to appear when that tab is first opened.

# Loading a track from your own data

Tracks are added by sending a message to a browser that already exists, so the
loaders are called from the server, take `session` and the widget `id`, and are
usually wired to an event:

```{r bed-track, eval=FALSE}
tbl <- data.frame(
    chrom = c("chr1", "chr1", "chr1"),
    start = c(7432000, 7437000, 7443000),
    end   = c(7436000, 7442000, 7447000),
    value = c(0.2, 0.9, 0.4),
    stringsAsFactors = FALSE
    )

server <- function(input, output, session) {
    output$igv <- renderIgvShiny({
        igvShiny(genomeOptions)
        })

    observeEvent(input$addTrack, {
        loadBedTrack(
            session,
            id = "igv",
            trackName = "my regions",
            tbl = tbl,
            color = "darkblue"
            )
        })
    }
```

`loadBedTrack()` needs `chrom`, `start` and `end` columns; anything else is
ignored. The other loaders follow the same shape — `loadBedGraphTrack()`,
`loadSegTrack()`, `loadVcfTrack()`, `loadGwasTrack()`,
`loadBamTrackFromLocalData()`, `loadSpliceJunctionTrackFromLocalData()` — and
each has a `*FromURL()` counterpart for data that already lives on a web
server:

```{r url-track, eval=FALSE}
gff3 <- paste0(
    "https://s3.amazonaws.com/igv.org.genomes/hg38/",
    "Homo_sapiens.GRCh38.94.chr.gff3.gz"
    )

loadGFF3TrackFromURL(
    session,
    id = "igv",
    trackName = "genes",
    gff3URL = gff3,
    indexURL = paste0(gff3, ".tbi"),
    color = "darkgreen",
    trackHeight = 100
    )
```

The `From URL` variants ask the *user's browser* to fetch the file, not R. A URL
that works in `download.file()` can still fail here if the server does not send
permissive CORS headers.

Tracks the user added during a session can be cleared with
`removeUserAddedTracks(session, id = "igv")`, or individually by name with
`removeTracksByName()`.

## Loading tracks at startup

Tracks that should always be present do not need an event. Pass them to
`igvShiny()` through `tracks`, as a list of igv.js track configurations:

```{r startup-tracks, eval=FALSE}
output$igv <- renderIgvShiny({
    igvShiny(
        genomeOptions,
        tracks = list(
            list(
                name = "genes",
                type = "annotation",
                format = "gff3",
                url = paste0(
                    "https://s3.amazonaws.com/igv.org.genomes/hg38/",
                    "Homo_sapiens.GRCh38.94.chr.gff3.gz"
                    ),
                indexed = FALSE
                )
            )
        )
    })
```

Keys that igv.js does not recognise are dropped with a warning rather than
passed through, so a typo produces a message instead of a silently empty track.

## Drawing two tracks in one panel

A `merged` track holds other tracks instead of a url of its own, and draws them
on top of each other in a single panel. Splice junction arcs over the coverage
they were spliced out of — the view sashimi plots are wanted for — is the case
it is worth knowing:

```{r merged-track, eval=FALSE}
sample <- paste0(
    "https://raw.githubusercontent.com/igvteam/igv-data/main/data/test/",
    "splice_junctions/splice_junction_track_test_cases_sampleA.",
    "chr15-92835700-93031800"
    )

output$igv <- renderIgvShiny({
    igvShiny(
        genomeOptions,
        tracks = list(
            list(
                name = "sampleA",
                type = "merged",
                height = 180,
                tracks = list(
                    list(
                        type = "wig",
                        format = "bigwig",
                        url = paste0(sample, ".bigWig"),
                        color = "rgb(190,190,190)"
                        ),
                    list(
                        type = "junction",
                        format = "bed",
                        url = paste0(sample, ".SJ.out.bed.gz"),
                        indexURL = paste0(sample, ".SJ.out.bed.gz.tbi"),
                        colorBy = "motif"
                        )
                    )
                )
            )
        )
    })
```

Member tracks go through the same option allowlist as any other track. A member
left with no usable url is dropped on its own, with a warning; the merged track
itself is dropped only once no member survives, rather than sent on empty, as
igv.js throws while building the browser if it is handed one.

# Moving around, and knowing where you are

To drive the browser from R, use `showGenomicRegion()`:

```{r navigate, eval=FALSE}
observeEvent(input$goto, {
    showGenomicRegion(
        session,
        id = "igv",
        region = "chr5:88,700,000-88,800,000"
        )
    })
```

Reading the position back is asynchronous, which is the one part of the API that
surprises people. `getGenomicRegion()` does not return the region; it asks the
browser for it, and the answer arrives later as a Shiny input named
`currentGenomicRegion.<id>`:

```{r read-region, eval=FALSE}
observeEvent(input$whereAmI, {
    getGenomicRegion(session, id = "igv")
    })

observeEvent(input[["currentGenomicRegion.igv"]], {
    message("now showing: ", input[["currentGenomicRegion.igv"]])
    })
```

The same input fires whenever the user pans or zooms, so an observer on it is
enough if you only want to follow the view rather than poll it.

Two other inputs are available: `igvReady`, which fires once the browser has
finished initialising — the right place to load tracks automatically — and
`trackClick`, which carries the feature the user clicked on.

# Inside a Shiny module

`igvShiny` works in modules with no extra arguments: the widget reads the
namespace from the session it is created in. The one thing to remember is that
the loaders address the browser by its *HTML element* id, not by the input name
— so they need the namespaced id, `ns("igv")`, and not `"igv"`:

```{r module, eval=FALSE}
igvModuleUI <- function(id) {
    ns <- NS(id)
    tagList(
        actionButton(ns("addTrack"), "Add track"),
        igvShinyOutput(ns("igv"))
        )
    }

igvModuleServer <- function(id, genomeOptions, tbl) {
    moduleServer(id, function(input, output, session) {
        ns <- session$ns

        output$igv <- renderIgvShiny({
            igvShiny(genomeOptions)
            })

        observeEvent(input$addTrack, {
            loadBedTrack(
                session,
                id = ns("igv"),
                trackName = "regions",
                tbl = tbl
                )
            })

        # the region input keeps its unnamespaced name inside the module
        observeEvent(input[["currentGenomicRegion.igv"]], {
            message("now showing: ", input[["currentGenomicRegion.igv"]])
            })
        })
    }
```

That observer follows the user panning and zooming. Asking for the region
explicitly with `getGenomicRegion()` is currently unreliable inside a module
unless the module id happens to be `"igv"` — see
[issue #134](https://github.com/gladkia/igvShiny/issues/134).

A runnable version is in `inst/demos/modules.R`.

# Stock genomes

Any genome igv.js knows can be requested by name — `get_css_genomes()` lists
them, reading the same registry the bundled igv.js does
([genomes3.json](https://igv.org/genomes/genomes3.json)).

Naming a stock genome means R fetches no sequence data: it reads the registry
only to validate the name, and the user's browser streams the sequence from
wherever the registry entry points. Since
igv.js 3.x that is a `.2bit` file, preferred over the entry's FASTA when both
are listed, and for the genomes igvShiny offers those `.2bit` files are hosted
by UCSC (`hgdownload.soe.ucsc.edu`). A UCSC outage therefore leaves a stock
genome blank, with 403s or timeouts in the JavaScript console — the package
itself is fine, and custom genomes (below) are unaffected because they carry
their own URLs.

# Custom genomes

For a genome outside the registry, supply the sequence yourself: a FASTA file,
its index, and optionally an annotation, either as local paths
(`dataMode = "localFiles"`) or URLs (`dataMode = "http"`). The package ships a
small example genome:

```{r custom-genome}
data_directory <- system.file(package = "igvShiny", "extdata")

customOptions <- parseAndValidateGenomeSpec(
    genomeName = "ribosomal RNA gene",
    initialLocus = "U13369.1:7,276-8,225",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = file.path(data_directory, "ribosomal-RNA-gene.fasta"),
    fastaIndex = file.path(data_directory, "ribosomal-RNA-gene.fasta.fai"),
    genomeAnnotation = file.path(data_directory, "ribosomal-RNA-gene.gff3")
    )
customOptions[c("genomeName", "dataMode", "stockGenome")]
```

The result goes to `igvShiny()` exactly like a stock genome specification.

The same genome served over HTTP instead, which is what you want once the app
runs somewhere other than your laptop. `dataMode = "http"` makes
`parseAndValidateGenomeSpec()` send a `HEAD` request to every URL, so this
chunk is shown rather than run — building the vignette stays offline:

```{r custom-genome-http, eval=FALSE}
base_url <- "https://gladki.pl/igvr/testFiles"

remoteOptions <- parseAndValidateGenomeSpec(
    genomeName = "ribosomal RNA gene",
    initialLocus = "U13369.1:7,276-8,225",
    stockGenome = FALSE,
    dataMode = "http",
    fasta = paste0(base_url, "/ribosomal-RNA-gene.fasta"),
    fastaIndex = paste0(base_url, "/ribosomal-RNA-gene.fasta.fai"),
    genomeAnnotation = paste0(base_url, "/ribosomal-RNA-gene.gff3")
    )
remoteOptions[c("genomeName", "dataMode", "stockGenome")]
```

`inst/demos/genomes.R` shows both variants side by side, picked with radio
buttons.

# When something does not show up

**The browser is blank.** Almost always the genome, not the track. Check the
JavaScript console: a 404 on the FASTA or its index leaves an empty browser.
Some upstream genome URLs have broken in the past without any change on our
side.

**A track loads but is empty.** Chromosome naming. If your data says `1` and the
genome says `chr1`, igv.js finds nothing and reports nothing. Compare one
interval against the locus shown in the search box.

**Nothing at all happens after a loader call.** The loaders are one-way messages
to the browser: no error comes back to R when they fail. Call the loader with
`quiet = FALSE` to log what was sent, and read the JavaScript console for what
happened to it.

**Local data files 404.** Tracks built from R data are written to a temporary
directory that Shiny serves. If you have set `TRACKS_DIR`, make sure the process
can write there; `get_tracks_dir()` reports the directory in use.

# Where to go next

* The *Track options reference* vignette — every track type and its options
* `inst/demos/` — one runnable app per feature, each under sixty lines
* `inst/showcase/igvShinyDemo.R` — most of the API in a single app, and what
  the public Posit Connect demo serves
* [igv.js documentation](https://github.com/igvteam/igv.js/wiki) — the
  underlying browser, useful when you need a track option this package passes
  through but does not document itself

# Session Info

```{r session-info}
sessionInfo()
```

---
title: "Track options reference"
author:
- name: Arkadiusz Gladki
package: igvShiny
output:
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Track options reference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
    )
```

```{r load-package, results="hide", message=FALSE, warning=FALSE}
library(igvShiny)
futile.logger::flog.threshold(futile.logger::WARN)
```

```{r stub-session, include = FALSE}
# The loaders are one-way messages to a browser: they take a shiny session and
# send to it, and there is no browser here. A session that prints instead of
# sending lets every example below run, and shows what igv.js receives.
session <- list(sendCustomMessage = function(type, message) {
    flatten <- function(v) paste(v, collapse = ",")
    fields <- vapply(message, flatten, character(1))
    cat(type, "\n", paste0("  ", names(fields), ": ", fields, collapse = "\n"),
        "\n", sep = "")
    })
```

# What this vignette is

Every track type `igvShiny` can load, and the options each one accepts. If you
are looking for how to get a browser onto the page in the first place, start
with the *Getting started with igvShiny* vignette instead.

The options are igv.js options: `igvShiny` passes them through rather than
reinventing them, and the names below are the ones the bundled igv.js
(currently 3.8.4) actually reads.

# How an option reaches the browser

Every loader takes a `trackConfig` argument: a named list merged into the track
configuration sent to igv.js.

The examples here run against a stand-in `session` that prints the message
instead of sending it to a browser, so you can see what igv.js is handed —
including the defaults the loader fills in:

```{r trackconfig}
tbl <- data.frame(
    chrom = "chr1",
    start = 7432000,
    end = 7436000
    )

loadBedTrack(
    session,
    id = "igv",
    trackName = "my regions",
    tbl = tbl,
    trackConfig = list(displayMode = "EXPANDED", maxRows = 20)
    )
```

The frequently used options are also plain arguments on the loaders that need
them — `color`, `trackHeight`, `displayMode` and so on. Both routes end up in
the same place, so use whichever reads better. Giving the same option twice is
not an error: the explicit argument wins, and you are told so.

Keys are checked against an allowlist before they are sent, and a key that is
not on it is dropped. Startup tracks go through the same check, and
`igvShiny()` builds outside a Shiny session, so the behaviour is visible here:

```{r rejected-key}
genomeOptions <- parseAndValidateGenomeSpec(
    genomeName = "hg38",
    initialLocus = "NDUFS2"
    )

widget <- igvShiny(
    genomeOptions,
    tracks = list(
        list(
            name = "genes",
            type = "annotation",
            url = "https://example.org/genes.bed",
            onClick = "alert('hi')"
            )
        )
    )

names(widget$x$tracks[[1]])
```

That is the mechanism, not a nuisance: a track configuration is a JavaScript
object built from user input, and only names known to be data options are let
through. The consequence to remember is that a typo does not reach igv.js and
does not raise an error — it produces a warning and a track drawn with
defaults.

# Loaders

| Function | Data comes from | igv.js track |
|---|---|---|
| `loadBedTrack()` | `data.frame` | annotation |
| `loadBedGraphTrack()` | `data.frame` | wig |
| `loadBedGraphTrackFromURL()` | URL | wig |
| `loadSegTrack()` | `data.frame` | seg |
| `loadGwasTrack()` | `data.frame` | gwas |
| `loadVcfTrack()` | `VCF` object | variant |
| `loadGFF3TrackFromLocalData()` | `data.frame` | annotation |
| `loadGFF3TrackFromURL()` | URL | annotation |
| `loadBamTrackFromLocalData()` | `GAlignments` | alignment |
| `loadBamTrackFromURL()` | URL | alignment |
| `loadCramTrackFromLocalData()` | file on disk | alignment |
| `loadCramTrackFromURL()` | URL | alignment |
| `loadSpliceJunctionTrackFromURL()` | URL | junction |

The `FromURL` variants ask the *user's browser* to fetch the file. A URL that
works in `download.file()` can still fail here if the server sends no
permissive CORS header — see the troubleshooting section of the getting started
vignette.

# Options every track accepts

| Option | Type | What it does |
|---|---|---|
| `name` | character | track label in the left panel |
| `url` | character | where the data is |
| `indexURL` | character | index for a bgzipped or binary file |
| `indexed` | logical | set `FALSE` to read an unindexed file whole |
| `format` | character | `"bed"`, `"gff3"`, `"bigwig"`, `"vcf"`, … |
| `type` | character | igv.js track class; usually implied by the loader |
| `order` | numeric | position among the other tracks |
| `height` | numeric | track height in pixels |
| `minHeight`, `maxHeight` | numeric | bounds when the track resizes itself |
| `autoHeight` | logical | grow to fit the features in view |
| `visibilityWindow` | numeric | above this span in bp, draw nothing |
| `removable` | logical | whether the user may close the track |
| `color` | character | any CSS colour |
| `altColor` | character | second colour, per track type |
| `displayMode` | character | `"COLLAPSED"`, `"EXPANDED"`, `"SQUISHED"` |
| `roi` | list | regions of interest drawn over this track |
| `oauthToken`, `headers` | character, list | sent with the data request |

`trackHeight` on the loaders and `height` in `trackConfig` are the same thing.

# Numeric tracks: bedGraph, wig, bigWig

| Option | Type | What it does |
|---|---|---|
| `autoscale` | logical | rescale to the data in view |
| `autoscaleGroup` | character | scale several tracks together, as one group |
| `min`, `max` | numeric | fixed data range when not autoscaling |
| `logScale` | logical | log the y axis |
| `graphType` | character | `"bar"` or `"points"` |
| `flipAxis` | logical | draw the y axis upside down |
| `color`, `altColor` | character | values above and below the baseline |

```{r wig}
coverage <- data.frame(
    chrom = "chr1",
    start = c(7432000, 7437000),
    end = c(7436000, 7442000),
    value = c(0.2, 0.9)
    )

loadBedGraphTrack(
    session, id = "igv", trackName = "coverage", tbl = coverage,
    autoscale = TRUE,
    trackConfig = list(
        graphType = "points",
        autoscaleGroup = "sampleA",
        logScale = FALSE
        )
    )
```

Two tracks sharing an `autoscaleGroup` are drawn on one scale, which is what
makes their heights comparable by eye.

# Annotation tracks: bed, gff3

| Option | Type | What it does |
|---|---|---|
| `colorBy` | character | feature attribute to take the colour from |
| `colorTable` | list | attribute value to colour |
| `featureHeight` | numeric | height of one feature row |
| `maxRows` | numeric | rows drawn before the rest are hidden |
| `searchable` | logical | let the locus box find features by name |
| `queryable` | logical | whether the track answers region queries |

On `loadGFF3TrackFromURL()` and `loadGFF3TrackFromLocalData()` the argument is
called `colorByAttribute`, and it becomes igv.js `colorBy`. Pass it as the
loader argument rather than in `trackConfig`, together with `colorTable`:

```{r gff3}
gff3.url <- paste0(
    "https://s3.amazonaws.com/igv.org.genomes/hg38/",
    "Homo_sapiens.GRCh38.94.chr.gff3.gz"
    )

loadGFF3TrackFromURL(
    session, id = "igv", trackName = "genes",
    gff3URL = gff3.url, indexURL = paste0(gff3.url, ".tbi"),
    colorByAttribute = "biotype",
    displayMode = "EXPANDED",
    visibilityWindow = 1000000,
    colorTable = list(
        protein_coding = "darkgreen",
        processed_transcript = "blue",
        default = "black"
        )
    )
```

A `default` entry in `colorTable` catches the values you did not list.

# Alignment tracks: bam, cram

| Option | Type | What it does |
|---|---|---|
| `showAllBases` | logical | draw every base, not just mismatches |
| `viewAsPairs` | logical | join mates on one row |
| `colorBy` | character | `"strand"`, `"firstOfPairStrand"`, `"tag"`, … |
| `colorTable` | list | value to colour, for `colorBy = "tag"` |
| `sort` | list | how reads are sorted when the track loads |
| `samplingWindowSize` | numeric | window the downsampler works over |
| `samplingDepth` | numeric | reads kept per window |

`sort` takes the same object the igv.js right-click menu builds, so sorting by
a tag at load time is:

```{r bam-sort}
bam.url <- "https://1000genomes.s3.amazonaws.com/phase3/NA12878.bam"

loadBamTrackFromURL(
    session, id = "igv", trackName = "reads",
    bamURL = bam.url, indexURL = paste0(bam.url, ".bai"),
    trackConfig = list(
        sort = list(
            option = "TAG", tag = "HP",
            chr = "chr8", position = 128750000
            )
        )
    )
```

Alignment tracks are the memory-hungry ones. `visibilityWindow` matters here
more than anywhere else: without it a user who zooms out asks the browser for
every read on the chromosome.

# Variant tracks: vcf

| Option | Type | What it does |
|---|---|---|
| `colorBy` | character | variant attribute to colour by |
| `colorTable` | list | attribute value to colour |
| `maxRows` | numeric | genotype rows drawn |
| `sort` | list | initial sort of the genotype rows |
| `displayMode` | character | `"COLLAPSED"` hides the genotypes |

# GWAS tracks

| Option | Type | What it does |
|---|---|---|
| `trait` | character | column holding the trait name |
| `columns` | list | **1-based** column positions in the source file |
| `min`, `max` | numeric | y range, in -log10(p) |
| `autoscale` | logical | rescale to the data in view |
| `colorTable` | list | chromosome to colour |

`columns` is worth its own note. The GWAS parser reads it 1-based, and without
it the parser guesses the layout from header names it recognises — any other
spelling draws an empty track rather than an error:

```{r gwas}
gwas <- data.frame(
    CHR = "chr1",
    BP = c(1000, 2000),
    SNP = c("rs1", "rs2"),
    P = c(1e-8, 1e-5)
    )

loadGwasTrack(
    session, id = "igv", trackName = "gwas", tbl.gwas = gwas,
    trackConfig = list(
        columns = list(chromosome = 12, position = 13, value = 28)
        )
    )
```

The `GWASTrack` class is the other way into the same track type, for data that
already lives at a URL; see `?GWASTrack`.

# Splice junction tracks

Junction options divide into filters, which decide whether an arc is drawn at
all, and appearance.

Filters:

| Option | Type | What it does |
|---|---|---|
| `minUniquelyMappedReads` | numeric | drop junctions below this count |
| `minTotalReads` | numeric | unique plus multi-mapped |
| `maxFractionMultiMappedReads` | numeric | drop mostly multi-mapped junctions |
| `minSplicedAlignmentOverhang` | numeric | shortest anchor accepted |
| `minJunctionEndsVisible` | numeric | 0, 1 or 2 ends in view |
| `minSamplesWithThisJunction` | numeric | across samples |
| `maxSamplesWithThisJunction` | numeric | across samples |
| `minPercentSamplesWithThisJunction` | numeric | as a percentage |
| `maxPercentSamplesWithThisJunction` | numeric | as a percentage |
| `hideAnnotatedJunctions` | logical | draw novel junctions only |
| `hideUnannotatedJunctions` | logical | draw annotated junctions only |
| `hideMotifs` | character | motifs to leave out, e.g. `c("GT/AG")` |
| `hideStrand` | character | `"+"` or `"-"` |

Appearance:

| Option | Type | What it does |
|---|---|---|
| `thicknessBasedOn` | character | what sets the arc thickness |
| `bounceHeightBasedOn` | character | what sets how high the arc rises |
| `colorBy` | character | what sets the arc colour |
| `colorByNumReadsThreshold` | numeric | split point for read-count colouring |
| `labelWith` | character | what the arc label shows |
| `labelWithInParen` | character | a second value, in parentheses |

The three `...BasedOn` and `colorBy` options take a fixed vocabulary:

* `thicknessBasedOn`: `"numUniqueReads"`, `"numReads"`,
  `"isAnnotatedJunction"`
* `bounceHeightBasedOn`: `"random"`, `"distance"`, `"thickness"`
* `colorBy`: `"numUniqueReads"`, `"numReads"`, `"isAnnotatedJunction"`,
  `"strand"`, `"motif"`

```{r junctions}
junctions.url <- paste0(
    "https://raw.githubusercontent.com/igvteam/igv-data/main/data/test/",
    "splice_junctions/sampleA.SJ.out.bed.gz"
    )

loadSpliceJunctionTrackFromURL(
    session, id = "igv", trackName = "sampleA junctions",
    url = junctions.url, indexURL = paste0(junctions.url, ".tbi"),
    trackHeight = 150,
    trackConfig = list(
        colorBy = "motif",
        labelWith = "uniquelyMapped",
        minUniquelyMappedReads = 5,
        hideAnnotatedJunctions = FALSE
        )
    )
```

The filter names come from the igv.js source rather than from the
`spliceJunctionTrack.html` example page that ships with it: that example sets
`labelUniqueReadCount` and four siblings, and no such option exists in the
library any more.

# Options on startup tracks

Tracks passed to `igvShiny()` through `tracks` are checked against the same
allowlist, so everything above applies to them too. What they additionally need
is a `url`, since there is no loader argument to carry the data: an entry
without a usable one is dropped with a warning.

```{r startup}
gff3.url <- paste0(
    "https://s3.amazonaws.com/igv.org.genomes/hg38/",
    "Homo_sapiens.GRCh38.94.chr.gff3.gz"
    )

startupTrack <- list(
    name = "genes",
    type = "annotation",
    format = "gff3",
    url = gff3.url,
    indexed = FALSE,
    displayMode = "EXPANDED"
    )

widget <- igvShiny(genomeOptions, tracks = list(startupTrack))
str(widget$x$tracks[[1]])
```

In an app that goes inside `renderIgvShiny()`:

```{r startup-app, eval=FALSE}
output$igv <- renderIgvShiny({
    igvShiny(genomeOptions, tracks = list(startupTrack))
    })
```

# When an option seems to do nothing

**Check for the warning first.** A dropped key says so on the R console. If
there is no warning, the option reached igv.js and the problem is elsewhere.

**An option igv.js does not know is ignored in silence** once past the
allowlist. igv.js reads the keys it recognises and never complains about the
rest, so a correct-looking option from IGV desktop, or from an older igv.js,
produces a track drawn with defaults and no message anywhere.

**Some options apply to one track type only.** `autoscale` on an annotation
track, `colorBy = "motif"` on anything that is not a junction track: both are
accepted, and both do nothing.

When you need an option this vignette does not list, the
[igv.js wiki](https://github.com/igvteam/igv.js/wiki) documents the underlying
browser. If it is genuinely missing from the allowlist rather than misspelled,
that is worth [an issue](https://github.com/gladkia/igvShiny/issues) — the list
covers the options that are useful from R, and it grows.

# Session Info

```{r session-info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genomeSpec.R
\name{get_css_genomes}
\alias{get_css_genomes}
\title{get_css_genomes}
\usage{
get_css_genomes(test = FALSE)
}
\arguments{
\item{test}{logical(1) defaults to FALSE}
}
\value{
an list of short genome codes, e.g., "hg38", "dm6", "tair10"
}
\description{
a helper function for mostly internal use,
obtains the genome codes (e.g. 'hg38') supported by igv.js
}
\examples{
css <- get_css_genomes(test = TRUE)

}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.registerTrackFile}
\alias{.registerTrackFile}
\title{Have a path removed when the session ends}
\usage{
.registerTrackFile(session, path)
}
\arguments{
\item{session}{a shiny session object}

\item{path}{character string, the path to remove when the session ends}
}
\value{
the path, unchanged
}
\description{
The companion of \code{.trackFile}, for the files a loader does
not name itself: \code{rtracklayer::export(format = "BAM")} writes an index
next to the bam it was given, and that index is served and leaks the same
way. The path need not exist yet, or ever - \code{unlink()} does not mind.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.trackFile}
\alias{.trackFile}
\title{Name a file in the served tracks directory, and have it removed with the
session that asked for it}
\usage{
.trackFile(session, ext)
}
\arguments{
\item{session}{a shiny session object}

\item{ext}{character string, the extension of the served file}
}
\value{
string with the path to the file
}
\description{
Loaders write a file per track load and nothing ever removes
it. In an interactive session that is harmless - the default directory sits
under \code{tempdir()} and goes with the process - but a deployed app keeps
one R process across many user sessions, and \code{TRACKS_DIR} may point
outside \code{tempdir()} altogether, where nothing removes the files at all.
Alignment loads make it concrete: a bam export or a cram copy is gigabytes.

The paths handed out to one session are collected in that session's
\code{userData}, and the first call registers a single
\code{session$onSessionEnded()} hook to unlink the set. Assignment goes
through a local binding to the \code{userData} environment on purpose:
\code{session$userData$x <- value} would call the assignment method of a
module's session proxy, which refuses to be written to.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadVcfTrack}
\alias{loadVcfTrack}
\title{load a VCF (variant) track provided as a Bioconductor
VariantAnnotation object}
\usage{
loadVcfTrack(
  session,
  id,
  trackName,
  vcfData,
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{vcfData}{VariantAnnotation object}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration options.}
}
\value{
nothing
}
\description{
load a VCF (variant) track provided as a Bioconductor
VariantAnnotation object
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "demos", "igvShinyDemo-withVCF.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadBedTrack}
\alias{loadBedTrack}
\title{load a bed track provided as a data.frame}
\usage{
loadBedTrack(
  session,
  id,
  trackName,
  tbl,
  color = "",
  trackHeight = 50,
  deleteTracksOfSameName = TRUE,
  quiet = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{tbl}{data.frame, with at least "chrom" "start" "end" columns}

\item{color}{character string, a legal CSS color, or "random",
"gray" by default}

\item{trackHeight}{an integer, 50 (pixels) by default}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{quiet}{logical, default TRUE, controls verbosity}

\item{trackConfig}{a named list of additional igv.js track configuration
options.}
}
\value{
nothing
}
\description{
load a bed track provided as a data.frame
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.tracksDir}
\alias{.tracksDir}
\title{Get the tracks directory and make sure shiny serves that very directory}
\usage{
.tracksDir()
}
\value{
string with the path to the tracks directory.
}
\description{
The "tracks" resource path is registered once, when the package
is loaded. The default directory lives under \code{tempdir()}, which is not
guaranteed to stay put for the life of the process: when it moves, files are
written to the new directory while shiny still serves the old one, and igv.js
gets a 404 for a file that exists on disk. Re-registering the path before
each write keeps the served directory and the written directory the same.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.stageTrackFile}
\alias{.stageTrackFile}
\title{Make a file on disk reachable by igv.js}
\usage{
.stageTrackFile(path, ext)
}
\arguments{
\item{path}{character string, an existing file}

\item{ext}{character string, the extension of the served file}
}
\value{
string with the path to the file, relative to the shiny app
}
\description{
Files igv.js reads by url have to sit in the directory shiny
serves as "tracks". Alignment files run to gigabytes, so the file is linked
rather than copied where the filesystem allows it (windows, and any mount
refusing links, falls back to a copy). The name is randomized: two loaders
may be handed same-named files from different directories.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genomeSpec.R
\name{parseAndValidateGenomeSpec}
\alias{parseAndValidateGenomeSpec}
\title{parseAndValidateGenomeSpec}
\usage{
parseAndValidateGenomeSpec(
  genomeName,
  initialLocus = "all",
  stockGenome = TRUE,
  dataMode = NA,
  fasta = NA,
  fastaIndex = NA,
  genomeAnnotation = NA
)
}
\arguments{
\item{genomeName}{character usually one short code of a supported ("stock")
genome (e.g., "hg38") or for a user-supplied custom genome,
the name you wish to use}

\item{initialLocus}{character default "all", otherwise "chrN:start-end"
or a recognized gene symbol}

\item{stockGenome}{logical default TRUE}

\item{dataMode}{character either "stock", "localFile" or "http"}

\item{fasta}{character when supplying a custom (non-stock) genome,
either a file path or a URL}

\item{fastaIndex}{character when supplying a custom (non-stock) genome,
either a file path or a URL,
essential for all but the very small custom genomes.}

\item{genomeAnnotation}{character when supplying a custom (non-stock) genome,
a file path or URL pointing to a genome annotation file in
a gff3 format}
}
\value{
an options list directly usable by igvApp.js, and thus igv.js
}
\description{
a helper function for internal use by the igvShiny constructor,
but possible also of use to those building an igvShiny app,
to test their genome specification for validity
}
\examples{
genomeSpec <-
  parseAndValidateGenomeSpec("hg38", "APOE")  # the simplest case
# a custom genome, read from the files shipped with the package: every
# resource is checked for existence, and "http" would check it over the
# network on each R CMD check
base.dir <-
  system.file(package = "igvShiny", "extdata", "sarsGenome")
fasta.file <-
  file.path(base.dir, "Sars_cov_2.ASM985889v3.dna.toplevel.fa")
fastaIndex.file <- paste0(fasta.file, ".fai")
annotation.file <-
  file.path(base.dir, "Sars_cov_2.ASM985889v3.101.gff3")
custom.genome.title <- "SARS-CoV-2"
genomeOptions <-
  parseAndValidateGenomeSpec(
    genomeName = custom.genome.title,
    initialLocus = "all",
    stockGenome = FALSE,
    dataMode = "localFiles",
    fasta = fasta.file,
    fastaIndex = fastaIndex.file,
    genomeAnnotation = annotation.file
  )

}
\seealso{
[get_css_genomes()] for stock genomes we support.
}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{.packJunctionAttributes}
\alias{.packJunctionAttributes}
\title{Pack junction attributes into a bed name column}
\usage{
.packJunctionAttributes(tbl)
}
\arguments{
\item{tbl}{A data.frame of junctions.}
}
\value{
A character vector, one packed name field per row.
}
\description{
Pack junction attributes into a bed name column
}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GWASTrack.R
\name{.sanitizeChromosomeColorMap}
\alias{.sanitizeChromosomeColorMap}
\title{Check a chromosome color map before it reaches igv.js}
\usage{
.sanitizeChromosomeColorMap(map)
}
\arguments{
\item{map}{a named list or character vector of colors, possibly empty}
}
\value{
the map as a named list, empty if nothing was supplied
}
\description{
Check a chromosome color map before it reaches igv.js
}
\keyword{GWASTrack_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genomeSpec.R
\name{get_cas_genomes}
\alias{get_cas_genomes}
\title{get_cas_genomes}
\usage{
get_cas_genomes()
}
\value{
an list of short genome codes, e.g., "hg38", "dm6"
}
\description{
a helper function for common always available stock genomes,
obtains the genome codes (e.g. 'hg38')
}
\examples{
cas <- get_cas_genomes()

}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GWASTrack.R
\docType{class}
\name{GWASTrack-class}
\alias{GWASTrack-class}
\alias{.GWASTrack}
\alias{GWASTrack}
\title{Constructor for GWASTrack}
\usage{
GWASTrack(
  trackName,
  data,
  chrom.col,
  pos.col,
  pval.col,
  chromosomeColorMap = list(),
  trackHeight = 50,
  autoscale = TRUE,
  minY = 0,
  maxY = 30
)
}
\arguments{
\item{trackName}{A character string, used as track label by igv,
we recommend unique names per track.}

\item{data}{a data.frame or a url pointing to one,
whose structure is described by chrom.col, pos.col, pval.col}

\item{chrom.col}{numeric, the column number of the chromosome column}

\item{pos.col}{numeric, the column number of the position column}

\item{pval.col}{numeric, the column number of the GWAS pvalue column}

\item{chromosomeColorMap}{a named list or character vector mapping
chromosome names, as they are spelled in the data, to colors,
e.g. \code{list("1" = "red", "2" = "blue")}. A \code{"*"} entry colors every
chromosome not named explicitly. Empty by default, which leaves the igv.js
chromosome palette in place}

\item{trackHeight}{numeric in pixels}

\item{autoscale}{logical}

\item{minY}{numeric for explicit (non-auto) scaling}

\item{maxY}{numeric for explicit (non-auto) scaling}
}
\value{
A GWASTrack object
}
\description{
\code{GWASTrack} creates an \code{IGV} manhattan track from GWAS data
}
\examples{
file <-
  # a local gwas file
  system.file(package = "igvShiny", "extdata", "gwas-5k.tsv.gz")
tbl.gwas <- read.table(file,
                       sep = "\t",
                       header = TRUE,
                       quote = "")
dim(tbl.gwas)
track <-
  GWASTrack(
    "gwas 5k",
    tbl.gwas,
    chrom.col = 12,
    pos.col = 13,
    pval.col = 28
  )
getUrl(track)

# a remote gwas file: the constructor checks that the url resolves, so this
# block reaches the network and stays out of R CMD check
\donttest{
url <- "https://gladki.pl/igvShiny/gwas_sample.tsv.gz"
track <- GWASTrack(
  "remote url gwas",
  url,
  chrom.col = 3,
  pos.col = 4,
  pval.col = 10,
  autoscale = FALSE,
  minY = 0,
  maxY = 300,
  trackHeight = 100
)
getUrl(track)
}

# colors picked per chromosome, with "*" covering the rest
track <- GWASTrack(
  "gwas 5k, custom colors",
  tbl.gwas,
  chrom.col = 12,
  pos.col = 13,
  pval.col = 28,
  chromosomeColorMap = list("1" = "red", "2" = "blue", "*" = "gray")
)


}
\keyword{GWASTrack_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadSpliceJunctionTrackFromURL}
\alias{loadSpliceJunctionTrackFromURL}
\title{load a splice junction track served up by http}
\usage{
loadSpliceJunctionTrackFromURL(
  session,
  id,
  trackName,
  url,
  indexURL = "",
  trackHeight = 100,
  displayMode = "COLLAPSED",
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{url}{character string http url for the bed file of junctions}

\item{indexURL}{character string http url for a tabix index, needed only
for a bgzipped bed; "" by default, which loads the file whole}

\item{trackHeight}{an integer, 100 (pixels) by default}

\item{displayMode}{character, "COLLAPSED", "EXPANDED" or "SQUISHED"}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration
options. The junction ones are read straight off it:
\code{minUniquelyMappedReads}, \code{minTotalReads},
\code{maxFractionMultiMappedReads}, \code{minSplicedAlignmentOverhang},
\code{thicknessBasedOn}, \code{bounceHeightBasedOn}, \code{colorBy},
\code{labelWith}, \code{hideAnnotatedJunctions},
\code{hideUnannotatedJunctions}, \code{hideMotifs} among them.}
}
\value{
nothing
}
\description{
load splice junctions from a BED file reachable by URL, as
written by STAR (\code{SJ.out.tab} converted to BED). Six columns, with the
per-junction attributes packed into the name column as \code{key=value}
pairs separated by semicolons: \code{motif}, \code{uniquely_mapped},
\code{multi_mapped}, \code{maximum_spliced_alignment_overhang} and
\code{annotated_junction}. The track reads its filters and labels from
those attributes.

igv.js draws junctions from a file of that shape only - it derives none of
them from a bam file, and it has no sashimi plot. Show this track above a
coverage track (bigWig, bedGraph) for the same sample to get the
arcs-over-coverage view sashimi plots are wanted for.
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadGwasTrack}
\alias{loadGwasTrack}
\title{load a GWAS (genome-wide association study) track
provided as a data.frame}
\usage{
loadGwasTrack(
  session,
  id,
  trackName,
  tbl.gwas,
  ymin = 0,
  ymax = 35,
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{tbl.gwas}{data.frame, with at least "chrom" "start" "end" columns}

\item{ymin}{numeric defaults to 0}

\item{ymax}{numeric defaults to 35}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration
options. A table whose headers igv.js does not recognize needs its layout
spelled out, 1-based, as
\code{columns = list(chromosome = 3, position = 4, value = 10)};
\code{colorTable} takes a chromosome-to-color map, see
\code{\link{GWASTrack}}}
}
\value{
nothing
}
\description{
load a GWAS (genome-wide association study) track
provided as a data.frame
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{.sanitizeTrack}
\alias{.sanitizeTrack}
\title{Sanitize one startup track specification}
\usage{
.sanitizeTrack(track, depth)
}
\arguments{
\item{track}{A named list, one igv.js track configuration.}

\item{depth}{An integer, how deep this track sits in a merged track.}
}
\value{
The sanitized track, or NULL if it cannot be used.
}
\description{
Sanitize one startup track specification
}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{showGenomicRegion}
\alias{showGenomicRegion}
\alias{getGenomicRegion}
\title{focus igv on a region}
\usage{
showGenomicRegion(session, id, region)

getGenomicRegion(session, id)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{region}{a character string, either e.g. "chr5:92,221,640-92,236,523"
or "MEF2C"}
}
\value{
a character string of format "chrom:start-end"
}
\description{
zoom in or out to show the nominated region, by chromosome locus
or gene symbol

return the current region displayed by your igv instance
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadSpliceJunctionTrackFromLocalData}
\alias{loadSpliceJunctionTrackFromLocalData}
\title{load a splice junction track from a data.frame}
\usage{
loadSpliceJunctionTrackFromLocalData(
  session,
  id,
  trackName,
  tbl,
  trackHeight = 100,
  displayMode = "COLLAPSED",
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{tbl}{data.frame, with at least "chrom" "start" "end" columns, in bed
coordinates: a 0-based start and an exclusive end}

\item{trackHeight}{an integer, 100 (pixels) by default}

\item{displayMode}{character, "COLLAPSED", "EXPANDED" or "SQUISHED"}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration
options, the junction ones among them; see
\code{\link{loadSpliceJunctionTrackFromURL}}}
}
\value{
nothing
}
\description{
load splice junctions held in R. The table is written as the
six-column bed igv.js draws junctions from, and served from the same
directory as the other local-data tracks.

Coordinates are taken as bed coordinates and written out unchanged, as the
other \code{*FromLocalData} loaders take theirs. STAR reports the first and
last intron base of a junction 1-based in \code{SJ.out.tab}, so a table read
straight from that file needs its start shifted by one first; a bed already
converted from it, which is what STAR wrappers and the igv.js test fixtures
ship, is passed in as is.

Columns \code{chrom} (or \code{chr}), \code{start} and \code{end} are
required. \code{score} and \code{strand} are used when present. The
per-junction attributes \code{motif}, \code{uniquely_mapped},
\code{multi_mapped}, \code{maximum_spliced_alignment_overhang} and
\code{annotated_junction} are packed into the bed name column, which is
where the track reads its filters and labels from; supply the ones you
want to filter or label on. A \code{name} column, if you have already
packed it yourself, is written through untouched.

Without \code{uniquely_mapped} or an explicit \code{score} the arcs all
draw at the same thickness: igv.js has nothing to size them by.
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadSegTrack}
\alias{loadSegTrack}
\alias{loadSEGTrack}
\title{load a seg track provided as a data.frame}
\usage{
loadSegTrack(
  session,
  id,
  trackName,
  tbl,
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{tbl}{data.frame, with at least "chrom" "start" "end" "score" columns}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration options.}
}
\value{
nothing
}
\description{
load a SEG track provided as a data.frame.  igv "displays
segmented data as a blue-to-red heatmap where the data range is
-1.5 to 1.5... The segmented data file format is the output of
the Circular Binary Segmentation algorithm (Olshen et al., 2004)".
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "demos", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GWASTrack.R
\name{show,GWASTrack-method}
\alias{show,GWASTrack-method}
\title{Display a short summary of a GWASTrack object}
\usage{
\S4method{show}{GWASTrack}(object)
}
\arguments{
\item{object}{An object of class GWASTrack}
}
\value{
nothing; called for the side effect of printing a summary
}
\description{
Display a short summary of a GWASTrack object
}
\examples{
file <-
  system.file(package = "igvShiny", "extdata", "gwas-5k.tsv.gz")
tbl.gwas <- read.table(file, sep = "\t", header = TRUE, quote = "")
track <-
  GWASTrack("gwas 5k", tbl.gwas, chrom.col = 12, pos.col = 13, pval.col = 28)
track

}
\keyword{GWASTrack_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadCramTrackFromLocalData}
\alias{loadCramTrackFromLocalData}
\title{load a cram file sitting on the same machine as the shiny app}
\usage{
loadCramTrackFromLocalData(
  session,
  id,
  trackName,
  cramFile,
  indexFile = paste0(cramFile, ".crai"),
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{cramFile}{character string, path to a cram file}

\item{indexFile}{character string, path to its crai index,
by default the cram file with ".crai" appended}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration
options, \code{sort} among them; see \code{\link{loadBamTrackFromURL}}.}
}
\value{
nothing
}
\description{
load a local cram track. Unlike \code{loadBamTrackFromLocalData}
this loader takes file paths, not an R object: no bioconductor package parses
cram, so the file and its index are handed to igv.js untouched, through the
directory shiny serves as "tracks".
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadBedGraphTrack}
\alias{loadBedGraphTrack}
\alias{loadGenomeAnnotationTrack}
\title{load a scored genome annotation track provided as a data.frame}
\usage{
loadBedGraphTrack(
  session,
  id,
  trackName,
  tbl,
  color = "gray",
  trackHeight = 30,
  autoscale,
  autoscaleGroup = -1,
  min = NA_real_,
  max = NA_real_,
  deleteTracksOfSameName = TRUE,
  quiet = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{tbl}{data.frame, with at least "chrom" "start" "end" "score" columns}

\item{color}{character string, a legal CSS color, or "random",
"gray" by default}

\item{trackHeight}{an integer, 30 (pixels) by default}

\item{autoscale}{logical}

\item{autoscaleGroup}{numeric(1) defaults to -1}

\item{min}{numeric, consulted when autoscale is FALSE}

\item{max}{numeric, consulted when autoscale is FALSE}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{quiet}{logical, default TRUE, controls verbosity}

\item{trackConfig}{a named list of additional igv.js track configuration options.}
}
\value{
nothing
}
\description{
load a genome annotation track provided as a data.frame
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "demos", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadBamTrackFromLocalData}
\alias{loadBamTrackFromLocalData}
\title{load GenomicAlignments data as an igv.js alignment track}
\usage{
loadBamTrackFromLocalData(
  session,
  id,
  trackName,
  data,
  deleteTracksOfSameName = TRUE,
  displayMode = "EXPANDED",
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{data}{GenomicAlignments object}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{displayMode}{character string, possible values are "EXPANDED"(default),
"SQUISHED" or "COLLAPSED"}

\item{trackConfig}{a named list of additional igv.js track configuration
options, \code{sort} among them; see \code{\link{loadBamTrackFromURL}}.}
}
\value{
nothing
}
\description{
load GenomicAlignments data  as an igv.js alignment track
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genomeSpec.R
\name{get_basic_genomes}
\alias{get_basic_genomes}
\title{get_basic_genomes}
\usage{
get_basic_genomes()
}
\value{
an list of short genome codes, e.g., "hg38", "dm6"
}
\description{
a helper function for basic genomes,
obtains the genome codes (e.g. 'hg38')
}
\examples{
bs <- get_basic_genomes()

}
\keyword{utils}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GWASTrack.R
\name{display,GWASTrack-method}
\alias{display,GWASTrack-method}
\alias{display}
\title{display the already constructed and configured track}
\usage{
\S4method{display}{GWASTrack}(obj, session, id, deleteTracksOfSameName = TRUE)
}
\arguments{
\item{obj}{An object of class GWASTrack}

\item{session}{a Shiny session object}

\item{id}{character the identifier of the target igv object in the browser}

\item{deleteTracksOfSameName}{logical to avoid duplications in track names}
}
\value{
nothing
}
\description{
display the already constructed and configured track
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{GWASTrack_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadCramTrackFromURL}
\alias{loadCramTrackFromURL}
\title{load a cram track which, with index, is served up by http}
\usage{
loadCramTrackFromURL(
  session,
  id,
  trackName,
  cramURL,
  indexURL,
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{cramURL}{character string http url for the bam file,
typically very large}

\item{indexURL}{character string http url for the bam file index,
typically small}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration
options, \code{sort} among them; see \code{\link{loadBamTrackFromURL}}.}
}
\value{
nothing
}
\description{
load a remote cram track
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadBamTrackFromURL}
\alias{loadBamTrackFromURL}
\title{load a bam track which, with index, is served up by http}
\usage{
loadBamTrackFromURL(
  session,
  id,
  trackName,
  bamURL,
  indexURL,
  deleteTracksOfSameName = TRUE,
  displayMode = "EXPANDED",
  showAllBases = FALSE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{bamURL}{character string http url for the bam file,
typically very large}

\item{indexURL}{character string http url for the bam file index,
typically small}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{displayMode}{character string, possible values are "EXPANDED"(default),
"SQUISHED" or "COLLAPSED"}

\item{showAllBases}{logical, show all bases in the alignment, default FALSE}

\item{trackConfig}{a named list of additional igv.js track configuration
options. Alignment tracks read \code{sort}: \code{list(sort =
list(chr = "chr1", position = 155160540, option = "TAG", tag = "HP"))}
sorts the reads at that 1-based position by the HP tag. Any igv.js sort
option works (BASE, STRAND, INSERT_SIZE, ...); reads sort descending
unless \code{direction = "ASC"}.}
}
\value{
nothing
}
\description{
load a remote bam track
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GWASTrack.R
\name{getUrl,GWASTrack-method}
\alias{getUrl,GWASTrack-method}
\alias{getUrl}
\title{the url of the gwas table}
\usage{
\S4method{getUrl}{GWASTrack}(obj)
}
\arguments{
\item{obj}{An object of class GWASTrack}
}
\value{
character
}
\description{
the url of the gwas table
}
\examples{
file <-
  # a local gwas file
  system.file(package = "igvShiny", "extdata", "gwas-5k.tsv.gz")
tbl.gwas <- read.table(file,
                       sep = "\t",
                       header = TRUE,
                       quote = "")
track <-
  GWASTrack(
    "gwas 5k",
    tbl.gwas,
    chrom.col = 12,
    pos.col = 13,
    pval.col = 28
  )
getUrl(track)

}
\keyword{GWASTrack_class}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadBedGraphTrackFromURL}
\alias{loadBedGraphTrackFromURL}
\title{load a bedgraph track from a URL}
\usage{
loadBedGraphTrackFromURL(
  session,
  id,
  trackName,
  url,
  color = "gray",
  trackHeight = 30,
  autoscale = TRUE,
  min = 0,
  max = 1,
  autoscaleGroup = -1,
  deleteTracksOfSameName = TRUE,
  quiet = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{url}{character}

\item{color}{character string, a legal CSS color, or "random",
"gray" by default}

\item{trackHeight}{an integer, 30 (pixels) by default}

\item{autoscale}{logical}

\item{min}{numeric, consulted when autoscale is FALSE}

\item{max}{numeric, consulted when autoscale is FALSE}

\item{autoscaleGroup}{numeric(1) defaults to -1}

\item{deleteTracksOfSameName}{logical(1) defaults to TRUE}

\item{quiet}{logical, default TRUE, controls verbosity}

\item{trackConfig}{a named list of additional igv.js track configuration options.}
}
\value{
nothing
}
\description{
load a bedgraph track provided as a data.frame
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "demos", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{renderIgvShiny}
\alias{renderIgvShiny}
\title{draw the igv genome browser element}
\usage{
renderIgvShiny(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{expr}{an expression that generates an HTML widget}

\item{env}{the environment in which to evaluate expr}

\item{quoted}{logical flag indicating if expr a quoted expression}
}
\value{
an output or render function that enables the use of the widget
within Shiny applications
}
\description{
This function is called in the server function of your shiny app
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{igvShinyOutput}
\alias{igvShinyOutput}
\title{create the UI for the widget}
\usage{
igvShinyOutput(outputId, width = "100\%", height = NULL)
}
\arguments{
\item{outputId}{a character string, specifies the html element id}

\item{width}{a character string, standard css notations,
either e.g., "1000px" or "95\%", "100\%" by default}

\item{height}{a character string, needs to be an explicit pixel measure,
e.g., "800px", "400px" by default}
}
\value{
the created widget's html
}
\description{
This function is called in the ui function of your shiny app
}
\examples{
io <- igvShinyOutput("igvOut")

}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{removeTracksByName}
\alias{removeTracksByName}
\title{remove tracks from the browser}
\usage{
removeTracksByName(session, id, trackNames)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackNames}{a vector of character strings}
}
\value{
nothing
}
\description{
delete tracks on the browser
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{loadGFF3TrackFromURL}
\alias{loadGFF3TrackFromURL}
\title{load a GFF3 track which, with index, is served up by http}
\usage{
loadGFF3TrackFromURL(
  session,
  id,
  trackName,
  gff3URL,
  indexURL,
  color = "gray",
  colorTable,
  colorByAttribute,
  displayMode,
  trackHeight = 50,
  visibilityWindow,
  deleteTracksOfSameName = TRUE,
  trackConfig = list()
)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}

\item{trackName}{character string}

\item{gff3URL}{character string http url for the bam file,
typically very large}

\item{indexURL}{character string http url for the bam file index,
typically small}

\item{color}{character #RGB or a recognized color name.  ignored if
colorTable and colorByAttribute provided}

\item{colorTable}{list, mapping a gff3 attribute, typically biotype,
to a color}

\item{colorByAttribute}{character, name of a gff3 attribute in column 9,
typically "biotype"}

\item{displayMode}{character,  "EXPANDED",  "SQUISHED" or "COLLAPSED"}

\item{trackHeight}{numeric defaults to 50}

\item{visibilityWindow}{numeric, Maximum window size in base pairs
for which indexed annotations or variants are displayed}

\item{deleteTracksOfSameName}{logical, default TRUE}

\item{trackConfig}{a named list of additional igv.js track configuration
options.}
}
\value{
nothing
}
\description{
load a remote GFF3 track
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "demos", "local-data.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{track_loaders}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{igvShiny}
\alias{igvShiny}
\title{Create an igvShiny instance}
\usage{
igvShiny(
  genomeOptions,
  width = NULL,
  height = NULL,
  elementId = NULL,
  displayMode = "squished",
  tracks = list()
)
}
\arguments{
\item{genomeOptions}{a list with these fields: genomeName, initialLocus,
annotation, dataMode, fasta, fastaIndex, stockGenome, validated}

\item{width}{a character string, standard css notations,
either e.g., "1000px" or "95\%"}

\item{height}{a character string, needs to be an explicit pixel measure,
e.g., "800px"}

\item{elementId}{a character string, the html element id within which
igv is created}

\item{displayMode}{a character string, default "SQUISHED".}

\item{tracks}{a list of track specifications to be created and displayed
at startup. Each element is itself a named list of igv.js track options
(e.g. \code{name}, \code{type}, \code{format}, \code{url}), for example:
\code{list(list(name="genes", type="annotation", format="gff3",
url="https://.../genes.gff3"))}. Unrecognized keys are dropped with a
warning; see \code{.validIgvTrackOptions} for the full allowlist.}
}
\value{
the created widget
}
\description{
This function is called in the server function of your shiny app
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{removeUserAddedTracks}
\alias{removeUserAddedTracks}
\title{remove only those tracks explicitly added by your app}
\usage{
removeUserAddedTracks(session, id)
}
\arguments{
\item{session}{an environment or list, provided and managed by shiny}

\item{id}{character string, the html element id of this widget instance}
}
\value{
nothing
}
\description{
remove only those tracks explicitly added by your app.
stock tracks (i.e., #' Refseq Genes) remain
}
\examples{
library(igvShiny)
demo_app_file <-
  system.file(package = "igvShiny", "showcase", "igvShinyDemo.R")
if (interactive()) {
  shiny::runApp(demo_app_file)
}

}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{.sanitizeAndMergeOptions}
\alias{.sanitizeAndMergeOptions}
\title{Sanitize and merge track configuration options}
\usage{
.sanitizeAndMergeOptions(baseOptions, userOptions)
}
\arguments{
\item{baseOptions}{A list of default options set by the R function.}

\item{userOptions}{A list of options provided by the user via trackConfig.}
}
\value{
A merged and sanitized list of options ready to be sent to
JavaScript.
}
\description{
Sanitize and merge track configuration options
}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igvShiny.R
\name{.sanitizeTracks}
\alias{.sanitizeTracks}
\title{Sanitize a list of startup track specifications}
\usage{
.sanitizeTracks(tracks, depth = 1L)
}
\arguments{
\item{tracks}{A list of named lists, each an igv.js track configuration.}

\item{depth}{An integer, how deep this list sits in a merged track; 1 for
the startup list itself.}
}
\value{
A sanitized list of track configurations; invalid entries or keys
are dropped with a warning.
}
\description{
Sanitize a list of startup track specifications
}
\keyword{igvShiny}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_tracks_dir}
\alias{get_tracks_dir}
\title{get_tracks_dir
Get the directory where tracks are stored. The directory can be defined
with environmental variable.
If not defined the default is a directory called "tracks" in the temp
directory.
We need a local directory to write files - for instance,
a vcf file representing a genomic region of interest.
We then tell shiny about that directory, so that shiny's built-in http server
can serve up files we write there, ultimately consumed by igv.js}
\usage{
get_tracks_dir(env_var = "TRACKS_DIR")
}
\arguments{
\item{env_var}{The name of the environmental variable to use.}
}
\value{
string with the path to the tracks directory.
}
\description{
get_tracks_dir
Get the directory where tracks are stored. The directory can be defined
with environmental variable.
If not defined the default is a directory called "tracks" in the temp
directory.
We need a local directory to write files - for instance,
a vcf file representing a genomic region of interest.
We then tell shiny about that directory, so that shiny's built-in http server
can serve up files we write there, ultimately consumed by igv.js
}
\examples{
gtd <- get_tracks_dir(env_var = "TRACKS_DIR")

}
\keyword{utils}
